Attribute VB_Name = "PictureFunctions"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E0698F7002B"
Option Explicit

'picture resource global variables
Private VisBlankData(26879) As Byte, PriBlankData(26879) As Byte
Private VisBuildData() As Byte, PriBuildData() As Byte

Private agPicData() As Byte
Private lngPos As Long
Private lngEndPos As Long

Private bytIn As Integer
Private CurrentPen As PenStatus
Private SavePen As PenStatus

Private Queue(26879) As Long

Private Enum CornerDirection
  cdX
  cdY
End Enum

Private CircleData(63) As Byte
Private InitPlotData As Boolean

Private Sub DrawPixel(ByVal xPos As Long, ByVal yPos As Long)

  On Error Resume Next
  
  If xPos + yPos * 160 <= 26879 Then
    If CurrentPen.VisColor < agNone Then
      VisBuildData(xPos + yPos * 160) = CurrentPen.VisColor
    End If
    If CurrentPen.PriColor < agNone Then
      PriBuildData(xPos + yPos * 160) = CurrentPen.PriColor
    End If
  End If
End Sub

Public Function GetColVal(lngEGAIn As Long) As Long
  'this one can't really be explained
  'basically it attempts to convert a long color Value
  'into the corresponding AGI color index with the least
  'amount of calculations
  'it is an empirically derived algorithm
  
  'NOTE: If this method is called with a color Value
  'other than the defined EGA color values for AGI, then
  'I have absolutely no idea what the return Value may
  'look like.
  
  Dim cR As Long, cG As Long, cB As Long
  Dim vR As Long, vG As Long, vB As Long
  
  'split the color up
  
  cR = lngEGAIn Mod 256&
  cG = (lngEGAIn \ 256&) Mod 256&
  cB = (lngEGAIn \ 65536) Mod 256&
  
  
  'convert to component numbers
  If cR = &HFF Then
    vR = 32
  ElseIf cR = 0 Then
    vR = 0
  Else
    vR = 16
  End If
  
  If cG = &HFF Then
    vG = 32
  ElseIf cG = 0 Then
    vG = 0
  Else
    vG = 16
  End If
  
  If cB = &HFF Then
    vB = 32
  ElseIf cB = 0 Then
    vB = 0
  Else
    vB = 16
  End If
  
  'build composite
  GetColVal = (vB + vG * 2 + vR * 4) / 16
  
  'if <5
  If GetColVal < 5 Then
    Exit Function
  End If
  
  'if >9
  If GetColVal > 9 Then
    GetColVal = GetColVal + 1
    Exit Function
  End If
  
  'if red is >80(&H50)
  If cR > &H50& Then
    Exit Function
  End If
  
  'three cases left:
  '7,8,5 corresponding to 8,9,10
  If GetColVal > 6 Then
    GetColVal = GetColVal + 1
    Exit Function
  End If
  
  'only one left is light green
  GetColVal = 10
End Function
Public Sub BuildBMPs(VisData() As Byte, PriData() As Byte, bytPicData() As Byte, ByVal EndPos As Long, ByVal StatusPos As Long)
  'converts the data extracted from the picture resource into
  'a 256 bit color DIBitmap
  
  Dim tmpX As Integer, tmpY As Integer
  Dim i As Long, j As Long
  
  On Error GoTo ErrHandler
  
  'if plot data not set yet
  If Not InitPlotData Then
    'initialize them
    InitializePlotData
  End If
  
  'if blank data not set yet,
  If VisBlankData(0) <> agWhite Then
    'build blank bitmaps
    For j = 0 To 167
      For i = 0 To 159
        PriBlankData(j * 160 + i) = agRed
        VisBlankData(j * 160 + i) = agWhite
      Next i
    Next j
  End If
  
  'copy picture data locally to increase speed
  'of picture drawing functions
  agPicData = bytPicData
  
  'load data arrays with blank info
  VisBuildData = VisBlankData
  PriBuildData = PriBlankData
  
  'if no end passed,
  If EndPos = -1 Then
    'get size of data
    EndPos = UBound(agPicData)
  End If
  
  'if no status pos passed
  If StatusPos = -1 Then
    'status pos is end pos
    StatusPos = EndPos
  End If
  
  'save endpos locally
  lngEndPos = EndPos
  
  'set default picture values
  With CurrentPen
    .VisColor = agNone
    .PriColor = agNone
    .PlotSize = 0
    .PlotShape = psCircle
    .PlotStyle = psSolid
  End With
  
  lngPos = 0
  
  'read first command byte
  bytIn = agPicData(lngPos)
  lngPos = lngPos + 1
  Do
    Select Case bytIn
    Case &HF6 'Absolute line (long lines).
      DrawAbsLine
      
    Case &HF7 'Relative line (short lines).
      DrawRelLine
      
    Case &HF8 'Fill.
      PicFloodFill
      
    Case &HF4 'Draw a Y corner.
      DrawCorner cdY
      
    Case &HF5 'Draw an X corner.
      DrawCorner cdX
      
    Case &HF9 'Change pen size and style.
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1
      
      CurrentPen.PlotStyle = (bytIn And &H20) / &H20
      CurrentPen.PlotShape = (bytIn And &H10) / &H10
      CurrentPen.PlotSize = bytIn And &H7
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1
      
    Case &HFA 'Plot with pen.
      BrushPlot
      
    Case &HF0 'Change picture color and enable picture draw.
      'get color
      CurrentPen.VisColor = agPicData(lngPos)
      lngPos = lngPos + 1
      'get next command byte
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1

    Case &HF1 'Disable picture draw.
      'disable visual drawing
      CurrentPen.VisColor = agNone
      'get next command byte
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1
      
    Case &HF2 'Change priority color and enable priority draw.
      'get color
      CurrentPen.PriColor = agPicData(lngPos)
      lngPos = lngPos + 1
      'get next command byte
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1

    Case &HF3 'Disable priority draw.
      'disable priority
      CurrentPen.PriColor = agNone
      'get next command byte
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1
      
    Case &HFF 'end of drawing
      'if pen status position is end of drawing
      If StatusPos = EndPos Then
        'save tool status
        SavePen = CurrentPen
        'reset status pos to aviod changing save status
        StatusPos = &H7FFFFFFF
      End If
      Exit Do
      
    Case Else
      On Error GoTo 0
      On Error GoTo 0: Err.Raise vbObjectError + 527, "WinAGI.BuildBMPs", Replace(Replace(LoadResString(527), ARG1, Hex$(bytIn)), "%2", Hex$(lngPos))
      Exit Sub
    End Select
    
    'if at Pen Status position,
    If lngPos > StatusPos Then
      'save tool status
      SavePen = CurrentPen
      'reset statuspos to avoid changing save status
      StatusPos = &H7FFFFFFF
      
    End If
  Loop Until lngPos > EndPos
  
  'if at end of resource, was last command end-of-resource flag?
  If lngPos > UBound(agPicData) Then
    If agPicData(UBound(agPicData)) <> &HFF Then
      On Error GoTo 0: Err.Raise vbObjectError + 682, "PictureFunctions.BuildBMPs", LoadResString(682)
      Exit Sub
    End If
  End If
  'copy resulting data back to calling function
  VisData = VisBuildData
  PriData = PriBuildData
  
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  On Error GoTo 0: Err.Raise vbObjectError + 609, "PictureFunctions.BuildBMPs", Replace(Replace(LoadResString(609), ARG1, CStr(lngError) & ":" & strError), "%2", Hex$(lngPos))
End Sub

Public Function GetToolStatus() As PenStatus
  'NOTE: this method should ONLY be called
  'by the form that just completed a buildbmp call
  'otherwise, the results are meaningless
  
  GetToolStatus = SavePen
End Function


Private Sub InitializePlotData()
  'circle data used by the
  'brush drawing functions to paint
  'pictures
  CircleData(0) = &H80
  CircleData(1) = &HC0
  CircleData(2) = &HC0
  CircleData(3) = &HC0
  CircleData(4) = &H40
  CircleData(5) = &HE0
  CircleData(6) = &HE0
  CircleData(7) = &HE0
  CircleData(8) = &H40
  CircleData(9) = &H60
  CircleData(10) = &H60
  CircleData(11) = &HF0
  CircleData(12) = &HF0
  CircleData(13) = &HF0
  CircleData(14) = &H60
  CircleData(15) = &H60
  CircleData(16) = &H20
  CircleData(17) = &H70
  CircleData(18) = &HF8
  CircleData(19) = &HF8
  CircleData(20) = &HF8
  CircleData(21) = &HF8
  CircleData(22) = &HF8
  CircleData(23) = &H70
  CircleData(24) = &H20
  CircleData(25) = &H30
  CircleData(26) = &H78
  CircleData(27) = &H78
  CircleData(28) = &H78
  CircleData(29) = &HFC
  CircleData(30) = &HFC
  CircleData(31) = &HFC
  CircleData(32) = &H78
  CircleData(33) = &H78
  CircleData(34) = &H78
  CircleData(35) = &H30
  CircleData(36) = &H38
  CircleData(37) = &H7C
  CircleData(38) = &H7C
  CircleData(39) = &H7C
  CircleData(40) = &HFE
  CircleData(41) = &HFE
  CircleData(42) = &HFE
  CircleData(43) = &HFE
  CircleData(44) = &HFE
  CircleData(45) = &H7C
  CircleData(46) = &H7C
  CircleData(47) = &H7C
  CircleData(48) = &H38
  CircleData(49) = &H18
  CircleData(50) = &H3C
  CircleData(51) = &H7E
  CircleData(52) = &H7E
  CircleData(53) = &H7E
  CircleData(54) = &HFF
  CircleData(55) = &HFF
  CircleData(56) = &HFF
  CircleData(57) = &HFF
  CircleData(58) = &HFF
  CircleData(59) = &H7E
  CircleData(60) = &H7E
  CircleData(61) = &H7E
  CircleData(62) = &H3C
  CircleData(63) = &H18
  
  'set flag
  InitPlotData = True
End Sub

Private Sub DrawLine(ByVal X1 As Long, ByVal Y1 As Long, ByVal X2 As Long, ByVal Y2 As Long)
  
  Dim xPos As Long, yPos As Long
  Dim DY As Long, DX As Long
  Dim vDir As Long, hDir As Long
  Dim XC As Long, YC As Long, MaxDelta As Long
  Dim i As Long
  
  'Sierra sucked at checking for overflows; if a bad value for coords was used that
  'overflowed the designated picture buffer, AGI didn't care; it would just draw the pixel
  'in invalid memory and plod on; so we're just going to ignore invalid pixels and
  'plod on too. ugh...
  
  '(would be nice if there was a way to warn the user; not sure I can do that though)
      
  On Error GoTo ErrHandler
  
  'determine delta x/delta y and direction
  DY = Y2 - Y1
  vDir = Sgn(DY)
  DX = X2 - X1
  hDir = Sgn(DX)
  
  'if a point
  If DY = 0 And DX = 0 Then
    'set the point
    DrawPixel X1, Y1
    
  'if horizontal
  ElseIf DY = 0 Then
    For i = X1 To X2 Step hDir
      'set point
      DrawPixel i, Y1
    Next i
  
  'if vertical
  ElseIf DX = 0 Then
    For i = Y1 To Y2 Step vDir
      'set point
      DrawPixel X1, i
    Next i
  
  Else
    'this line drawing function EXACTLY matches the Sierra
    'drawing function
    
    'set the starting point
    DrawPixel X1, Y1
    
    xPos = X1
    yPos = Y1
  
    'invert DX and DY if they are negative
    If DY < 0 Then
      DY = DY * -1
    End If
    If (DX < 0) Then
      DX = DX * -1
    End If
  
    'set up the loop, depending on which direction is largest
    If DX >= DY Then
      MaxDelta = DX
      YC = DX \ 2
      XC = 0
    Else
      MaxDelta = DY
      XC = DY \ 2
      YC = 0
    End If
  
    For i = 1 To MaxDelta
      YC = YC + DY
      If YC >= MaxDelta Then
        YC = YC - MaxDelta
        yPos = yPos + vDir
      End If
    
      XC = XC + DX
      If XC >= MaxDelta Then
        XC = XC - MaxDelta
        xPos = xPos + hDir
      End If
      
      DrawPixel xPos, yPos
    Next i
  End If
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 662, strErrSrc, Replace(LoadResString(662), ARG1, CStr(lngError) & ":" & strError)
End Sub

Private Sub DrawCorner(CurAxis As CornerDirection)
  Dim X1 As Byte, Y1 As Byte
  Dim X2 As Byte, Y2 As Byte

  'read in start coordinates
  X1 = agPicData(lngPos)
  lngPos = lngPos + 1
  Y1 = agPicData(lngPos)
  lngPos = lngPos + 1
  
  'draw first pixel
  DrawLine X1, Y1, X1, Y1
  
  'get next byte
  bytIn = agPicData(lngPos)
  lngPos = lngPos + 1
  
  Do Until bytIn >= &HF0 Or lngPos > lngEndPos
    
    If CurAxis = cdX Then
      X2 = bytIn
      Y2 = Y1
      DrawLine X1, Y1, X2, Y2
      CurAxis = cdY
      X1 = X2
    Else
      Y2 = bytIn
      X2 = X1
      DrawLine X1, Y1, X2, Y2
      CurAxis = cdX
      Y1 = Y2
    End If
    
    'get next byte
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
  Loop
End Sub


Private Sub DrawAbsLine()
  Dim X1 As Byte, Y1 As Byte
  Dim X2 As Byte, Y2 As Byte

  'read in start position
  X1 = agPicData(lngPos)
  lngPos = lngPos + 1
  Y1 = agPicData(lngPos)
  lngPos = lngPos + 1

  'draw first pixel
  DrawLine X1, Y1, X1, Y1
  
  'get next potential coordinate
  bytIn = agPicData(lngPos)
  lngPos = lngPos + 1
  
  Do Until bytIn >= &HF0 Or lngPos > lngEndPos
  
    X2 = bytIn
    Y2 = agPicData(lngPos)
    lngPos = lngPos + 1
    DrawLine X1, Y1, X2, Y2
    X1 = X2
    Y1 = Y2
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
  Loop
End Sub


Private Sub DrawRelLine()
  Dim xdisp As Integer, ydisp As Integer
  Dim X1 As Byte, Y1 As Byte
  Dim rtn As Long
  
  'read in starting position
  X1 = agPicData(lngPos)
  lngPos = lngPos + 1
  Y1 = agPicData(lngPos)
  lngPos = lngPos + 1
  'set pixel of starting position
  DrawPixel X1, Y1
  
  'get next potential command
  bytIn = agPicData(lngPos)
  lngPos = lngPos + 1
  
  Do Until bytIn >= &HF0 Or lngPos > lngEndPos
    'if horizontal high bit set
    If (bytIn And &H80) Then
      'displacement is negative
      xdisp = 0 - ((bytIn And &H70) / &H10)
    Else
      xdisp = ((bytIn And &H70) / &H10)
    End If
    'if vertical high bit set
    If (bytIn And &H8) Then
      'displacement is negative
      ydisp = 0 - (bytIn And &H7)
    Else
      ydisp = (bytIn And &H7)
    End If
    DrawLine X1, Y1, X1 + xdisp, Y1 + ydisp
    X1 = X1 + xdisp
    Y1 = Y1 + ydisp
    
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
  Loop
End Sub


Private Sub BrushPlot()
  Dim PlotX As Byte, PlotY As Byte
  Dim PatternNum As Byte ''', newPatt As Byte
  Dim intBrushCeil As Integer, intBrushFloor As Integer
  Dim rtn As Long
  Dim X As Byte, Y As Byte
  
  On Error GoTo ErrHandler
  
  bytIn = agPicData(lngPos)
  
  lngPos = lngPos + 1
  Do Until bytIn >= &HF0 Or lngPos > lngEndPos
    If CurrentPen.PlotStyle = psSplatter Then
      PatternNum = bytIn Or 1
      
      bytIn = agPicData(lngPos)
      lngPos = lngPos + 1
    End If
    
    'get brush position
    PlotX = bytIn
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
    PlotY = bytIn
    
    'calculate brush ceiling/floor values
    intBrushCeil = (CurrentPen.PlotSize + 1) \ 2
    intBrushFloor = CurrentPen.PlotSize \ 2
    
    'adjust position for edges of picture
    If PlotX < intBrushCeil Then
      PlotX = intBrushCeil
    ElseIf PlotX > 159 - intBrushFloor Then
      PlotX = 159 - intBrushFloor
    End If
    If PlotY < CurrentPen.PlotSize Then
      PlotY = CurrentPen.PlotSize
    ElseIf PlotY > 167 - CurrentPen.PlotSize Then
      PlotY = 167 - CurrentPen.PlotSize
    End If

    'if brush is a circle
    If CurrentPen.PlotShape = psCircle Then
      For Y = PlotY - CurrentPen.PlotSize To PlotY + CurrentPen.PlotSize
        For X = PlotX - intBrushCeil To PlotX + intBrushFloor
          'if pixel is within circle shape,
          If (CircleData(CurrentPen.PlotSize ^ 2 + Y - PlotY + CurrentPen.PlotSize) \ 2 ^ (7 - (CLng(X) - PlotX + intBrushCeil)) And 1) Then
            'if style is splatter
            If CurrentPen.PlotStyle = psSplatter Then
              'only draw if pattern bit is set
'''              newPatt = PatternNum \ 2
'''              If (PatternNum Mod 2) Then
'''                newPatt = newPatt Xor &HB8
'''              End If
'''              PatternNum = newPatt
              If (PatternNum And 1) Then
                PatternNum = PatternNum \ 2 Xor &HB8
              Else
                PatternNum = PatternNum \ 2
              End If
              
              If (PatternNum And 3) = 2 Then
                DrawPixel X, Y
              End If
            Else  'solid
              'set all pixels
              DrawPixel X, Y
            End If
          End If
        Next X
      Next Y
    Else 'square
      For Y = PlotY - CurrentPen.PlotSize To PlotY + CurrentPen.PlotSize
        For X = PlotX - intBrushCeil To PlotX + intBrushFloor
          'if style is splatter
          If CurrentPen.PlotStyle = psSplatter Then
            'only draw if pattern bit is set
'''            newPatt = PatternNum \ 2
'''            If (PatternNum Mod 2) Then
'''              newPatt = newPatt Xor &HB8
'''            End If
'''            PatternNum = newPatt
            If (PatternNum And 1) Then
              PatternNum = PatternNum \ 2 Xor &HB8
            Else
              PatternNum = PatternNum \ 2
            End If
            If (PatternNum And 3) = 2 Then
              DrawPixel X, Y
            End If
          Else  'solid
            'set all pixels
            DrawPixel X, Y
          End If
        Next X
      Next Y
    End If
    
    'get next byte
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
  Loop
  
Exit Sub

ErrHandler:
'''  'if error is due to going past end of data, clear the error and just exit
'''  If lngPos > lngEndPos Then
'''    Err.Clear
'''  Else
    'failure! pass error along
    Debug.Assert False
    Err.Raise Err.Number, Err.Source, Err.Description
'''  End If
End Sub




Private Sub PicFloodFill()
  
  Dim QueueStart As Long, QueueEnd As Long
  Dim lngOffset As Long
  Dim X As Byte, Y As Byte
  
  On Error GoTo ErrHandler
  
  'get next byte
  bytIn = agPicData(lngPos)
  lngPos = lngPos + 1
  
  Do Until bytIn >= &HF0 Or lngPos > lngEndPos
    'save x, get Y
    X = bytIn
    Y = agPicData(lngPos)
    lngPos = lngPos + 1

    'if visual OR priority but not both
    If (CurrentPen.VisColor < agNone) Xor (CurrentPen.PriColor < agNone) Then
      'use flood fill
      'if drawing visual
      If CurrentPen.VisColor < agNone Then
        'if color is not white, and current pixel IS white,
        If CurrentPen.VisColor <> agWhite And VisBuildData(X + 160 * Y) = agWhite Then
          'store the starting point in first queue position
          QueueStart = 0
          QueueEnd = 1
          lngOffset = Y * 160 + X
          Queue(QueueStart) = lngOffset
          'set first point
          VisBuildData(lngOffset) = CurrentPen.VisColor
          
          Do
            lngOffset = Queue(QueueStart)
            X = lngOffset Mod 160
            Y = lngOffset \ 160
            QueueStart = QueueStart + 1
  
            'if pixel above is white,
            If Y > 0 Then
              lngOffset = (Y - 1) * 160 + X
              If VisBuildData(lngOffset) = agWhite Then
                'set it
                VisBuildData(lngOffset) = CurrentPen.VisColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel to left is white,
            If X > 0 Then
              lngOffset = Y * 160 + X - 1
              If VisBuildData(lngOffset) = agWhite Then
                'set it
                VisBuildData(lngOffset) = CurrentPen.VisColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel to right is white,
            If X < 159 Then
              lngOffset = Y * 160 + X + 1
              If VisBuildData(lngOffset) = agWhite Then
                'set it
                VisBuildData(lngOffset) = CurrentPen.VisColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel below is white
            If Y < 167 Then
              lngOffset = (Y + 1) * 160 + X
              If VisBuildData(lngOffset) = agWhite Then
                'set it
                VisBuildData(lngOffset) = CurrentPen.VisColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
          Loop Until QueueStart = QueueEnd
        End If
      Else
        'if color is not red, and current pixel IS red,
        If CurrentPen.PriColor <> agRed And PriBuildData(X + 160 * Y) = agRed Then
          'store the starting point in first queue position
          QueueStart = 0
          QueueEnd = 1
          lngOffset = Y * 160 + X
          Queue(QueueStart) = lngOffset
          'set first point
          PriBuildData(lngOffset) = CurrentPen.PriColor
          
          Do
            lngOffset = Queue(QueueStart)
            X = lngOffset Mod 160
            Y = lngOffset \ 160
            QueueStart = QueueStart + 1
  
            'if pixel above is red,
            If Y > 0 Then
              lngOffset = (Y - 1) * 160 + X
              If PriBuildData(lngOffset) = agRed Then
                'set it
                PriBuildData(lngOffset) = CurrentPen.PriColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel to left is red,
            If X > 0 Then
              lngOffset = Y * 160 + X - 1
              If PriBuildData(lngOffset) = agRed Then
                'set it
                PriBuildData(lngOffset) = CurrentPen.PriColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel to right is red,
            If X < 159 Then
              lngOffset = Y * 160 + X + 1
              If PriBuildData(lngOffset) = agRed Then
                'set it
                PriBuildData(lngOffset) = CurrentPen.PriColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
            'if pixel below is red
            If Y < 167 Then
              lngOffset = (Y + 1) * 160 + X
              If PriBuildData(lngOffset) = agRed Then
                'set it
                PriBuildData(lngOffset) = CurrentPen.PriColor
                'add to queue
                Queue(QueueEnd) = lngOffset
                QueueEnd = QueueEnd + 1
              End If
            End If
          Loop Until QueueStart = QueueEnd
        End If
      End If
    'if drawing both
    ElseIf (CurrentPen.VisColor < agNone) And (CurrentPen.VisColor < agNone) Then
      'need to manually fill to ensure priority fill matches visual boundaries
      'if picture draw color is NOT white, and current pixel is white
      If CurrentPen.VisColor <> agWhite And VisBuildData(X + 160 * Y) = agWhite Then
        'store the starting point in first queue position
        QueueStart = 0
        QueueEnd = 1
        lngOffset = Y * 160 + X
        Queue(QueueStart) = lngOffset
        'set first point
        VisBuildData(lngOffset) = CurrentPen.VisColor
        PriBuildData(lngOffset) = CurrentPen.PriColor
        
        Do
          lngOffset = Queue(QueueStart)
          X = lngOffset Mod 160
          Y = lngOffset \ 160
          QueueStart = QueueStart + 1

          'if pixel above is white,
          If Y > 0 Then
            lngOffset = (Y - 1) * 160 + X
            If VisBuildData(lngOffset) = agWhite Then
              'set it
              VisBuildData(lngOffset) = CurrentPen.VisColor
              PriBuildData(lngOffset) = CurrentPen.PriColor
              'add to queue
              Queue(QueueEnd) = lngOffset
              QueueEnd = QueueEnd + 1
            End If
          End If
          'if pixel to left is white,
          If X > 0 Then
            lngOffset = Y * 160 + X - 1
            If VisBuildData(lngOffset) = agWhite Then
              'set it
              VisBuildData(lngOffset) = CurrentPen.VisColor
              PriBuildData(lngOffset) = CurrentPen.PriColor
              'add to queue
              Queue(QueueEnd) = lngOffset
              QueueEnd = QueueEnd + 1
            End If
          End If
          'if pixel to right is white,
          If X < 159 Then
            lngOffset = Y * 160 + X + 1
            If VisBuildData(lngOffset) = agWhite Then
              'set it
              VisBuildData(lngOffset) = CurrentPen.VisColor
              PriBuildData(lngOffset) = CurrentPen.PriColor
              'add to queue
              Queue(QueueEnd) = lngOffset
              QueueEnd = QueueEnd + 1
            End If
          End If
          'if pixel below is white
          If Y < 167 Then
            lngOffset = (Y + 1) * 160 + X
            If VisBuildData(lngOffset) = agWhite Then
              'set it
              VisBuildData(lngOffset) = CurrentPen.VisColor
              PriBuildData(lngOffset) = CurrentPen.PriColor
              'add to queue
              Queue(QueueEnd) = lngOffset
              QueueEnd = QueueEnd + 1
            End If
          End If
        Loop Until QueueStart = QueueEnd
      End If
    End If
    
    'get next byte
    bytIn = agPicData(lngPos)
    lngPos = lngPos + 1
  Loop
Exit Sub

ErrHandler:
  Debug.Assert False
  'no way to handle if a bad pixel is encountered; just give up
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 662, strErrSrc, Replace(LoadResString(662), ARG1, CStr(lngError) & ":" & strError)
End Sub

