VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AGIGame"
Attribute VB_GlobalNameSpace = True
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E06727600AD"
'    WinAGI Game Engine
'    Copyright (C) 2005 - 2019 Andrew Korson
'
'    This program is free software; you can redistribute it and/or modify
'    it under the terms of the GNU General Public License as published by
'    the Free Software Foundation; either version 2 of the License, or
'    (at your option) any later version.
'
'    This program is distributed in the hope that it will be useful,
'    but WITHOUT ANY WARRANTY; without even the implied warranty of
'    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'    GNU General Public License for more details.
'
'    You should have received a copy of the GNU General Public License
'    along with this program; if not, write to the Free Software
'    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'enums (have to put them in a class module so
'they will be visible throughout the project

Option Base 0
Option Explicit

Public Enum AGIResType
  rtLogic = 0
  rtPicture = 1
  rtSound = 2
  rtView = 3
  rtObjects = 4
  rtWords = 5
  rtLayout = 6
  rtMenu = 7
  rtGlobals = 8
  rtGame = 9
  rtText = 10
  rtWarnings = 11
  rtNone = 255
End Enum

Public Enum AGIColors
  agBlack
  agBlue
  agGreen
  agCyan
  agRed
  agMagenta
  agBrown
  agLtGray
  agDkGray
  agLtBlue
  agLtGreen
  agLtCyan
  agLtRed
  agLtMagenta
  agYellow
  agWhite
  agNone
End Enum

Public Enum ObjDirection
  odStopped
  odUp
  odUpRight
  odRight
  odDownRight
  odDown
  odDownLeft
  odLeft
  odUpLeft
End Enum

Public Enum EPlotShape
  psCircle
  psRectangle
End Enum

Public Enum EPlotStyle
  psSolid
  psSplatter
End Enum
  
Public Enum DrawFunction
  dfEnableVis = &HF0    'Change picture color and enable picture draw.
  dfDisableVis = &HF1   'Disable picture draw.
  dfEnablePri = &HF2    'Change priority color and enable priority draw.
  dfDisablePri = &HF3   'Disable priority draw.
  dfYCorner = &HF4      'Draw a Y corner.
  dfXCorner = &HF5      'Draw an X corner.
  dfAbsLine = &HF6      'Absolute line (long lines).
  dfRelLine = &HF7      'Relative line (short lines).
  dfFill = &HF8         'Fill.
  dfChangePen = &HF9    'Change pen size and style.
  dfPlotPen = &HFA      'Plot with pen.
  dfEnd = &HFF          'end of drawing
End Enum

Public Enum LogicErrorLevel
  leHigh    'all compile/decompile problems are returned as errors
  leMedium  'only errors that prevent compilation/decompilation
            'are passed; warnings embedded in
            'source code on compilation
  leLow     'only errors that prevent compilation/decompiliation
            'are passed; no warnings are given
End Enum

Public Enum ECStatus  'used to update editor as components are completed,
  csCompWords
  csCompObjects
  csAddResource
  csDoneAdding
  csCompileComplete
  csWarning
  csResError
  csLogicError
  csCanceled
End Enum

Public Enum ELStatus  'used to update editor during a game load
  lsInitialize
  lsDecompiling
  lsPropertyFile
  lsResources
  lsFinalizing
End Enum

Public Enum SoundFormat
  sfUndefined
  sfAGI
  sfMIDI
  sfScript
End Enum

Public Type PenStatus
  VisColor As AGIColors
  PriColor As AGIColors
  PlotShape As EPlotShape
  PlotStyle As EPlotStyle
  PlotSize As Long
End Type

Public Type AGIWord
  WordText As String
  Group As Long
End Type

Public Type FreeSpaceInfo
  VOL As Byte
  Start As Long
  End As Long
End Type

'type used for defined names
Public Type TDefine
  Name As String
  Default As String 'for reserved, this is default name; not used for other defines
  Value As String
  Type As ArgTypeEnum
End Type

Private strErrSource As String


Public Sub CancelCompile()
  
  'if compiling
  If agCompGame Then
    'reset to force cancel
    agCompGame = False
    'set flag indicating cancellation
    agCancelComp = True
  End If
End Sub
Public Property Get Commands() As AGICommands

  Set Commands = agCmdCol
End Property

Public Property Let DefResDir(NewDir As String)

  'validate- cant have  \/:*?<>|
  
  NewDir = Trim$(NewDir)
  
  If LenB(NewDir) = 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "\") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "/") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, ":") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "*") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "?") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "<") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, ">") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, "|") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewDir, " ") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  
  'save new resdir name
  agDefResDir = NewDir
End Property

Public Property Get DefResDir() As String
  DefResDir = agDefResDir
End Property

Public Property Let DOSExec(NewVal As String)

  'no validation required
  agDOSExec = NewVal
  
  'if a game is loaded,
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "DOSExec", agDOSExec
  End If
End Property

Public Property Get DOSExec() As String

  'return stored value
  DOSExec = agDOSExec
  
End Property

Public Property Let EGAColor(ByVal Index As Integer, ByVal NewColor As Long)
  
  Dim bytRed As Byte, bytBlue As Byte, bytGreen As Byte
  
  'in API calls, colors are a four byte number:
  '         xxrrggbb
  '(perhaps because of way CopyMemory works?)
  '
  'in VB (and other languages?) colors are also four byte:
  '         xxbbggrr
  
  'validate index
  If Index < 0 Or Index > 15 Then
    On Error GoTo 0: Err.Raise 9, strErrSource, "Subscript out of range"
    Exit Property
  End If
  
  'store the new color
  lngEGACol(Index) = NewColor
  
  'now invert red and blue components for revcolor
  lngEGARevCol(Index) = (NewColor And &HFF000000) + (NewColor And &HFF) * &H10000 + (NewColor And &HFF00&) + (NewColor And &HFF0000) / &H10000
  

End Property

Public Property Get EGAColor(ByVal Index As Integer) As Long
  
  'in API calls, colors are a four byte number:
  '         xxrrggbb
  '(perhaps because of way CopyMemory works?)
  '
  'in VB (and other languages?) colors are also four byte:
  '         xxbbggrr
  
  'validate index
  If Index < 0 Or Index > 15 Then
    On Error GoTo 0: Err.Raise 9, strErrSource, "Subscript out of range"
    Exit Property
  End If
  
  EGAColor = lngEGACol(Index)
End Property

Public Property Get EGARevColor(ByVal Index As Integer) As Long

  'in API calls, colors are a four byte number:
  '         xxrrggbb
  '(perhaps because of way CopyMemory works?)
  '
  'in VB (and other languages?) colors are also four byte:
  '         xxbbggrr
  
  'validate index
  If Index < 0 Or Index > 15 Then
    On Error GoTo 0: Err.Raise 9, strErrSource, "Subscript out of range"
    Exit Property
  End If
  
  EGARevColor = lngEGARevCol(Index)
End Property


Public Property Let GameAbout(NewAbout As String)
  
  'limit to 4096 characters
  If Len(NewAbout) > 4096 Then
    agAbout = Left$(NewAbout, 4096)
  Else
    agAbout = NewAbout
  End If
  
  'if a game is loaded,
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "About", agAbout
  End If
End Property

Public Property Get GameAbout() As String
  
  'game about doesn't depend on game load status
  'just return what it is...
  GameAbout = agAbout
End Property

Public Property Let GameAuthor(NewAuthor As String)

  'limit author to 256 bytes
  If Len(NewAuthor) > 256 Then
    agAuthor = Left$(NewAuthor, 256)
  Else
    agAuthor = NewAuthor
  End If
  
  'if game loaded,
  If agGameLoaded Then
    'write property
    WriteGameSetting "General", "Author", agAuthor
  End If
End Property


Public Property Get GameAuthor() As String
  
  'game about doesn't depend on game load status
  'just return what it is...
  GameAuthor = agAuthor
End Property

Public Property Let GameDescription(NewDescription As String)
  
  'comments limited to 4K
  If Len(NewDescription) > 4096 Then
    agDescription = Left$(NewDescription, 4096)
  Else
    agDescription = NewDescription
  End If
  
  'if a game is loaded
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "Description", agDescription
  End If
End Property


Public Property Get GameDescription() As String
  
  'game about doesn't depend on game load status
  'just return what it is...
  GameDescription = agDescription
End Property

Public Property Let GameDir(NewDir As String)

  'changing directory is only allowed if a game
  'is loaded, but....
  '
  'changing gamedir directly may cause problems
  'because the game may not be able to find the
  'resource files it is looking for;
  'also, changing the gamedir directly does not
  'update the resource directory
  '
  'It is responsibility of calling function to
  'make sure that all game resources/files/
  'folders are also moved/renamed as needed to
  'support the new directory; exception is the
  'agGameFile property, which gets updated
  'in this property (but not moved or created)
  
  If agGameLoaded Then
    'validate gamedir
    If Not FileExists(cDir(NewDir), vbDirectory) Then
      'return error
      On Error GoTo 0: Err.Raise vbObjectError + 630, strErrSource, Replace(LoadResString(630), ARG1, NewDir)
      Exit Property
    End If
    
    'change the directory
    agGameDir = cDir(NewDir)
    
    'update gamefile name
    agGameFile = agGameDir & agGameID & ".wag"
  End If
End Property

Public Property Get GameEvents() As AGIGameEvents
  
  Set GameEvents = agGameEvents
End Property

Public Property Get PlatformType() As Long
  
  PlatformType = agPlatformType
End Property

Public Property Let PlatformType(NewVal As Long)

  'only 0 - 4 are valid
  If NewVal < 0 Or NewVal > 4 Then
    agPlatformType = 0
  Else
    agPlatformType = NewVal
  End If
  
  'if a game is loaded,
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "PlatformType", agPlatformType
  End If
End Property

Public Property Let Platform(NewVal As String)
  
  'no validation required
  agPlatformFile = NewVal

  'if a game is loaded,
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "Platform", agPlatformFile
  End If
End Property

Public Property Let PlatformOpts(NewVal As String)
  
  'no validation required
  agPlatformOpts = NewVal

  'if a game is loaded,
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "PlatformOpts", agPlatformOpts
  End If
End Property


Public Property Get Platform() As String
  
  'return stored Value
  Platform = agPlatformFile
End Property

Public Property Get PlatformOpts() As String
  
  'return stored Value
  PlatformOpts = agPlatformOpts
End Property

Public Property Get GameVersion() As String
  
  'game version doesn't depend on current load status;
  'just return what it is
  GameVersion = agGameVersion
End Property

Public Property Let GameVersion(NewVersion As String)
  
  'limit to 256 bytes
  If Len(NewVersion) > 256 Then
    agGameVersion = Left$(NewVersion, 256)
  Else
    agGameVersion = NewVersion
  End If
  
  'if game loaded
  If agGameLoaded Then
    'write new property
    WriteGameSetting "General", "GameVersion", agGameVersion
  End If
End Property

Public Property Get GlobalDefines() As TDefine()

  Dim dtFileMod As Date
  
  On Error Resume Next
  'initialize global defines
  
  dtFileMod = FileLastMod(agGameDir & "globals.txt")
  If CRC32(StrConv(CStr(dtFileMod), vbFromUnicode)) <> agGlobalCRC Then
    GetGlobalDefines
  End If
  
  GlobalDefines = agGlobal()
End Property

Public Property Get InterpreterVersion() As String
  
  'if a game is loaded
  If agGameLoaded Then
    InterpreterVersion = agIntVersion
  Else
    'if no version yet
    If Len(agIntVersion) = 0 Then
'      'use 2.917
      InterpreterVersion = "2.917"
    End If
  End If
End Property
Public Sub CloseGame()
  
  On Error GoTo ErrHandler
  
  'if no game is currently loaded
  If Not agGameLoaded Then
    'don't do anything
    Exit Sub
  End If
  
  'unload and remove all resources
  agViews.Clear
  agLogs.Clear
  agPics.Clear
  agSnds.Clear
  If agInvObj.Loaded Then
    agInvObj.Unload
  End If
  agInvObj.InGame = False
  If agVocabWords.Loaded Then
    agVocabWords.Unload
  End If
  agVocabWords.InGame = False
  
  On Error Resume Next
  
  'write colors for palette
  SavePalette
  RestoreDefaultColors
  
  'write useresname property
  WriteGameSetting "General", "UseResNames", agUseRes
  
  'write lastdate of editing
  WriteGameSetting "General", "LastEdit", agLastDate
  
  'now save it
  SaveSettingList agGameProps
  Set agGameProps = Nothing
  
  
  'then clear it
  agLastDate = 0
  
  'clear out other properties
  agIntVersion = vbNullString
  agIsVersion3 = False
  agDirExpanded = False
  agDescription = vbNullString
  agGameID = vbNullString
  agAuthor = vbNullString
  agGameVersion = vbNullString
  agAbout = vbNullString
  agResDirName = vbNullString
  agPlatformType = 0
  agPlatformFile = vbNullString
  agPlatformOpts = vbNullString
  agDOSExec = vbNullString
  
  agGameFile = vbNullString
  
  'clear flag
  agGameLoaded = False
  
  'reset global CRC
  agGlobalCRC = 0
  
  'finally, reset directories
'''  agGameDir = cDir(CurDir$())
'''  agResDir = cDir(CurDir$())
  agGameDir = vbNullString
  agResDir = vbNullString
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 644, strErrSrc, Replace(LoadResString(644), ARG1, CStr(lngError) & ":" & strError)
End Sub

Public Sub CompileGame(ByVal RebuildOnly As Boolean, Optional ByVal NewGameDir As String)
  'compiles the game into NewGameDir
  '
  'if RebuildOnly is true, the VOL files are
  'rebuilt without recompiling all logics
  'and WORDS.TOK and OBJECT are not recompiled
  '
  'WARNING: if NewGameDir is same as current directory
  'this WILL overwrite current game files
  '

  Dim tmpLogic As AGILogic, tmpPicture As AGIPicture
  Dim tmpView As AGIView, tmpSound As AGISound
  Dim CurResType As AGIResType, CurResNum As Byte
  Dim tmpLoaded As Boolean, tmpSourceLoaded As Boolean
  Dim DirByte(0 To 2) As Byte, strID As String
  Dim i As Long, blnReplace As Boolean
  Dim intTmpDir As Integer, lngOffset As Long
  Dim NewIsV3 As Boolean, strFileName As String

  On Error Resume Next
  
  'only loaded games can be compiled
  If Not agGameLoaded Then
    Exit Sub
  End If

  'set compiling flag
  agCompGame = True
  'reset cancel flag
  agCancelComp = False

  'if no directory passed,
  If LenB(NewGameDir) = 0 Then
    NewGameDir = agGameDir
  End If

  'validate new directory
  NewGameDir = cDir(NewGameDir)
  If LenB(Dir(NewGameDir, vbDirectory)) = 0 Then
    'this isn't a directory
    CompleteCancel True
    On Error GoTo 0: Err.Raise vbObjectError + 561, Replace(LoadResString(561), ARG1, NewGameDir)
    Exit Sub
  End If

  'set flag if game is being compiled in its current directory
  blnReplace = (LCase$(NewGameDir) = LCase$(agGameDir))

  'save compile dir so rebuild method can access it
  agCompileDir = NewGameDir

  'set new game version
  If agChangeVersion Then
    NewIsV3 = Not agIsVersion3
  Else
    NewIsV3 = agIsVersion3
  End If

  'ensure switch flag is reset
  agChangeVersion = False

  'if version 3
  If NewIsV3 Then
    'version 3 ids limited to 5 characters
    If Len(agGameID) > 5 Then
      'truncate it
      strID = Left$(agGameID, 5)
      
      'rename wag and wal files
      On Error Resume Next
      Name agGameDir & agGameID & ".wag" As agGameDir & strID & ".wag"
      Name agGameDir & agGameID & ".wal" As agGameDir & strID & ".wal"
      'update agGameFile
      agGameFile = agGameDir & strID & ".wag"
      agGameID = strID
      
      'and save it
      WriteGameSetting "General", "GameID", agGameID
    End If
    strID = agGameID
  Else
    strID = vbNullString
  End If

  'if not rebuildonly
  If Not RebuildOnly Then
    'set status
    agGameEvents.RaiseEvent_CompileStatus csCompWords, 0, 0, vbNullString
    'check for cancellation
    If Not agCompGame Then
      CompleteCancel
      Exit Sub
    End If

    'compile WORDS.TOK if dirty
    If agVocabWords.IsDirty Then
      agVocabWords.Save
    End If
    If Err.Number <> 0 Then
      'note it
      agGameEvents.RaiseEvent_CompileStatus csResError, 0, 0, "Error during compilation of WORDS.TOK (" & Err.Description & ")"
      Err.Clear
      'check for cancellation
      If Not agCompGame Then
        CompleteCancel
        Exit Sub
      End If
    End If

    'if compiling to a different directory
    If Not blnReplace Then
      Kill NewGameDir & "WORDS.OLD"
      Name NewGameDir & "WORDS.TOK" As NewGameDir & "WORDS.OLD"
      'copy into new directory
      FileCopy agVocabWords.ResFile, NewGameDir & "WORDS.TOK"
      If Err.Number <> 0 Then
        'note it
        agGameEvents.RaiseEvent_CompileStatus csResError, 0, 0, "Error while creating WORDS.TOK file (" & Err.Description & ")"
        Err.Clear
        'check for cancellation
        If Not agCompGame Then
          CompleteCancel
          Exit Sub
        End If
      End If
    End If
    
    'clear errors
    Err.Clear

    'set status
    agGameEvents.RaiseEvent_CompileStatus csCompObjects, 0, 0, vbNullString
    'check for cancellation
    If Not agCompGame Then
      CompleteCancel
      Exit Sub
    End If

    'compile OBJECT file if dirty
    If agInvObj.IsDirty Then
      agInvObj.Save
    End If
    If Err.Number <> 0 Then
      'note it
      agGameEvents.RaiseEvent_CompileStatus csResError, 0, 0, "Error during compilation of OBJECT (" & Err.Description & ")"
      'check for cancellation
      If Not agCompGame Then
        CompleteCancel
        Exit Sub
      End If
      Err.Clear
    End If

    'if compiling to a different directory
    If Not blnReplace Then
      Kill NewGameDir & "OBJECT.OLD"
      Name NewGameDir & "OBJECT" As NewGameDir & "OBJECT.OLD"
      'copy into new directory
      FileCopy agInvObj.ResFile, NewGameDir & "OBJECT"
      If Err.Number <> 0 Then
        'note it
        agGameEvents.RaiseEvent_CompileStatus csResError, 0, 0, "Error while creating OBJECT file (" & Err.Description & ")"
        Err.Clear
        'check for cancellation
        If Not agCompGame Then
          CompleteCancel
          Exit Sub
        End If
      End If
    End If
  End If

  'ensure error object is cleared
  Err.Clear

  'reset game compiler variables
  lngCurrentVol = 0
  lngCurrentLoc = 0
  strNewDir = NewGameDir

  'create temp dir files
  If NewIsV3 Then
    'ensure temp dir file does not already exist
    Kill NewGameDir & agGameID & "DIR.NEW"
    If Err.Number <> 0 And Err.Number <> 53 Then
      strError = Err.Number
      strError = Err.Description
      strErrSrc = Err.Source
      lngError = Err.Number
      CompleteCancel True
      
      On Error GoTo 0: Err.Raise vbObjectError + 503, strErrSrc, Replace(LoadResString(503), ARG1, CStr(lngError) & ":" & strError)
      Exit Sub
    End If
    Err.Clear
    
    'open temp dir file
    intDirFile = FreeFile()
    Open NewGameDir & agGameID & "DIR.NEW" For Binary As intDirFile
    If Err.Number <> 0 Then
      strError = Err.Number
      strError = Err.Description
      strErrSrc = Err.Source
      lngError = Err.Number
      CompleteCancel True
      
      On Error GoTo 0: Err.Raise vbObjectError + 503, strErrSrc, Replace(LoadResString(503), ARG1, CStr(lngError) & ":" & strError)
      Exit Sub
    End If
    
    'add header (sets offsets to 8, 768+8, etc)
    Put intDirFile, 1, CLng(&H3080008)
    Put intDirFile, , CLng(&H9080608)
    'add data
    Put intDirFile, , String$(3072, &HFF)
  Else
    GetNextDir NewGameDir & "LOGDIR.NEW"
  End If
  
  'ensure temp vol files are removed
  Kill NewGameDir & "NEW_VOL.*"
  Err.Clear

  'open first new vol file
  intVolFile = FreeFile()
  Open NewGameDir & "NEW_VOL.0" For Binary As intVolFile
  If Err.Number <> 0 Then
    strError = Err.Number
    strError = Err.Description
    strErrSrc = Err.Source
    lngError = Err.Number
    CompleteCancel True
    
    On Error GoTo 0: Err.Raise vbObjectError + 503, strErrSrc, Replace(LoadResString(503), ARG1, CStr(lngError) & ":" & strError)
    Exit Sub
  End If

  'add all logic resources
  CompileResCol agLogs, rtLogic, RebuildOnly, NewIsV3
  
  'pass along any errors
  If Err.Number <> 0 Then
    lngError = Err.Number
    strError = Err.Description
    CompleteCancel True
    On Error GoTo 0: Err.Raise lngError, strErrSource, strError
    Exit Sub
  End If
  'check for cancellation
  'if a resource error (or user canceled) encountered, just exit
  If Not agCompGame Then
    Exit Sub
  End If

 If Not NewIsV3 Then
    'open picdir
    GetNextDir NewGameDir & "PICDIR.NEW"
    'pass along any errors
    If Err.Number <> 0 Then
      lngError = Err.Number
      strError = Err.Description
      CompleteCancel True
      On Error GoTo 0: Err.Raise lngError, strErrSource, strError
      Exit Sub
    End If
  End If

  'add picture resources
  CompileResCol agPics, rtPicture, RebuildOnly, NewIsV3
  'pass along any errors
  If Err.Number <> 0 Then
    lngError = Err.Number
    strError = Err.Description
    CompleteCancel True
    On Error GoTo 0: Err.Raise lngError, strErrSource, strError
    Exit Sub
  End If
  'check for cancellation
  'if a resource error (or user canceled) encountered, just exit
  If Not agCompGame Then
    Exit Sub
  End If

  If Not NewIsV3 Then
    'open viewdir
    GetNextDir NewGameDir & "VIEWDIR.NEW"
    'pass along any errors
    If Err.Number <> 0 Then
      lngError = Err.Number
      strError = Err.Description
      CompleteCancel True
      On Error GoTo 0: Err.Raise lngError, strErrSource, strError
      Exit Sub
    End If
  End If

  'add view resources
  CompileResCol agViews, rtView, RebuildOnly, NewIsV3
  'pass along any errors
  If Err.Number <> 0 Then
    lngError = Err.Number
    strError = Err.Description
    CompleteCancel True
    On Error GoTo 0: Err.Raise lngError, strErrSource, strError
    Exit Sub
  End If
  'check for cancellation
  'if a resource error (or user canceled) encountered, just exit
  If Not agCompGame Then
    Exit Sub
  End If

  If Not NewIsV3 Then
    'open snddir
    GetNextDir NewGameDir & "SNDDIR.NEW"
    'pass along any errors
    If Err.Number <> 0 Then
      lngError = Err.Number
      strError = Err.Description
      CompleteCancel True
      On Error GoTo 0: Err.Raise lngError, strErrSource, strError
      Exit Sub
    End If
  End If

  'add sound resources
  CompileResCol agSnds, rtSound, RebuildOnly, NewIsV3
  'pass along any errors
  If Err.Number <> 0 Then
    lngError = Err.Number
    strError = Err.Description
    CompleteCancel True
    On Error GoTo 0: Err.Raise lngError, strErrSource, strError
    Exit Sub
  End If
  'check for cancellation
  'if a resource error (or user canceled) encountered, just exit
  If Not agCompGame Then
    Exit Sub
  End If

  'close files
  Close intDirFile
  Close intVolFile

  'update status
  agGameEvents.RaiseEvent_CompileStatus csDoneAdding, 0, 0, vbNullString
  'check for cancellation
  If Not agCompGame Then
    CompleteCancel
    Exit Sub
  End If

  'remove any existing old dirfiles
  If NewIsV3 Then
    Kill NewGameDir & agGameID & "DIR.OLD"
  Else
    Kill NewGameDir & "LOGDIR.OLD"
    Kill NewGameDir & "PICDIR.OLD"
    Kill NewGameDir & "SNDDIR.OLD"
    Kill NewGameDir & "VIEWDIR.OLD"
  End If
  
  'remove any existing old vol files
  For i = 0 To 15
    If NewIsV3 Then
      Kill NewGameDir & agGameID & "VOL." & CStr(i) & ".OLD"
    Else
      Kill NewGameDir & "VOL." & CStr(i) & ".OLD"
    End If
  Next i
  Err.Clear
  
  'rename existing dir and vol files as .OLD
  Name NewGameDir & agGameID & "DIR" As NewGameDir & agGameID & "DIR.OLD"
  Name NewGameDir & "LOGDIR" As NewGameDir & "LOGDIR.OLD"
  Name NewGameDir & "PICDIR" As NewGameDir & "PICDIR.OLD"
  Name NewGameDir & "VIEWDIR" As NewGameDir & "VIEWDIR.OLD"
  Name NewGameDir & "SNDDIR" As NewGameDir & "SNDDIR.OLD"
  For i = 0 To 15
    'rename current vol files
    Name NewGameDir & agGameID & "VOL." & CStr(i) As NewGameDir & agGameID & "VOL." & CStr(i) & ".OLD"
    Name NewGameDir & "VOL." & CStr(i) As NewGameDir & "VOL." & CStr(i) & ".OLD"
  Next i
  Err.Clear
  
  'use errhandler to identify any issues with creating new files
  On Error GoTo ErrHandler
  
  If NewIsV3 Then
    'rename temp dir file
    strFileName = agGameID & "DIR"
    Name NewGameDir & agGameID & "DIR.NEW" As NewGameDir & strFileName
  Else
    'rename temp dir files
    strFileName = "LOGDIR"
    Name NewGameDir & "LOGDIR.NEW" As NewGameDir & strFileName
    strFileName = "PICDIR"
    Name NewGameDir & "PICDIR.NEW" As NewGameDir & strFileName
    strFileName = "SNDDIR"
    Name NewGameDir & "SNDDIR.NEW" As NewGameDir & strFileName
    strFileName = "VIEWDIR"
    Name NewGameDir & "VIEWDIR.NEW" As NewGameDir & strFileName
  End If

  Err.Clear
  For i = 0 To lngCurrentVol
    strFileName = strID & "VOL." & CStr(i)
    Name NewGameDir & "NEW_VOL." & CStr(i) As NewGameDir & strFileName
  Next i

  On Error Resume Next
  
  'update status to indicate complete
  agGameEvents.RaiseEvent_CompileStatus csCompileComplete, 0, 0, vbNullString

  'ensure errors cleared
  Err.Clear

  'if replacing
  If blnReplace Then
    'update all resources with correct vol and loc positions
    If NewIsV3 Then
      'open new dir file
      intTmpDir = FreeFile()
      Open NewGameDir & agGameID & "DIR" For Binary As intTmpDir

      'step through logics
      For Each tmpLogic In agLogs
        'extract directory info
        Get intTmpDir, tmpLogic.Number * 3 + 9, DirByte 'offet is 8; add one because get is '1' based
        'extract volume and loc
        tmpLogic.Resource.Volume = DirByte(0) \ &H10
        tmpLogic.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'step through pictures
      For Each tmpPicture In agPics
        'extract directory info
        Get intTmpDir, tmpPicture.Number * 3 + &H309, DirByte 'offet is &H308; add one because get is '1' based
        'extract volume and loc
        tmpPicture.Resource.Volume = DirByte(0) \ &H10
        tmpPicture.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'step through sounds
      For Each tmpSound In agSnds
        'extract directory info
        Get intTmpDir, tmpSound.Number * 3 + &H909, DirByte 'offet is &H908; add one because get is '1' based
        'extract volume and loc
        tmpSound.Resource.Volume = DirByte(0) \ &H10
        tmpSound.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'step through views
      For Each tmpView In agViews
        'extract directory info
        Get intTmpDir, tmpView.Number * 3 + &H609, DirByte 'offet is 8; add one because get is '1' based
        'extract volume and loc
        tmpView.Resource.Volume = DirByte(0) \ &H10
        tmpView.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next

      Close intTmpDir
    Else
      'open new logdir file
      intTmpDir = FreeFile()
      Open NewGameDir & "LOGDIR" For Binary As intTmpDir

      'step through logics
      For Each tmpLogic In agLogs
        'extract directory info
        Get intTmpDir, tmpLogic.Number * 3 + 1, DirByte
        'extract volume and loc
        tmpLogic.Resource.Volume = DirByte(0) \ &H10
        tmpLogic.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'close logdir
      Close intTmpDir

      'open picdir
      intTmpDir = FreeFile()
      Open NewGameDir & "PICDIR" For Binary As intTmpDir
      'step through pictures
      For Each tmpPicture In agPics
        'extract directory info
        Get intTmpDir, tmpPicture.Number * 3 + 1, DirByte
        'extract volume and loc
        tmpPicture.Resource.Volume = DirByte(0) \ &H10
        tmpPicture.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'close picdir
      Close intTmpDir

      'open snddir
      intTmpDir = FreeFile()
      Open NewGameDir & "SNDDIR" For Binary As intTmpDir
      'step through sounds
      For Each tmpSound In agSnds
        'extract directory info
        Get intTmpDir, tmpSound.Number * 3 + 1, DirByte
        'extract volume and loc
        tmpSound.Resource.Volume = DirByte(0) \ &H10
        tmpSound.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'close snddir
      Close intTmpDir

      'open viewdir
      intTmpDir = FreeFile()
      Open NewGameDir & "VIEWDIR" For Binary As intTmpDir
      'step through views
      For Each tmpView In agViews
        'extract directory info
        Get intTmpDir, tmpView.Number * 3 + 1, DirByte
        'extract volume and loc
        tmpView.Resource.Volume = DirByte(0) \ &H10
        tmpView.Resource.Loc = (DirByte(0) And &HF) * &H10000 + DirByte(1) * &H100& + DirByte(2)
      Next
      'close viewdir
      Close intTmpDir
    End If
  End If

  'ensure error object is cleared
  Err.Clear

  'save the wag file
  SaveSettingList agGameProps
  
  'reset compiling flag
  agCompGame = False
Exit Sub

ErrHandler:
  'file error when creating new DIR and VOL files
  'pass along any errors
  lngError = Err.Number
  Select Case lngError
  Case 58 'file already exists
    'this means the previous version of the file we are trying to create
    'wasn't able to be renamed; most likely due to being open
    strError = "Unable to delete existing " & strFileName & ". New " & strFileName & " file was not created."
  Case Else
    'some other error - who knows what's going on
    strError = Err.Description
  End Select
  
  CompleteCancel True
  On Error GoTo 0: Err.Raise lngError, strErrSource, strError
End Sub


Public Property Let GameID(NewID As String)
  'limit gameID to 6 characters for v2 games and 5 characters for v3 games
  
  Dim strVolFile As String
  Dim strExtension() As String
  
  On Error GoTo ErrHandler
  
  'id is undefined if a game is not loaded
  If Not agGameLoaded Then
    On Error GoTo 0: Err.Raise vbObjectError + 677, strErrSource, LoadResString(677)
    Exit Property
  End If
  
  'version 3 games limited to 5 characters
  If agIsVersion3 Then
    If Len(NewID) > 5 Then
      NewID = Left$(NewID, 5)
    End If
  Else
    If Len(NewID) > 6 Then
      NewID = Left$(NewID, 6)
    End If
  End If
  
  'if no change
  If agGameID = NewID Then
    Exit Property
  End If
  
  'if version 3
  If agIsVersion3 Then
    'need to rename the dir file
    Name agGameDir & agGameID & "DIR" As agGameDir & NewID & "DIR"
    'and vol files
    strVolFile = Dir(agGameDir & agGameID & "VOL.*")
    
    Do Until LenB(strVolFile) = 0
      'get extension
      strExtension = Split(strVolFile, ".")
      'rename
      Name agGameDir & strVolFile As agGameDir & NewID & "VOL." & strExtension(1)
      'get next
      strVolFile = Dir()
    Loop
  End If
  
  'set new id
  agGameID = UCase$(NewID)
  
  'write new property
  WriteGameSetting "General", "GameID", NewID
  
  'rename game property file
  On Error Resume Next
  Name agGameFile As agGameDir & NewID & ".wag"
  Err.Clear
  agGameFile = agGameDir & agGameID & ".wag"
  'save the game prop file
  agGameProps.StringLine(0) = agGameFile
  SaveSettingList agGameProps
  
Exit Property

ErrHandler:
  'file renaming or property writing are only sources of error
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 530, strErrSrc, Replace(LoadResString(530), ARG1, CStr(lngError) & ":" & strError)
End Property
Public Property Get GameID() As String
    
  'id is undefined if a game is not loaded
  If Not agGameLoaded Then
    On Error GoTo 0: Err.Raise vbObjectError + 677, strErrSource, LoadResString(677)
    Exit Property
  End If
  
  GameID = agGameID
End Property


Public Property Get GameLoaded() As Boolean
  
  'returns state of game load
  GameLoaded = agGameLoaded
End Property


Public Property Get LastDate() As Date
  
  'if game loaded,
  If agGameLoaded Then
    LastDate = agLastDate
  Else
    LastDate = Now()
  End If
End Property


Public Property Set LogicSourceSettings(NewLogSettings As AGILogicSourceSettings)

  Set agMainLogSettings = NewLogSettings
End Property

Public Property Get LogicSourceSettings() As AGILogicSourceSettings
  
  Set LogicSourceSettings = agMainLogSettings
End Property


Public Property Get MaxVol0Size() As Long
  
  MaxVol0Size = agMaxVol0
End Property

Public Property Let MaxVol0Size(ByVal NewSize As Long)
  
  'validate
  If NewSize < 32768 Then
    NewSize = 32768
  ElseIf NewSize > agMaxVolSize Then
    NewSize = agMaxVolSize
  End If
  
  agMaxVol0 = NewSize
End Property

Private Function SavePalette() As String

  'write the palette bytes to the game property list
  Dim i As Long
  Dim byte1 As Byte, byte2 As Byte
  Dim byte3 As Byte, byte4 As Byte
  
  On Error GoTo ErrHandler
  
  For i = 0 To 15
    'extract the four byte values
    byte1 = lngEGACol(i) \ &H1000000
    byte2 = (lngEGACol(i) And &HFF0000) \ &H10000
    byte3 = (lngEGACol(i) And &HFF00&) \ &H100&
    byte4 = (lngEGACol(i) And &HFF)
    
    WriteGameSetting "Palette", "Color" & CStr(i), "&H" & CStr(byte1) & ",&H" & CStr(byte2) & ",&H" & CStr(byte3) & ",&H" & CStr(byte4)
  Next i
Exit Function


ErrHandler:
  Debug.Assert False
  Resume Next
End Function

Public Property Get Pictures() As AGIPictures
    
  Set Pictures = agPics
End Property





Public Property Let InterpreterVersion(NewVersion As String)
  
  'attempts to set version to new Value
  
  Dim i As Long
  Dim OldVersion As String
  
  On Error Resume Next
  
  'validate new version
  If ValidateVersion(NewVersion) Then
    'if in a game, need to adjust resources if necessary
    If agGameLoaded Then
      'if new version and old version are not same major version:
      If Asc(agIntVersion) <> Asc(NewVersion) Then
        'set flag to switch version on rebuild
        agChangeVersion = True
        
        'use compiler to rebuild new vol and dir files
        '(it is up to calling program to deal with dirty and invalid resources)
        CompileGame True
        
        'check for errors, or cancellation
        If Err.Number <> 0 Then
          strError = Err.Description
          strErrSrc = Err.Source
          lngError = Err.Number
          
          'reset cancel flag
          agCancelComp = False
          'pass along the error
          On Error GoTo 0: Err.Raise vbObjectError + 641, strErrSrc, Replace(LoadResString(641), ARG1, CStr(lngError) & ":" & strError)
          Exit Property
        
        'if canceled,
        ElseIf agCancelComp Then
          'reset cancel flag
          agCancelComp = False
          Exit Property
        End If
      End If
      Err.Clear
      
      'write new version to file
      WriteGameSetting "General", "Interpreter", NewVersion
      'and save it
      SaveSettingList agGameProps
      
    End If
    
    'set new version
    agIntVersion = NewVersion
    
    'set v3 flag
    agIsVersion3 = (Val(agIntVersion) > 3)
    
    'adjust commands
    CorrectCommands agIntVersion
  Else
    If Val(NewVersion) < 2 Or Val(NewVersion) > 3 Then
      'not a version 2 or 3 game
      On Error GoTo 0: Err.Raise vbObjectError + 597, strErrSource, LoadResString(597)
    Else
      'unsupported version 2 or 3 game
      On Error GoTo 0: Err.Raise vbObjectError + 543, strErrSource, LoadResString(543)
    End If
  End If
  
End Property
Public Property Get ResDir() As String
  
  'if a game is loaded,
  If agGameLoaded Then
    'use resdir property
    ResDir = agResDir
  Else
    'use current directory
    ResDir = cDir(CurDir$())
  End If
End Property


Public Property Get ResDirName() As String

  ResDirName = agResDirName
End Property
Public Property Let ResDirName(NewResDirName As String)

  'validate- cant have  \/:*?<>|
  
  NewResDirName = Trim$(NewResDirName)
  
  If LenB(NewResDirName) = 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "\") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "/") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, ":") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "*") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "?") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "<") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, ">") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, "|") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  If InStr(1, NewResDirName, " ") <> 0 Then
    On Error GoTo 0: Err.Raise 380, strErrSource, "Invalid property Value"
    Exit Property
  End If
  
  'save new resdir name
  agResDirName = NewResDirName
  'update the actual resdir
  agResDir = agGameDir & agResDirName & "\"
  
  'save resdirname
  WriteGameSetting "General", "ResDir", NewResDirName
End Property


Public Property Get ResTypeName(ByVal ResType As AGIResType)
  
  'returns the resource name
  ResTypeName = agResTypeName(ResType)
End Property

Public Sub SaveProperties()
  
  'this forces WinAGI to save the property file, rather than waiting until
  'the game is unloaded
  
  Debug.Assert Not (agGameProps Is Nothing)
  If Not (agGameProps Is Nothing) Then
    SaveSettingList agGameProps
  End If
End Sub

Public Property Let TemplateDir(NewDir As String)

  'validate gamedir
  If Not FileExists(cDir(NewDir), vbDirectory) Then
    'return error
    On Error GoTo 0: Err.Raise vbObjectError + 630, strErrSource, Replace(LoadResString(630), ARG1, NewDir)
    Exit Property
  End If
  
  agTemplateDir = cDir(NewDir)
End Property

Public Property Get TemplateDir() As String
  
  TemplateDir = agTemplateDir
End Property


Public Property Get TestCommands() As AGITestCommands
  
  Set TestCommands = agTestCmdCol
End Property


Public Property Get Views() As AGIViews
    
  Set Views = agViews
End Property




Public Property Get Sounds() As AGISounds
    
  Set Sounds = agSnds
End Property

Public Property Get Logics() As AGILogics
    
  Set Logics = agLogs
End Property

Public Property Get GameDir() As String
  
  'if a game is loaded,
  If agGameLoaded Then
    'use resdir property
    GameDir = agGameDir
  Else
    'use current directory
    GameDir = cDir(CurDir$())
  End If
End Property
Public Property Get InventoryObjects() As AGIInventoryObjects
  'if nothing,
  If agInvObj Is Nothing Then
    Set InventoryObjects = Nothing
  Else
    Set InventoryObjects = agInvObj
  End If
End Property

Public Sub LoadGame(GameDir As String)
  'opens a WinAGI game file or a Sierra game directory
  'depending on format
  'returns true if no errors encountered
  'returns false if errors occurred
  
  Dim blnWarnings As Boolean, lngError As Long
  Dim strTempDir As String, lngDirCount As Long
  Dim strTempFile As String, strWAGFiles() As String
  Dim i As Long, j As Long, lngFileCount As Long
  Dim strVer As String
  
  On Error Resume Next
  
  'if a game is already open,
  If agGameLoaded Then
    'can't open a game if one is already open
    On Error GoTo 0: Err.Raise vbObjectError + 501, strErrSource, LoadResString(501)
    Exit Sub
  End If
  
  Debug.Assert agGameID = ""
  
  'set game directory
  agGameDir = cDir(GameDir)
  
  'look for a WinAGI property file
  
  '*****
  ' stupid Dir() function will return files that have extensions longer than 3 characters; this can
  ' cause problems; so, rewriting the WAG file search feature;
  '
  ' new strategy is to make array of all possible matches, delete non-wag files,
  ' then if one and only one is left, we're good
  '*****
  
  strTempFile = Dir(agGameDir & "*.wag")
  Do Until LenB(strTempFile) = 0
    ReDim Preserve strWAGFiles(lngFileCount)
    strWAGFiles(lngFileCount) = strTempFile
    lngFileCount = lngFileCount + 1
    strTempFile = Dir()
  Loop
  
  'if at least one found,
  If lngFileCount > 0 Then
    'check for any files that don't end in 'wag'
    For i = lngFileCount - 1 To 0 Step -1
      If UCase(Right(strWAGFiles(i), 4)) <> ".WAG" Then
        'it don't count
        strWAGFiles(i) = ""
        lngFileCount = lngFileCount - 1
      End If
    Next i
  
    'now we have good list of WAG files
    Select Case lngFileCount
    Case 0  'no file
      'do nothing; a wag file will be created later
      
    Case 1
      'save path
      For i = 0 To UBound(strWAGFiles())
        If LenB(strWAGFiles(i)) <> 0 Then
          agGameFile = strWAGFiles(i)
          Exit For
        End If
      Next i
      agGameFile = agGameDir & agGameFile
      
      'open the property file
      Set agGameProps = OpenSettingList(agGameFile)
      'check to see if it's valid
      strVer = ReadSettingString(agGameProps, "General", "WinAGIVersion", "")
      If Len(strVer) = 0 Then
        'not a valid file; maybe it's an old binary version
        ConvertWag
      Else
        Select Case strVer
        Case "1.2.2" ' need to convert it to 1.2.3
          WriteGameSetting "General", "WinAGIVersion", "1.2.3"
        Case "1.2.3"
          'ok
        Case Else
          'if another winagi version is ever released,
          'yeah, right; just consider this file invalid
          Set agGameProps = Nothing
        End Select
      End If
      
      'if not a valid wag file, propfile will be nothing
      'if valid, get the gameid
      If Not (agGameProps Is Nothing) Then
        agGameID = ReadSettingString(agGameProps, "General", "GameID", "")
      End If
      
      'id found?
      If Len(agGameID) > 0 Then
        'got ID; now get interpreter version from propfile
        agIntVersion = ReadSettingString(agGameProps, "General", "Interpreter", "")
        
        If Len(agIntVersion) = 0 Then
          'invalid game file; reset ID to blank
          'and then try to extract it by loading Sierra files
          agGameID = ""
          Set agGameProps = Nothing
        End If
      Else
        'clear propfile
        Set agGameProps = Nothing
      End If
      
    Case Else 'too many
      'clear variables
      agGameDir = vbNullString
      'raise error
      On Error GoTo 0: Err.Raise vbObjectError + 508, strErrSource, LoadResString(508)
      Exit Sub
    End Select
  End If
  
  'if a valid wag file was found, we now have agGameID, agGameFile and correct interpreter version
  'if no wag file found (or invalid wag file), these three key properties are not yet
  'set; we will need to try to extract them from Sierra files (i.e. from DIR files
  'and VOL files)
  
  'if no id (meaning couldn't find ID in a .wag file)
  If LenB(agGameID) = 0 Then
    Debug.Assert agGameProps Is Nothing
    'inform user no wag file found
    agGameEvents.RaiseEvent_LoadStatus lsDecompiling, rtNone, 0, vbNullString
    
    'if this game is not a sierra AGI game
    '(which gets gameid, and sets version 3 status flag)
    If Not IsSierraGameDir(agGameDir) Then
      'save dir as error string
      strError = agGameDir
      'clear game variables
      agGameDir = vbNullString
      agGameFile = vbNullString
      agGameID = vbNullString
      agIntVersion = vbNullString
      agIsVersion3 = False
      'invalid game directory
      On Error GoTo 0
      On Error GoTo 0: Err.Raise vbObjectError + 541, strErrSource, Replace(LoadResString(541), ARG1, strError)
      Exit Sub
    End If
    
    'build default propfile name
    agGameFile = agGameDir & agGameID & ".wag"
    'delete any existing game file, as it is corrupt
    Kill agGameFile
    Err.Clear
    
    Set agGameProps = OpenSettingList(agGameFile)
    agGameProps.Add "# WinAGI Game Property File"
    agGameProps.Add "#"
    WriteGameSetting "General", "WinAGIVersion", "1.2.3"
    WriteGameSetting "General", "GameID", agGameID
    SaveSettingList agGameProps
  End If
  
  'informs user we are checking property file to set game parameters
  agGameEvents.RaiseEvent_LoadStatus lsPropertyFile, rtNone, 0, vbNullString
  'if no version
  If Len(agIntVersion) = 0 Then
    'get version number (version3 flag already set)
    agIntVersion = GetIntVersion()
  
    'if not valid
    If Len(agIntVersion) = 0 Then
      'clear game variables
      agGameDir = vbNullString
      agGameFile = vbNullString
      agGameID = vbNullString
      agIsVersion3 = False
      Set agGameProps = Nothing
      'invalid number found
      On Error GoTo 0: Err.Raise vbObjectError + 543, strErrSource, LoadResString(543)
      Exit Sub
    End If
    'save version
    WriteGameSetting "General", "Interpreter", agIntVersion
  End If
  
  'set v3 flag
  agIsVersion3 = (Val(agIntVersion) > 3)

  'get resdir before loading resources
  agResDirName = ReadSettingString(agGameProps, "General", "ResDir", "")
  
  'if none, use default
  If LenB(agResDirName) = 0 Then
    'look for an existing directory
    strTempDir = Dir(agGameDir, vbDirectory)
    Do Until Len(strTempDir) = 0
      If strTempDir <> "." And strTempDir <> ".." Then
        If (GetAttr(agGameDir & strTempDir) And vbDirectory) = vbDirectory Then
          lngDirCount = lngDirCount + 1
          If lngDirCount > 1 Then
            Exit Do
          End If
          strTempFile = strTempDir
        End If
      End If
      strTempDir = Dir(, vbDirectory)
    Loop
    
    'is there only one subfolder?
    If lngDirCount = 1 Then
      'assume it is resource directory
        agResDirName = strTempFile
    Else
      'either no subfolders, or more than one
      'so use default
      agResDirName = agDefResDir
    End If
    
    WriteGameSetting "General", "ResDir", agResDirName
  End If
  
  'now create full resdir from name
  agResDir = agGameDir & agResDirName & "\"

  'ensure resource directory exists
  If Not FileExists(agGameDir & agResDirName, vbDirectory) Then
    MkDir agGameDir & agResDirName
    'if can't create the resources directory
    If Err.Number <> 0 Then
      'note the problem in the error log as a warning
      RecordLogEvent 0, "Can't create " & agResDir
      'use game directory
      agResDir = agGameDir
      'set warning
      blnWarnings = True
    End If
  End If
  
  'load resources
  blnWarnings = LoadResources()
  'if there was an error
  If Err.Number <> 0 Then
    'can't continue
    GoTo ErrHandler
  End If
  
  'load vocabulary word list
  agGameEvents.RaiseEvent_LoadStatus lsResources, rtWords, 0, vbNullString
  
  Set agVocabWords = New AGIWordList
  agVocabWords.Init
  agVocabWords.Load agGameDir & "WORDS.TOK"
  'if there was an error,
  If Err.Number <> 0 Then
    'note the problem in the error log as a warning
    RecordLogEvent 1, "An error occurred while loading WORDS.TOK: " & Err.Description
    'reset warning flag
    blnWarnings = True
  End If
  
  'load inventory objects list
  agGameEvents.RaiseEvent_LoadStatus lsResources, rtObjects, 0, vbNullString
  
  Set agInvObj = New AGIInventoryObjects
  agInvObj.Init
  agInvObj.Load agGameDir & "OBJECT"
  'if there was an error,
  If Err.Number <> 0 Then
    'note the problem in the error log as a warning
    RecordLogEvent 1, "An error occurred while loading OBJECT: " & Err.Description
    'reset warning flag
    blnWarnings = True
  End If
  
  agGameEvents.RaiseEvent_LoadStatus lsFinalizing, rtNone, 0, vbNullString
  
  'assign commands
  AssignCommands
  
  'and adust commands based on AGI version
  CorrectCommands agIntVersion

  'clear other game properties
  agLastDate = 0
  agAuthor = vbNullString
  agDescription = vbNullString
  agGameVersion = vbNullString
  agAbout = vbNullString
  agPlatformType = 0
  agPlatformFile = vbNullString
  agPlatformOpts = vbNullString
  agDOSExec = vbNullString
  
  'get rest of game properties
  Debug.Assert Not (agGameProps Is Nothing)
  GetGameProperties
  
  'if errors
  If Err.Number <> 0 Then
    'record event
    RecordLogEvent 0, "Error while loading WAG file; some properties not loaded. (Error number: " & CStr(Err.Number) & ")"
    Err.Clear
    blnWarnings = True
  End If
  
  'force id reset
  blnSetIDs = False
  
  'we've established that the game can be opened
  'so set the loaded flag now
  agGameLoaded = True
  
  'if errors
  If blnWarnings Then
    On Error GoTo 0: Err.Raise vbObjectError + 636, strErrSource, LoadResString(636)
  End If
Exit Sub

ErrHandler:
    'save error information
    strError = Err.Description
    lngError = Err.Number
    
    'clear game variables
    agGameDir = vbNullString
    agGameFile = vbNullString
    agGameID = vbNullString
    agIntVersion = vbNullString
    agIsVersion3 = False
    
    'reset collections & objects
    Set agLogs = Nothing
    Set agViews = Nothing
    Set agPics = Nothing
    Set agSnds = Nothing
    Set agInvObj = Nothing
    Set agVocabWords = Nothing
    Set agLogs = New AGILogics
    Set agViews = New AGIViews
    Set agPics = New AGIPictures
    Set agSnds = New AGISounds
    Set agInvObj = New AGIInventoryObjects
    Set agVocabWords = New AGIWordList
    
    'raise error
    On Error GoTo 0: Err.Raise lngError, strErrSource, strError
End Sub

Public Sub NewGame(NewID As String, ByVal NewVersion As String, NewGameDir As String, NewResDir As String, Optional ByVal TemplateDir As String = vbNullString)

  'creates a new game in NewGameDir
  'if a template directory is passed,
  'use the resources from that template directory
  
  Dim strDirData As String, intFile As Integer
  Dim i As Long, strTempDir As String, lngDirCount As Long
  Dim blnWarnings As Boolean, strTmplResDir As String
  
  On Error Resume Next
  
  'if a game is already open,
  If agGameLoaded Then
    'can't open a game if one is already open
    On Error GoTo 0: Err.Raise vbObjectError + 501, strErrSource, LoadResString(501)
    Exit Sub
  End If
  
  'if not a valid directory
  If Not FileExists(NewGameDir, vbDirectory) Then
    'raise error
    On Error GoTo 0: Err.Raise vbObjectError + 630, strErrSource, Replace(LoadResString(630), ARG1, NewGameDir)
    Exit Sub
  End If
  
  'if a game already exists
  If LenB(Dir(cDir(NewGameDir) & "*.wag")) <> 0 Then
    'game file exists;
    On Error GoTo 0: Err.Raise vbObjectError + 687, strErrSource, LoadResString(687)
    Exit Sub
  
  ElseIf IsSierraGameDir(cDir(NewGameDir)) Then
    'game files exist;
    agGameDir = vbNullString
    agIsVersion3 = False
    On Error GoTo 0: Err.Raise vbObjectError + 687, strErrSource, LoadResString(687)
    Exit Sub
  End If
  
  'set game directory
  agGameDir = cDir(NewGameDir)
  
  'ensure resdir is valid
  If LenB(NewResDir) = 0 Then
    'if blank use default
    NewResDir = agDefResDir
  End If
  
  'if using template
  If LenB(TemplateDir) <> 0 Then
    'template should include dir files, vol files, logic
    'source, words and objects lists;
    'globals lists and layouts
    
    '(first, make sure it's formatted as a directory
    TemplateDir = cDir(TemplateDir)
    'we need to know name of the resource subdir so we can
    'rename it later, if needed
    
    'find first subdir, use that name
    strTmplResDir = Dir(TemplateDir, vbDirectory)
    Do Until Len(strTmplResDir) = 0
      If strTmplResDir <> "." And strTmplResDir <> ".." Then
        If (GetAttr(TemplateDir & strTmplResDir) And vbDirectory) = vbDirectory Then
          'this is it
          lngDirCount = 1
          Exit Do
        End If
      End If
      strTmplResDir = Dir()
    Loop
    
    'if no resdir found, reset template resdir name so it gets handled correctly
    If lngDirCount <> 1 Then
      strTmplResDir = vbNullString
    End If
      
    'copy all files from the templatedir into gamedir
    If Not CopyFolder(TemplateDir, agGameDir, False) Then
      On Error GoTo 0: Err.Raise vbObjectError + 683, strErrSource, Replace(LoadResString(683), ARG1, Err.Description)
      Exit Sub
    End If
    
    'open the game in gamedir
    LoadGame agGameDir
    If Err.Number <> 0 Then
      On Error GoTo 0: Err.Raise vbObjectError + 684, strErrSource, Replace(LoadResString(684), ARG1, Err.Description)
      Exit Sub
    End If
    
    'we need to rename the resdir
    '(have to do this AFTER load, because loading will keep the current
    'resdir that's in the WAG file)
    If NewResDir <> strTmplResDir Then
      'we need to change it
        Name agGameDir & strTmplResDir As agGameDir & NewResDir
    End If
    'then change the resource directory property
    agResDirName = NewResDir
    'update the actual resdir
    agResDir = agGameDir & agResDirName & "\"
    
    'change gameid
    GameID = NewID
    If Err.Number <> 0 Then
      On Error GoTo 0: Err.Raise vbObjectError + 685, strErrSource, Replace(LoadResString(685), ARG1, Err.Description)
      Exit Sub
    End If
      
  'if not using template,
  Else
    'validate new version
    If ValidateVersion(NewVersion) Then
      'ok; set version
      agIntVersion = NewVersion
      'set version3 flag
      agIsVersion3 = (Val(NewVersion) > 3)
    Else
      If Val(NewVersion) < 2 Or Val(NewVersion) > 3 Then
        'not a version 2 or 3 game
        On Error GoTo 0: Err.Raise vbObjectError + 597, strErrSource, LoadResString(597)
      Else
        'unsupported version 2 or 3 game
        On Error GoTo 0: Err.Raise vbObjectError + 543, strErrSource, LoadResString(543)
      End If
      Exit Sub
    End If
    
    'set game id (limit to 6 characters for v2, and 5 characters for v3
    '(don't use the GameID property; gameloaded flag is not set yet
    'so using GameID property will cause error)
    If agIsVersion3 Then
      agGameID = UCase$(Left$(NewID, 5))
    Else
      agGameID = UCase$(Left$(NewID, 6))
    End If
    
    'create empty property file
    agGameFile = agGameDir & agGameID & ".wag"
    Kill agGameFile
    Err.Clear
    Set agGameProps = OpenSettingList(agGameFile)
    agGameProps.Add "# WinAGI Game Property File for " & agGameID
    agGameProps.Add "#"
    WriteGameSetting "General", "WinAGIVersion", "1.2.3"
    
    'set the resource directory name so it can be set up
    ResDirName = NewResDir
    
    'errors after this cause failure
    On Error GoTo ErrHandler
    
    'create default resource directories
    strDirData = String$(768, &HFF)
    If agIsVersion3 Then
      intFile = FreeFile()
      Open agGameDir & agGameID & "DIR" For Binary As intFile
      Put intFile, 1, Chr$(8) & Chr$(0) & Chr$(8) & Chr$(3) & Chr$(8) & Chr$(6) & Chr$(8) & Chr$(9)
      Put intFile, , strDirData
      Put intFile, , strDirData
      Put intFile, , strDirData
      Put intFile, , strDirData
      Close intFile
    Else
      intFile = FreeFile()
      Open agGameDir & "LOGDIR" For Binary As intFile
      Put intFile, 1, strDirData
      Close intFile
      
      intFile = FreeFile()
      Open agGameDir & "PICDIR" For Binary As intFile
      Put intFile, 1, strDirData
      Close intFile
      
      intFile = FreeFile()
      Open agGameDir & "SNDDIR" For Binary As intFile
      Put intFile, 1, strDirData
      Close intFile
      
      intFile = FreeFile()
      Open agGameDir & "VIEWDIR" For Binary As intFile
      Put intFile, 1, strDirData
      Close intFile
    End If
  
    'set flag indicating directories expanded
    agDirExpanded = True
    
    'create default vocabulary word list
    Set agVocabWords = New AGIWordList
    'use loaded argument to force load of the new wordlist
    agVocabWords.Init True
    agVocabWords.AddWord "a", 0
    agVocabWords.AddWord "anyword", 1
    agVocabWords.AddWord "rol", 9999
    agVocabWords.Save
    
    'create inventory objects list
    Set agInvObj = New AGIInventoryObjects
    'use loaded argument to force load of new inventory list
    agInvObj.Init True
    agInvObj.Add "?", 0
    agInvObj.Save
    
    'assign and adust commands based on AGI version
    AssignCommands
    CorrectCommands agIntVersion
  
    'add logic zero
    agLogs.Add 0
    agLogs(0).Clear
    agLogs(0).Save
    
    'force id reset
    blnSetIDs = False
    
    'set open flag, so properties can be updated
    agGameLoaded = True
  End If
  
  'set resource directory
  Debug.Assert LenB(agResDirName) <> 0
  Debug.Assert agResDir = agGameDir & agResDirName & "\"
  
  On Error Resume Next
  'ensure resource directory exists
  If Not FileExists(agGameDir & agResDirName, vbDirectory) Then
    MkDir agGameDir & agResDirName
    'if can't create the resources directory
    If Err.Number <> 0 Then
      'note the problem in the error log as a warning
      RecordLogEvent 0, "Can't create " & agResDir
      'use main directory
      agResDir = agGameDir
      'set warning flag
      blnWarnings = True
    End If
  End If
  
  'for non-template games, save the source code for logic 0
  If LenB(TemplateDir) = 0 Then
    agLogs(0).SaveSource
  End If
  
  'save gameID, version, directory resource name to the property file;
  'rest of properties need to be set by the calling function
  WriteGameSetting "General", "GameID", agGameID
  WriteGameSetting "General", "Interpreter", agIntVersion
  WriteGameSetting "General", "ResDir", agResDirName
  
  'if errors
  If blnWarnings Then
    On Error GoTo 0: Err.Raise vbObjectError + 637, strErrSource, LoadResString(637)
  End If
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  'reset collections & objects
  Set agLogs = Nothing
  Set agViews = Nothing
  Set agPics = Nothing
  Set agSnds = Nothing
  Set agInvObj = Nothing
  Set agVocabWords = Nothing
  Set agLogs = New AGILogics
  Set agViews = New AGIViews
  Set agPics = New AGIPictures
  Set agSnds = New AGISounds
  Set agInvObj = New AGIInventoryObjects
  Set agVocabWords = New AGIWordList
  
  On Error GoTo 0: Err.Raise vbObjectError + 631, strErrSrc, Replace(LoadResString(631), ARG1, CStr(lngError) & ":" & strError)
End Sub

Public Property Get VocabularyWords() As AGIWordList
  
  Set VocabularyWords = agVocabWords
End Property
Private Sub Class_Initialize()
  Dim rtn As Long
  
  On Error GoTo ErrHandler
  
  'save copy of this object to allow internal access
  Set agMainGame = Me
  
  'set events object
  Set agGameEvents = New AGIGameEvents
  
  'initialize arrays that serve as constants
  InitializeAGI
  
  'set collections
  Set agLogs = New AGILogics
  Set agPics = New AGIPictures
  Set agSnds = New AGISounds
  Set agViews = New AGIViews
  Set agInvObj = New AGIInventoryObjects
  Set agVocabWords = New AGIWordList
  
  'initialize logic source settings object and command info
  Set agMainLogSettings = New AGILogicSourceSettings
  Set agCmdCol = New AGICommands
  Set agTestCmdCol = New AGITestCommands
  'assign commands
  AssignCommands
  
  'assign default reserved defines
  AssignReservedDefines
  
  strErrSource = "agiGame"
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub



Private Sub Class_Terminate()

  'release objects
  Set agLogs = Nothing
  Set agPics = Nothing
  Set agSnds = Nothing
  Set agViews = Nothing
  
  Set agInvObj = Nothing
  Set agVocabWords = Nothing

  Set agGameEvents = Nothing
  Set agMainGame = Nothing
End Sub


