Attribute VB_Name = "AGIObjects"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E0671810195"
Option Explicit

'variables needed across all objects in the
'game system

'version used by save/load games
'be sure to pad it to be 16 characters long!!
'constants for loading/saving game & resource properties
Public Const SP_SEPARATOR = 128 '???separator???
Public Const PC_GAMEDESC = 129
Public Const PC_GAMEAUTHOR = 130
Public Const PC_GAMEID = 131
Public Const PC_INTVERSION = 132
Public Const PC_GAMELAST = 133
Public Const PC_GAMEVERSION = 134
Public Const PC_GAMEABOUT = 135
Public Const PC_GAMEEXEC = 136
Public Const PC_RESDIR = 137
Public Const PC_DEFSYNTAX = 138 'not used anymore...
Public Const PC_INVOBJDESC = 144
Public Const PC_VOCABWORDDESC = 160
Public Const PC_PALETTE = 172
Public Const PC_USERESNAMES = 180
Public Const PC_LOGIC = 192
Public Const PC_PICTURE = 208
Public Const PC_SOUND = 224
Public Const PC_VIEW = 240
Public Const PT_ID = 0
Public Const PT_DESC = 1
Public Const PT_COMPCRC32 = 2
Public Const PT_CRC32 = 3
Public Const PT_KEY = 4
Public Const PT_INST0 = 5
Public Const PT_INST1 = 6
Public Const PT_INST2 = 7
Public Const PT_MUTE0 = 8
Public Const PT_MUTE1 = 9
Public Const PT_MUTE2 = 10
Public Const PT_MUTE3 = 11
Public Const PT_TPQN = 12
Public Const PT_ROOM = 13
Public Const PT_VIS0 = 14
Public Const PT_VIS1 = 15
Public Const PT_VIS2 = 16
Public Const PT_VIS3 = 17
Public Const PT_BKIMG = 18
Public Const PT_BKTRANS = 19
Public Const PT_BKPOS = 20
Public Const PT_BKSZ = 21
Public Const PT_SIZE = 254
Public Const PT_ALL = 255

Public agMainGame As AGIGame
Public agMainLogSettings As AGILogicSourceSettings
Public agCmdCol As AGICommands
Public agTestCmdCol As AGITestCommands
Public agGameEvents As AGIGameEvents

Public agGameProps As StringList

'status of game load
Public agGameLoaded As Boolean

'game compile variables
Public agCompGame As Boolean
Public agCancelComp As Boolean
Public agChangeVersion As Boolean

'local variable(s) to hold property Value(s)
'for game properties which need to be accessible
'from all objects in the game system
Public agLogs As AGILogics
Public agSnds As AGISounds
Public agViews As AGIViews
Public agPics As AGIPictures
Public agInvObj As AGIInventoryObjects
Public agVocabWords As AGIWordList

Public agGameDir As String
Public agResDir As String
Public agResDirName As String
Public agDefResDir As String
Public agTemplateDir As String
Public agGameID As String
Public agAuthor As String
Public agLastDate As Date
Public agDescription As String
Public agIntVersion As String
Public agIsVersion3 As Boolean
Public agAbout As String
Public agGameVersion As String
Public agGameFile As String

  'settings strategy: open the file, store in a stringlist; access the stringlist
  'as needed to make changes; include a ForceWrite option so results can be written periodically
  'when game (or app) closes, final save of the file
Public agGameSettings As StringList

Public agMaxVol0 As Long
Public agMaxVolSize As Long
Public agCompileDir As String
Public agDirExpanded As Boolean
Public agPlatformType As Long
  ' 0 = none
  ' 1 = DosBox
  ' 2 = ScummVM
  ' 3 = NAGI
  ' 4 = Other
Public agPlatformFile As String
Public agPlatformOpts As String
Public agDOSExec As String

Public agSrcExt As String

'error number and string to return error values
'from various functions/subroutines
Public lngError As Long
Public strError As String
Public strErrSrc As String

'arrays which will be treated as constants
'rev colors have red and blue components switched
'so api functions using colors work correctly
Public lngEGARevCol(15) As Long
Public lngEGACol(15) As Long
Public agColor(15) As AGIColors
Public bytEncryptKey(10) As Byte ' = "Avis Durgan"
Public ResTypeAbbrv(3) As String '("LOG", "PIC", "VIEW, "SND"
Public agResTypeName(3) As String '("Logic", "Picture", "View", "Sound")
Public agVersion(15) As String '(2.917, etc)

'game constants
Public Const MAX_RES_SIZE = 65530
Public Const MAX_LOOPS = 16
Public Const MAX_CELS = 32
Public Const MAX_ITEMS = 256
Public Const MAX_VOL_FILES = 65
Public Const MAX_CEL_WIDTH = 160
Public Const MAX_CEL_HEIGHT = 168
Public Const MAX_GROUP_NUM = 65535
Public Const MAX_WORD_GROUPS = 10000
'Public Const MAX_VOLSIZE = 1047552  '= 1024 * 1023
Public Const WORD_SEPARATOR As String = " | "

'temp file location
Public TempFileDir As String
  

Public Sub ConvertWag()

  'converts a v1.2.1 propfile to v1.2.2 proplist
  
  '1.2.1 properties use the following format for the property file:
  ' CPRLL<data>
  'where C= PropCode, P=PropNum, R=ResNum, LL=length of data (as integer)
  '      <data>= property data
  'last line of file should be version code
  
  Dim intFile As Integer
  Dim bytData() As Byte
  Dim lngCount As Long, lngPos As Long
  Dim strValue As String
  Dim i As Long
  Dim ResNum As Byte, PropType As Byte
  Dim PropCode As Byte, PropSize As Long
  Dim blnFoundID As Boolean, blnFoundVer As Boolean
  
  On Error GoTo ErrHandler
  
  'remove everything except first line in wag file
  Do Until agGameProps.Count = 1
    agGameProps.Delete agGameProps.Count - 1
  Loop
  agGameProps.Add "#"
  agGameProps.Add "# WinAGI Game Property File"
  agGameProps.Add "# converted from version 1.2.1"
  agGameProps.Add "#"
  agGameProps.Add "[General]"
  agGameProps.Add "   WinAGIVersion = 1.2.3"
  
  'open old file
  intFile = FreeFile()
  Open agGameFile For Binary As intFile
  
  'verify version
  strValue = String$(16, 0)
  'adjust position to compensate for length of variable
  '(and fact that get is '1' based): LOF - 16 + 1
  Get intFile, LOF(intFile) - 15, strValue
  
  'if version is incompatible
  Select Case strValue
  Case WINAGI_VERSION_1_2, WINAGI_VERSION_1_0, WINAGI_VERSION_BETA
    'ok
  Case Else
    Close intFile
    'return nothing
    Set agGameProps = Nothing
    Exit Sub
  End Select
  
  'don't copy version line into buffer
  lngCount = LOF(intFile) - 16
  If lngCount > 0 Then
    ReDim bytData(lngCount - 1)
    Get intFile, 1, bytData
  Else
    'set to zero
    lngCount = 0
  End If
  Close intFile
  lngPos = 0
   
  'get codes
  Do Until lngPos >= lngCount
    'reset propval
    strValue = vbNullString
    'get prop code
    PropCode = bytData(lngPos)
    PropType = bytData(lngPos + 1)
    ResNum = bytData(lngPos + 2)
    PropSize = bytData(lngPos + 3) + 256 * bytData(lngPos + 4)
    For i = 1 To PropSize
      strValue = strValue & Chr$(bytData(lngPos + 4 + i))
    Next i
      
    Select Case PropCode
    Case Is >= PC_LOGIC
      Select Case PropCode
      Case PC_LOGIC
        Select Case PropType
        Case PT_ID
          WriteGameSetting "Logic" & CStr(ResNum), "ID", strValue
        Case PT_DESC
          WriteGameSetting "Logic" & CStr(ResNum), "Description", strValue
        Case PT_CRC32
          WriteGameSetting "Logic" & CStr(ResNum), "CRC32", "&H" & strValue
        Case PT_COMPCRC32
          WriteGameSetting "Logic" & CStr(ResNum), "CompCRC32", "&H" & strValue
        Case PT_ROOM
          If ResNum = 0 Then
            'force to false
            strValue = "False"
          End If
          WriteGameSetting "Logic" & CStr(ResNum), "IsRoom", strValue
        Case PT_SIZE
          WriteGameSetting "Logic" & CStr(ResNum), "Size", strValue
          
        Case Else
          'unknown code; ignore it
          Debug.Assert False
        End Select
      
      Case PC_PICTURE
        Select Case PropType
        Case PT_ID
          WriteGameSetting "Picture" & CStr(ResNum), "ID", strValue
        Case PT_DESC
          WriteGameSetting "Picture" & CStr(ResNum), "Description", strValue
        Case PT_SIZE
          WriteGameSetting "Picture" & CStr(ResNum), "Size", strValue
        Case PT_BKIMG
          WriteGameSetting "Picture" & CStr(ResNum), "BkgdImg", strValue
        Case PT_BKPOS
          WriteGameSetting "Picture" & CStr(ResNum), "BkgdPosn", strValue
        Case PT_BKSZ
          WriteGameSetting "Picture" & CStr(ResNum), "BkgdSize", strValue
        Case PT_BKTRANS
          WriteGameSetting "Picture" & CStr(ResNum), "BkgdTrans", strValue
        Case Else
          'unknown code; ignore it
          Debug.Assert False
        End Select
        
      Case PC_SOUND
        Select Case PropType
        Case PT_ID
          WriteGameSetting "Sound" & CStr(ResNum), "ID", strValue
        Case PT_DESC
          WriteGameSetting "Sound" & CStr(ResNum), "Description", strValue
        Case PT_SIZE
          WriteGameSetting "Sound" & CStr(ResNum), "Size", strValue
        Case PT_KEY
          WriteGameSetting "Sound" & CStr(ResNum), "Key", strValue
        Case PT_TPQN
          WriteGameSetting "Sound" & CStr(ResNum), "TQPN", strValue
        Case PT_INST0
          WriteGameSetting "Sound" & CStr(ResNum), "Inst0", strValue
        Case PT_INST1
          WriteGameSetting "Sound" & CStr(ResNum), "Inst1", strValue
        Case PT_INST2
          WriteGameSetting "Sound" & CStr(ResNum), "Inst2", strValue
        Case PT_MUTE0
          WriteGameSetting "Sound" & CStr(ResNum), "Mute0", strValue
        Case PT_MUTE1
          WriteGameSetting "Sound" & CStr(ResNum), "Mute1", strValue
        Case PT_MUTE2
          WriteGameSetting "Sound" & CStr(ResNum), "Mute2", strValue
        Case PT_MUTE3
          WriteGameSetting "Sound" & CStr(ResNum), "Mute3", strValue
        Case PT_VIS0
          WriteGameSetting "Sound" & CStr(ResNum), "Visible0", strValue
        Case PT_VIS1
          WriteGameSetting "Sound" & CStr(ResNum), "Visible1", strValue
        Case PT_VIS2
          WriteGameSetting "Sound" & CStr(ResNum), "Visible2", strValue
        Case PT_VIS3
          WriteGameSetting "Sound" & CStr(ResNum), "Visible3", strValue
        Case Else
          'unknown code; ignore it
          Debug.Assert False
        End Select
        
      Case PC_VIEW
        Select Case PropType
        Case PT_ID
          WriteGameSetting "View" & CStr(ResNum), "ID", strValue
        Case PT_DESC
          WriteGameSetting "View" & CStr(ResNum), "Description", strValue
        Case PT_SIZE
          WriteGameSetting "View" & CStr(ResNum), "Size", strValue
        Case Else
          'unknown code; ignore it
          Debug.Assert False
        End Select
      End Select
      
    Case PC_GAMEDESC
      WriteGameSetting "General", "Description", strValue
      
    Case PC_GAMEAUTHOR
      WriteGameSetting "General", "Author", strValue
      
    Case PC_GAMEID
    blnFoundID = Len(strValue) > 0
      WriteGameSetting "General", "GameID", strValue
      
    Case PC_INTVERSION
      WriteGameSetting "General", "Interpreter", strValue
      blnFoundVer = Len(strValue) > 0
      
    Case PC_GAMEABOUT
      WriteGameSetting "General", "About", strValue
    
    Case PC_GAMEVERSION
      WriteGameSetting "General", "GameVersion", strValue
    
    Case PC_RESDIR
      WriteGameSetting "General", "ResDir", strValue
      
    Case PC_GAMELAST
      WriteGameSetting "General", "LastEdit", strValue
      
    Case PC_INVOBJDESC
      WriteGameSetting "OBJECT", "Description", strValue
      
    Case PC_VOCABWORDDESC
      WriteGameSetting "WORDS.TOK", "Description", strValue
      
    Case PC_GAMEEXEC
      WriteGameSetting "General", "Exec", strValue
    
    Case PC_PALETTE
      'convert strvalue into 16 4 byte numbers
      '(actually it's more efficient to use the
      'data array and extract the numbers directly)
      For i = 0 To 15
        strValue = "&H" & Hex(bytData(lngPos + 5 + 4 * i)) & ",&H" & Hex(bytData(lngPos + 6 + 4 * i)) & ",&H" & Hex(bytData(lngPos + 7 + 4 * i)) & ",&H" & Hex(bytData(lngPos + 8 + 4 * i))
        WriteGameSetting "Palette", "Color" & CStr(i), strValue
      Next i
        
    Case PC_USERESNAMES
      WriteGameSetting "General", "UseResNames", strValue
      
    Case Else
      'ignore
    End Select
    
    'add offset to next code (length +5)
    lngPos = lngPos + PropSize + 5
  Loop
  
  'if no id and no intver
  If Not blnFoundID Or Not blnFoundVer Then
    Set agGameProps = Nothing
  End If
  
Exit Sub

ErrHandler:
    Debug.Assert False
    Resume Next
    Set agGameProps = Nothing
'''  End If
End Sub

Public Function ValidateVersion(ByVal NewVersion As String) As Boolean
  'checks NewVersion against list of supported versions
  'returns TRUE if ok
  'returns FALSE if not version 2 or 3 or if unsupported version 2 or version 3
  
  Dim i As Long
  
  'validate against list of numbers
  For i = 0 To UBound(agVersion)
    If NewVersion = agVersion(i) Then
      'ok-
      ValidateVersion = True
      Exit Function
    End If
  Next i
  
  'not found
  ValidateVersion = False
End Function
Public Sub GetGameProperties()

  'what's loaded BEFORE we get here:
  ' General:
  '  GameID
  '  Interpreter
  '  ResDir
  
  
  'ASSUMES a valid game property file has been loaded
  'loads only these properties:
  '
  '  Palette:
  '     all colors
  '
  '  General:
  '     description
  '     author
  '     about
  '     game version
  '     exec program
  '     use res names property
  
  
  Dim i As Long, strSection As String
  Dim strLine As String, strProp() As String
   
  On Error GoTo ErrHandler
  
  Debug.Assert Not (agGameProps Is Nothing)
  
  'Palette: (make sure defaults set first)
  RestoreDefaultColors
  For i = 0 To 15
    strLine = ReadSettingString(agGameProps, "Palette", "Color" & CStr(i), lngEGACol(i))
    'set color
    strProp = Split(strLine, ",")
    'must be 4;
    If UBound(strProp) = 3 Then
      'in case file was edited, trap and ignore errors
      On Error Resume Next
      lngEGACol(i) = &H1000000 * CByte(Val(strProp(0))) + &H10000 * CByte(Val(strProp(1))) + &H100& * CByte(Val(strProp(2))) + CByte(Val(strProp(3)))
      'switch bytes 2 & 4 for reverse color
      lngEGARevCol(i) = &H1000000 * CByte(Val(strProp(0))) + CByte(Val(strProp(1))) + &H100& * CByte(Val(strProp(2))) + &H10000 * CByte(Val(strProp(3)))
      On Error GoTo ErrHandler
    End If
  Next i
  
  'description
  agDescription = ReadSettingString(agGameProps, "General", "Description")
  
  'author
  agAuthor = ReadSettingString(agGameProps, "General", "Author")
  
  'about
  agAbout = ReadSettingString(agGameProps, "General", "About")

  'game version
  agGameVersion = ReadSettingString(agGameProps, "General", "GameVersion")

  'platform
  agPlatformType = ReadSettingLong(agGameProps, "General", "PlatformType", 0)
  
  'platform program
  agPlatformFile = ReadSettingString(agGameProps, "General", "Platform")
  
  'dos executable
  agDOSExec = ReadSettingString(agGameProps, "General", "DOSExec")
  
  'platform options
  agPlatformOpts = ReadSettingString(agGameProps, "General", "PlatformOpts")
  
  'use res names property
  agUseRes = ReadSettingBool(agGameProps, "General", "UseResNames", agUseRes)
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub

Public Function IsSierraGameDir(strDir As String) As Boolean

  'this function will determine if the strDir is a
  'valid sierra AGI game directory
  'it also sets the gameID, if one is found and the version3 flag
  
  Dim strFileName As String
  Dim intFile As Integer, strChunk As String
  
  'search for 'DIR' files
  strFileName = UCase$(Dir(strDir & "*DIR"))
    
  'if files named '*DIR' exist,
  If LenB(strFileName) <> 0 Then
    'this is an AGI game directory-
    IsSierraGameDir = True
    'check for existence of a version three gameid:
    If FileExists(strDir & Left$(strFileName, InStr(1, strFileName, "DIR") - 1) & "VOL.0") Then
      'set gameID
      agGameID = UCase$(Left$(strFileName, InStr(1, strFileName, "DIR") - 1))
      'set version3 flag
      agIsVersion3 = True
    Else
      'if errors occur, exit without setting ID
      On Error GoTo ErrHandler
      
      'clear version3 flag
      agIsVersion3 = False
      'clear ID
      agGameID = vbNullString
      'look for loader file to find ID
      strFileName = UCase$(Dir(strDir & "*.COM"))
      Do Until LenB(strFileName) = 0
        'open file and get chunk
        strChunk = String$(6, 32)
        intFile = FreeFile()
        Open strDir & strFileName For Binary As intFile
        Get intFile, 4, strChunk
        Close intFile
        'if this is a Sierra loader
        If strChunk = "LOADER" Then
          'if not SIERRA.COM
          If strFileName <> "SIERRA.COM" Then
            'use this filename as ID
            agGameID = UCase$(Left$(strFileName, Len(strFileName) - 4))
            Exit Do
          End If
        End If
        'get next file
        strFileName = UCase$(Dir())
      Loop
      'if no loader file found,
      If LenB(agGameID) = 0 Then
        'use default
        agGameID = "AGI"
      End If
    End If
  End If
Exit Function

ErrHandler:
  Err.Clear
  'clear ID
  agGameID = vbNullString
End Function


Function GetIntVersion() As String
  'this function gets the version number of a Sierra AGI game
  'if found, it is validated against list of versions
  'that WinAGI recognizes
  '
  'returns version number for a valid number
  'returns null string for invalid number
  
  Dim strFileName As String, intFile As Integer
  Dim bytBuffer() As Byte, lngPos As Integer
  Dim strVersion As String
  Dim i As Long
  
  On Error GoTo ErrHandler
  
  strFileName = agGameDir & "AGIDATA.OVL"
  
  If FileExists(strFileName) Then
    'open AGIDATA.OVL, coy to buffer, and close
    intFile = FreeFile()
    Open strFileName For Binary As intFile
    ReDim bytBuffer(LOF(intFile))
    Get intFile, 1, bytBuffer()
    Close intFile
  Else
    'no agidata.ovl
    'if version3 is set
    If agIsVersion3 Then
      'use default v3
      GetIntVersion = "3.002149" 'most common version 3
    Else
      'use default version  2.917
      GetIntVersion = "2.917"
    End If
    Exit Function
  End If

  lngPos = 0
  'go until a '2' or '3' is found
  Do
    'check char
    Select Case bytBuffer(lngPos)
    Case 50 '2.xxx format
      strVersion = Chr(bytBuffer(lngPos))
      'get next four chars
      For i = 1 To 4
        lngPos = lngPos + 1
        'just in case, check for end of buffer
        If lngPos > UBound(bytBuffer) Then
          Exit For
        End If
        'add this char
        strVersion = strVersion & Chr(bytBuffer(lngPos))
      Next i
      
      'validate this version
      If ValidateVersion(strVersion) Then
        'return it
        GetIntVersion = strVersion
        Exit Function
      End If
      
    Case 51 '3.xxx.xxx format (for easier manipulation, the second '.' is
            'removed, so result can be converted to a single precision number)
      strVersion = Chr(bytBuffer(lngPos))
      'get next seven chars
      For i = 1 To 7
        lngPos = lngPos + 1
        'just in case, check for end of buffer
        If lngPos > UBound(bytBuffer) Then
          Exit For
        End If
        'add this char (unless it's the second period)
        If lngPos <> 4 Then
          strVersion = strVersion & Chr(bytBuffer(lngPos))
        End If
      Next i
      
      'validate this version
      If ValidateVersion(strVersion) Then
        'return it
        GetIntVersion = strVersion
        Exit Function
      End If
    End Select
    
    'increment pointer
    lngPos = lngPos + 1
  Loop While lngPos <= UBound(bytBuffer)
  
  'if version info not found in AGIDATA.OVL;
  
  'check for version 3:
  If LenB(Dir(agGameDir & agGameID & "DIR")) <> 0 And LenB(Dir(agGameDir & agGameID & "VOL.*")) <> 0 Then
    'use default version 3
    agIsVersion3 = True
    GetIntVersion = "3.002149" 'most common version 3?
  Else
    agIsVersion3 = False
    GetIntVersion = "2.917"  ' This is what we use if we can't find the version number.
                                  ' Version 2.917 is the most common interpreter and
                                  ' the one that all the "new" AGI games should be based on.
  End If
Exit Function

ErrHandler:
  Debug.Assert False
  Resume Next
End Function
Public Sub InitializeAGI()

  'this function loads all the 'constant' arrays
  'with their values
  
  Dim i As Integer, j As Integer
  Dim rtn As Long
  
  'get tempfile directory
  TempFileDir = GetTempFileDir()
  
'''  'get default resource dir name
'''  agDefResDir = Trim$(GetSetting(sAPPNAME, "General", "DefResDir", "src"))
'''
'''  'validate directory
'''  If agDefResDir = vbNullString Then
'''    agDefResDir = "src"
'''  Else
'''    For i = 1 To Len(agDefResDir)
'''      Select Case AscW(Mid$(agDefResDir, i))
'''      Case 32, 92, 47, 58, 42, 63, 34, 60, 62, 124
'''        'invalid character; reset to default
'''        agDefResDir = "src"
'''        Exit For
'''      End Select
'''    Next i
'''  End If
  
  'build crc table
  CRC32Setup
  
  'build extended character table
  InitExtChars
  
  'get max vol size
  agMaxVolSize = 1023 * 1024&
  
  'set max vol0 size
  agMaxVol0 = agMaxVolSize
  
  'abbreviations for version 2 directory files
  ResTypeAbbrv(rtLogic) = "LOG"
  ResTypeAbbrv(rtPicture) = "PIC"
  ResTypeAbbrv(rtSound) = "SND"
  ResTypeAbbrv(rtView) = "VIEW"
  'abbreviations for resource file names
  agResTypeName(rtLogic) = "Logic"
  agResTypeName(rtPicture) = "Picture"
  agResTypeName(rtSound) = "Sound"
  agResTypeName(rtView) = "View"
  
  'colors
  RestoreDefaultColors
  agColor(0) = agBlack
  agColor(1) = agBlue
  agColor(2) = agGreen
  agColor(3) = agCyan
  agColor(4) = agRed
  agColor(5) = agMagenta
  agColor(6) = agBrown
  agColor(7) = agLtGray
  agColor(8) = agDkGray
  agColor(9) = agLtBlue
  agColor(10) = agLtGreen
  agColor(11) = agLtCyan
  agColor(12) = agLtRed
  agColor(13) = agLtMagenta
  agColor(14) = agYellow
  agColor(15) = agWhite
 
 'encryption key values
  bytEncryptKey(0) = 65 'Asc("A")
  bytEncryptKey(1) = 118 'Asc("v")
  bytEncryptKey(2) = 105 'Asc("i")
  bytEncryptKey(3) = 115 'Asc("s")
  bytEncryptKey(4) = 32 'Asc(" ")
  bytEncryptKey(5) = 68 'Asc("D")
  bytEncryptKey(6) = 117 'Asc("u")
  bytEncryptKey(7) = 114 'Asc("r")
  bytEncryptKey(8) = 103 'Asc("g")
  bytEncryptKey(9) = 97 'Asc("a")
  bytEncryptKey(10) = 110 'Asc("n")
  'load versions
  agVersion(0) = "2.089"
  agVersion(1) = "2.272"
  agVersion(2) = "2.411"
  agVersion(3) = "2.425"
  agVersion(4) = "2.426"
  agVersion(5) = "2.435"
  agVersion(6) = "2.439"
  agVersion(7) = "2.440"
  agVersion(8) = "2.915"
  agVersion(9) = "2.917"
  agVersion(10) = "2.936"
  agVersion(11) = "3.002086"
  agVersion(12) = "3.002098"
  agVersion(13) = "3.002102"
  agVersion(14) = "3.002107"
  agVersion(15) = "3.002149"
End Sub

Public Sub RestoreDefaultColors()

  '(note that reverse colors  are in RRGGBB format)
  lngEGARevCol(0) = 0&         'black
  lngEGARevCol(1) = &HA0&      'blue
  lngEGARevCol(2) = &HA000&    'green
  lngEGARevCol(3) = &HA0A0&    'cyan
  lngEGARevCol(4) = &HA00000   'red
  lngEGARevCol(5) = &H8000A0   'magenta
  lngEGARevCol(6) = &HA05000   'brown
  lngEGARevCol(7) = &HA0A0A0   'light gray
  lngEGARevCol(8) = &H505050   'dark gray
  lngEGARevCol(9) = &H5050FF   'light blue
  lngEGARevCol(10) = &HFF50&   'light green
  lngEGARevCol(11) = &H50FFFF  'light cyan
  lngEGARevCol(12) = &HFF5050  'light red
  lngEGARevCol(13) = &HFF50FF  'light magenta
  lngEGARevCol(14) = &HFFFF50  'yellow
  lngEGARevCol(15) = &HFFFFFF  'white
  'note regular colors are 'bbggrr' format
  lngEGACol(0) = 0&         'black
  lngEGACol(1) = &HA00000   'blue
  lngEGACol(2) = &HA000&    'green
  lngEGACol(3) = &HA0A000     'cyan
  lngEGACol(4) = &HA0&   'red
  lngEGACol(5) = &HA00080   'magenta
  lngEGACol(6) = &H50A0     'brown
  lngEGACol(7) = &HA0A0A0   'light gray
  lngEGACol(8) = &H505050   'dark gray
  lngEGACol(9) = &HFF5050   'light blue
  lngEGACol(10) = &H50FF00    'light green
  lngEGACol(11) = &HFFFF50  'light cyan
  lngEGACol(12) = &H5050FF  'light red
  lngEGACol(13) = &HFF50FF  'light magenta
  lngEGACol(14) = &H50FFFF  'yellow
  lngEGACol(15) = &HFFFFFF  'white
End Sub
Public Sub RecordLogEvent(intType As Integer, ByRef strMessage As String)
  'open the log file and write the message
  'intType =0 means warning
  'inttype =1 means error
  Dim intLogFile As Integer
  Dim strType As String
  
  On Error GoTo ErrHandler
  
  intLogFile = FreeFile()
  
  'set type of msg
  If intType = 0 Then
    strType = "WARNING: "
  Else
    strType = "ERROR: "
  End If
    
  Open agGameDir & "errlog.txt" For Append As intLogFile
  Print #intLogFile, Format$(Now(), "Medium Date") & ": " & strType & strMessage
  Close intLogFile
  Err.Clear
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub


Public Sub WriteGameSetting(ByVal Section As String, ByVal Key As String, ByVal Value As String)

  Debug.Assert Not (agGameProps Is Nothing)
  
  WriteAppSetting agGameProps, Section, Key, Value
  
  'if not changing date
  If Key <> "LastDate" Then
    'change lastdate
    agLastDate = Now()
  End If
End Sub


