VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AGIResource"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Description = "Public Const MaxResourceSize = 65530\r\nPublic Type agiResource\r\n    Data(MaxResourceSize - 1) As Byte\r\n    Size As Integer\r\nEnd Type\r\n"
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E06719E0157"
Option Explicit

'
'variables holding internal property values
'

Private mResNum As Byte
Private mResType As AGIResType
Private mResFile As String
Private mVolume As Long
Private mLoc As Long

Private mLoaded As Boolean
Private mInGame As Boolean

Private mSize As Long
Private mSizeInVol As Long

Private mbytData() As Byte 'variable size array that holds resource data
Private mblnEORes As Boolean 'flag indicating pointer is at beginning of resource (CurPos=0)
Private mlngCurPos As Long  'current position of pointer in resource data
                            
'other
Private strErrSource As String

'change event
Public Event Change()
Friend Sub Clear()
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'clears the resource data
  ReDim mbytData(0)
  mlngCurPos = 0
  'size is correctly set to zero on clear; no bytes in resource
  mSize = 0
  'mSizeInVol is undefined
  mSizeInVol = -1
  mblnEORes = False
End Sub

Friend Property Let InGame(blnNew As Boolean)
  mInGame = blnNew
End Property

Public Property Get InGame() As Boolean
  InGame = mInGame
  
End Property


Friend Sub Init(ByVal ResNum As Byte, Optional ByVal VOL As Long = -1, Optional ByVal Loc As Long = -1)
  'Init method only called when a resource is added to a game
  'via the Add method of the resource collections (Views, Logics, Pictures, Sounds)
  'or via the Load<restype> method

  mResNum = ResNum
  mVolume = VOL
  mLoc = Loc
  mInGame = True
End Sub

Public Sub InsertData(NewData As Variant, Optional ByVal InsertPos As Long = -1)
  'inserts newdata into resource at insertpos, if passed,
  'or at end of resource, if not passed
  
  Dim i As Long, lngResEnd As Long
  Dim lngNewDatLen As Long
  Dim bNewData() As Byte
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'array of bytes = OK
  'single byte = OK
  'anything that can be converted to single byte = OK
  
  'not array of bytes and not single byte?
  If VarType(NewData) <> vbArray + vbByte And VarType(NewData) <> vbByte Then
    Do
      'not array anything else?
      If (VarType(NewData) And vbArray) = 0 Then
        'is it a number?
        If IsNumeric(NewData) Then
          'an integer number?
          If Int(NewData) = Val(NewData) Then
            'between 0 and 255?
            If Val(NewData) >= 0 And Val(NewData) <= 255 Then
              'ok!
              Exit Do
            End If
          End If
        End If
      End If
      'error
      On Error GoTo 0: Err.Raise 13, strErrSource, "Type Mismatch"
      Exit Sub
   Loop Until True
  End If
  
  'if an array of bytes:
  If VarType(NewData) = vbArray + vbByte Then
    bNewData = NewData
  'otherwise treat as single byte
  Else
    ReDim bNewData(0)
    bNewData(0) = NewData
  End If
  
  
  'get resource end and length of data being inserted
  Debug.Assert mSize = UBound(mbytData) + 1
  lngResEnd = mSize - 1
  lngNewDatLen = UBound(bNewData) + 1
  
  'if no insert pos passed,
  If InsertPos = -1 Then
    'insert at end
    mSize = mSize + lngNewDatLen
    ReDim Preserve mbytData(mSize - 1)
    For i = 1 To lngNewDatLen
      mbytData(lngResEnd + i) = bNewData(i - 1)
    Next i
  Else
    'validate insert pos
    If InsertPos < 0 Then
      'error
      On Error GoTo 0: Err.Raise vbObjectError + 620, strErrSource, LoadResString(620)
      Exit Sub
    End If
    If InsertPos >= mSize Then
      'error
      On Error GoTo 0: Err.Raise vbObjectError + 620, strErrSource, LoadResString(620)
      Exit Sub
    End If
    
    'increase array size
    mSize = mSize + lngNewDatLen
    ReDim Preserve mbytData(mSize - 1)
    'move current data forward to make room for inserted data
    For i = lngResEnd + lngNewDatLen To InsertPos + lngNewDatLen Step -1
      mbytData(i) = mbytData(i - lngNewDatLen)
    Next i
    'add newdata at insertpos
    For i = 0 To lngNewDatLen - 1
      mbytData(InsertPos + i) = bNewData(i)
    Next i
  End If
  
  'raise change event
  RaiseEvent Change
End Sub


Public Sub NewResource(Optional ByVal Reset As Boolean = False)
  'clears any data for the resource,
  'and marks it as loaded
  'this is needed so new resources can be created and edited
  
  On Error Resume Next
  
  'cant call NewResource if already in a game;
  'clear it instead
  If mInGame Then
    On Error GoTo 0: Err.Raise vbObjectError + 510, strErrSource, LoadResString(510)
    Exit Sub
  End If
  
  'if need to reset
  If Reset Then
    'use unload to force reset
    Unload
  End If
  
  'mark as loaded
  mLoaded = True
  
  'set loc and vol to undefined
  mVolume = -1
  mLoc = -1
  
End Sub

Friend Property Let Number(ByVal NewNum As Byte)
  'should only be called when a resource
  'is being reunumbered or resource is being copied
  
  mResNum = NewNum
End Property


Public Property Get Number() As Byte
  Number = mResNum
 
End Property

Public Sub RemoveData(ByVal RemovePos As Long, Optional ByVal RemoveCount As Long = 1)
  'removes data from RemovePos; if a Count
  'is passed, that number of bytes removed; if no
  'Count is passed, then only one byte removed
  
  Dim i As Long
  Dim lngResEnd As Long
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  lngResEnd = mSize - 1
  
  'validate Count
  If RemoveCount <= 0 Then
    'force back to 1
    RemoveCount = 1
  End If
  
  'validate removepos
  If RemovePos < 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 620, strErrSource, LoadResString(620)
    Exit Sub
  End If
  If RemovePos >= mSize Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 620, strErrSource, LoadResString(620)
    Exit Sub
  End If
  
  'remove by moving data backwards
  For i = RemovePos To mSize - RemoveCount - 1
    mbytData(i) = mbytData(i + RemoveCount)
  Next i
  
  mSize = mSize - RemoveCount
  ReDim Preserve mbytData(mSize - 1)
  
  'raise change event
  RaiseEvent Change
End Sub


Friend Property Let ResFile(ByVal NewFile As String)

  'called by parent resource to set a non-ingame resource file before loading
  mResFile = NewFile
End Property

Public Property Get ResType() As AGIResType
  ResType = mResType

End Property


Friend Sub Save(Optional SaveFile As String)
  'saves a resource into a VOL file if in a game
  'exports the resource if not in a game
  
  On Error GoTo ErrHandler
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'if in game,
  If mInGame Then
    'add resource to VOL file to save it
    AddToVol Me, agIsVersion3
    
    'if error,
    If Err.Number <> 0 Then
      'pass error along
      lngError = Err.Number
      strError = Err.Description
      strErrSrc = Err.Source
      
      On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
      Exit Sub
    End If
  Else
    'export it
    Export SaveFile
  End If
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub

Friend Sub SetRes(NewRes As AGIResource)
  'called by setview, setlog, setpic and setsnd
  'copies entire resource structure
  
  With NewRes
    mResNum = .Number
    mResType = .ResType
    mResFile = .ResFile
    'if resource being copied is in a game,
    If NewRes.InGame Then
      'copy vol and loc info
      mVolume = .Volume
      mLoc = .Loc
    End If
    mLoaded = .Loaded
    mSize = .Size
    
    'ingame property is never copied; only way to
    'change ingame status is to do so through
    'appropriate methods for adding/removing
    'resources to/from game
    
    'resresource data is copied manually as necessary by calling method
    'EORes and CurPos are calculated; don't need to copy them
  End With
End Sub


Friend Sub SetType(ByVal ResType As AGIResType)
  'should only be called when a new resource is initialized

  mResType = ResType
End Sub


Friend Property Let Size(NewSize As Long)

  'only called during game load
  mSize = NewSize
End Property

Public Property Get SizeInVol() As Long
  'returns the size of the resource on the volume
  
  'if in a game
  If mInGame Then
    'if not established yet
    If mSizeInVol = -1 Then
      'get Value
      mSizeInVol = GetSizeInVOL(mVolume, mLoc)
      'if valid Value not returned,
      If mSizeInVol = -1 Then
        On Error GoTo 0: Err.Raise vbObjectError + 625, strErrSource, LoadResString(625)
        Exit Sub
      End If
    End If
  Else
    mSizeInVol = -1
  End If
  
  'return Value
  SizeInVol = mSizeInVol
End Property
Friend Property Let SizeInVol(ByVal NewSize As Long)
  'only AddToVOL calls this
  'after saving a resource to a VOL file
  mSizeInVol = NewSize
  
End Property


Public Property Get Volume() As Long
   
  'if not in a game
  If Not mInGame Then
    Volume = -1
  Else
    Volume = mVolume
  End If
End Property
Friend Property Let Volume(ByVal NewVOL As Long)
  
  mVolume = NewVOL
End Property


Public Property Get ResFile() As String
  'only applies for resources that are not in a game
  
  If Not mInGame Then
    ResFile = mResFile
  Else
    'return nothing
    ResFile = vbNullString
  End If
End Property

Public Property Get Loaded() As Boolean
  Loaded = mLoaded
  
End Property



Public Property Get Loc() As Long
  
  If Not mInGame Then
    Loc = -1
  Else
    Loc = mLoc
  End If
End Property
Friend Property Let Loc(ByVal NewLoc As Long)
  
  mLoc = NewLoc
End Property



Friend Sub Unload()
  mLoaded = False
  'reset resource variables
  ReDim mbytData(0)
  'don't mess with sizes though! they remain accessible even when unloaded
  mblnEORes = True
  mlngCurPos = 0
End Sub

Friend Property Let AllData(NewData() As Byte)

  'only called internally when copying a resource for editing
  'or for creating new compiled resources
  
  'whenever data is set this way, load flag is automatically set
  'to true
  
  On Error GoTo ErrHandler
  
  'assign data
  mbytData = NewData
  mSize = UBound(mbytData) + 1
  
  'reset position
  mlngCurPos = 0
  'reset end of resource
  mblnEORes = False
  'set status as loaded
  mLoaded = True
  
  'raise change event
  RaiseEvent Change
Exit Property

ErrHandler:
  Debug.Assert False
  Resume Next
End Property

Public Sub SetData(NewData() As Byte)

  'allows the entire resource byte array to be replaced
  
  On Error GoTo ErrHandler
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'assign data
  mbytData = NewData
  mSize = UBound(mbytData) + 1
  
  'reset position
  mlngCurPos = 0
  'reset end of resource
  mblnEORes = False
  
  'raise change event
  RaiseEvent Change
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub

Public Property Get AllData() As Byte()
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  'return entire array of data
  AllData = mbytData
End Property


Public Sub Export(ExportFile As String)
  'exports resource to file
  'doesn't affect the current resource filename
  'MUST specify a valid export file
  'if export file already exists, it is overwritten
  'caller is responsible for verifying overwrite is ok or not
  
  Dim strTempFile As String
  Dim intFile As Integer, blnUnload As Boolean
  
  'if no filename passed
  If LenB(ExportFile) = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 599, strErrSource, LoadResString(599)
    Exit Sub
  End If

  On Error Resume Next
  
  'if not loaded
  If Not mLoaded Then
    blnUnload = True
    Load
    'check for error
    If Err.Number <> 0 Then
      On Error GoTo 0: Err.Raise vbObjectError + 601, strErrSource, LoadResString(601)
      Exit Sub
    End If
  End If
  
  'get temporary file
  strTempFile = TempFileName()
  
  'open file for output
  intFile = FreeFile()
  Open strTempFile For Binary As intFile
  'write data
  Put intFile, , mbytData
  
  'if error,
  If Err.Number <> 0 Then
    'close file
    Close intFile
    'erase the temp file
    Kill strTempFile
    'unload if necessary
    If blnUnload Then
      Unload
    End If
    'return error condition
    On Error GoTo 0: Err.Raise vbObjectError + 582, strErrSource, LoadResString(582)
    Exit Sub
  End If
  
  'close file,
  Close intFile
  'unload if necessary
  If blnUnload Then
    Unload
  End If
  
  'if savefile exists
  If FileExists(ExportFile) Then
    'delete it
    Kill ExportFile
    Err.Clear
  End If
  
  'copy tempfile to savefile
  FileCopy strTempFile, ExportFile
  
  'if error,
  If Err.Number <> 0 Then
    'close file
    Close intFile
    'erase the temp file
    Kill strTempFile
    'return error condition
    On Error GoTo 0: Err.Raise vbObjectError + 582, strErrSource, LoadResString(582)
    Exit Sub
  End If
    
  'delete temp file
  Kill strTempFile
  Err.Clear
  
  'if NOT in a game,
  If Not mInGame Then
    'change resfile to match new export file
    mResFile = ExportFile
  End If
End Sub


Public Function GetPos() As Long

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Function
  End If
  
  'returns current position
  GetPos = mlngCurPos
End Function




Friend Sub Import(ImportFile As String)
  'imports resource from a file, and loads it
  'if in a game, it also saves the new
  'resource in a VOL file
  '
  'import does not check if the imported
  'resource is valid or not, nor if it is
  'of the correct resource type
  '
  'the calling program is responsible for
  'that check
  
  Dim intFile As Integer
  
  On Error GoTo ErrHandler
  
  'if no filename passed
  If LenB(ImportFile) = 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 604, strErrSource, LoadResString(604)
    Exit Sub
  End If
  
  'if file doesn't exist
  If Not FileExists(ImportFile) Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 524, strErrSource, Replace(LoadResString(524), ARG1, ImportFile)
    Exit Sub
  End If
  
  'if resource loaded,
  If mLoaded Then
    Unload
  End If
  
  On Error Resume Next
  'open file for binary
  intFile = FreeFile()
  Open ImportFile For Binary As intFile
  'check for file access error
  If Err.Number <> 0 Then
    Err.Clear
    On Error GoTo 0: Err.Raise vbObjectError + 605, strErrSource, Replace(LoadResString(605), ARG1, ImportFile)
    Exit Sub
  End If
  
  If LOF(intFile) = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 605, strErrSource, Replace(LoadResString(605), ARG1, ImportFile)
    Exit Sub
  End If
  
  On Error GoTo ErrHandler
  'load resource from file
  mSize = LOF(intFile)
  ReDim mbytData(mSize - 1)
  Get intFile, 1, mbytData
  
  'close file
  Close intFile
  
  'if in a game
  If mInGame Then
    'save resource
    Save
  Else
    'save the resource filename
    mResFile = ImportFile
  End If
  
  'reset resource markers
  mlngCurPos = 0
  mblnEORes = False
  mLoaded = True
  
  'raise change event
  RaiseEvent Change
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub

Public Function ReadByte(Optional ByVal Pos As Long = MAX_RES_SIZE + 1) As Byte

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Function
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
  
  'if a position is passed
  If Pos <> MAX_RES_SIZE + 1 Then
    'validate position
    Select Case Pos
    Case Is < 0 'no negatives allowed
      lngError = 514
      GoTo ErrHandler
    Case Is >= mSize 'cant read past end
      lngError = 513
      GoTo ErrHandler
    End Select
    'new position is ok; assign it
    mlngCurPos = Pos
  End If
  
  'check for end of resource
  mblnEORes = (mlngCurPos = mSize)
  
  'if at eores,
  If mblnEORes Then
    lngError = 513
    GoTo ErrHandler
  End If
  
  ReadByte = mbytData(mlngCurPos)
  mlngCurPos = mlngCurPos + 1
  
  'set end of resource Value
  mblnEORes = (mlngCurPos = mSize)
Exit Function

ErrHandler:
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Function


Friend Sub Load()
  'loads the data for this resource
  'from its VOL file, if in a game
  'or from its resfile
  
  Dim intFile As Integer
  Dim bytLow As Byte, bytHigh As Byte
  Dim bytVolNum As Byte, blnIsPicture As Byte
  Dim lngExpandedSize As Long
  Dim strLoadResFile As String
  
  'NOTE: because GET is '1' based, need to correct all 'GET' statements
  'by adding one to position
  
  On Error Resume Next
  
  'if already loaded,
  If mLoaded Then
    'error- resource already loaded
    On Error GoTo 0: Err.Raise vbObjectError + 511, strErrSource, LoadResString(511)
    Exit Sub
  End If
  
  'if in a game,
  If mInGame Then
    'resource data is loaded from the AGI VOL file
    'build filename
    If agIsVersion3 Then
      strLoadResFile = agGameDir & agGameID & "VOL." & CStr(mVolume)
    Else
      strLoadResFile = agGameDir & "VOL." & CStr(mVolume)
    End If
  Else
    'if no filename
    If LenB(mResFile) = 0 Then
      'error- nothing to load
      On Error GoTo 0: Err.Raise vbObjectError + 626, strErrSource, LoadResString(626)
      Exit Sub
    Else
      'use resource filename
      strLoadResFile = mResFile
    End If
  End If
  
  'verify file exists
  If Not FileExists(strLoadResFile) Then
    On Error GoTo 0: Err.Raise vbObjectError + 606, strErrSource, Replace(LoadResString(606), ARG1, JustFileName(strLoadResFile))
    Exit Sub
  End If
    
  'open volfile
  intFile = FreeFile
  Open strLoadResFile For Binary Access Read As intFile
    
  'verify file opened okay
  If Err.Number <> 0 Then
    Err.Clear
    Close intFile
    lngError = 502
    strError = Replace(Replace(LoadResString(502), ARG1, CStr(Err.Number)), "%2", JustFileName(strLoadResFile))
    GoTo ErrHandler
  End If
  
  'from here, all errors are trapped
  'also, from this point on, need to do cleanup if
  'an error is found
  On Error GoTo ErrHandler
  
  'verify resource is within file bounds
  If mLoc > LOF(intFile) Then
    lngError = 505
    strError = Replace(Replace(LoadResString(505), ARG1, CStr(mLoc)), "%2", JustFileName(strLoadResFile))
    GoTo ErrHandler
  End If
    
  'if loading from a VOL file (i.e. is in a game)
  If mInGame Then
    'read header bytes
    Get intFile, mLoc + 1, bytHigh
    Get intFile, , bytLow
  
    If (Not ((bytHigh = &H12) And (bytLow = &H34))) Then
      lngError = 506
      strError = Replace(Replace(LoadResString(506), ARG1, CStr(mLoc)), "%2", JustFileName(strLoadResFile))
      GoTo ErrHandler
    End If
  
    'get volume where this resource is stored
    Get intFile, , bytVolNum
  
    'determine if this resource is a picture
    blnIsPicture = ((bytVolNum And &H80) = &H80)
  
    'get size info
    Get intFile, , bytLow
    Get intFile, , bytHigh
    mSize = CLng(bytHigh) * 256 + CLng(bytLow)
  
    'if version3,
    If agIsVersion3 Then
      'the size retreived is the expanded size
      lngExpandedSize = mSize
      'now get compressed size
      Get intFile, , bytLow
      Get intFile, , bytHigh
      mSize = CLng(bytHigh) * 256 + CLng(bytLow)
    End If
  Else
    'get size from total file length
    mSize = LOF(intFile)
    'version 3 files are never compressed when
    'loaded as individual files
    lngExpandedSize = mSize
  End If
    
  'get resource data
  ReDim mbytData(mSize - 1)
  
  Get intFile, , mbytData
  
  'close file
  Close intFile
  
  'if version3
  If agIsVersion3 Then
    'if resource is a compressed picture
    If blnIsPicture Then
      'pictures use this decompression
      DecompressPicture mbytData()
      'set size after decompression
      mSize = UBound(mbytData) + 1
      
    'if resource is 'regular' compressed,
    ElseIf mSize <> lngExpandedSize Then
      'all other resources use LZW compression
      ExpandV3ResData mbytData(), lngExpandedSize
      'set true size after expanding
      mSize = UBound(mbytData) + 1
      Debug.Assert mSize = lngExpandedSize
      
      'if expanding a logic
      If mResType = rtLogic Then
        Dim lngMsgStart As Long, NumMessages As Byte
        Dim lngPos As Long, lngMsgEnd As Long
        Dim EncryptionStart As Long, i As Long
        
        'need to toggle encryption of message section
        '(since compressed version three logics don't
        'have their message sections encrypted)
        '
        'get start of msg section
        lngMsgStart = mbytData(1) * 256 + mbytData(0) + 2
        
        'read in number of messages
        NumMessages = mbytData(lngMsgStart)
        If NumMessages > 0 Then
          'retrieve and adjust end of message section
          lngMsgEnd = lngMsgStart + 256 * mbytData(lngMsgStart + 2) + mbytData(lngMsgStart + 1)
          'set start of encryption (at end of msg section table)
          EncryptionStart = lngMsgStart + 3 + NumMessages * 2
          'mark start of encrypted data (to align encryption string)
          'step through all bytes and toggle encryption
          For lngPos = EncryptionStart To lngMsgEnd
            mbytData(lngPos) = mbytData(lngPos) Xor bytEncryptKey((lngPos - EncryptionStart) Mod 11)
          Next lngPos
        End If
      End If
    End If
  End If
    
  'reset resource markers
  mlngCurPos = 0
  mblnEORes = False
  mLoaded = True
Exit Sub

ErrHandler:
  'if an AGI error, error number and string already set;
  'if got here because of a system error,
  If Err.Number <> 0 Then
    lngError = 507
    strError = Replace(Replace(LoadResString(507), ARG1, CStr(Err.Number)), "%2", JustFileName(strLoadResFile))
    Err.Clear
  End If
  On Error GoTo 0
  'close file
  Close intFile
  
  'reset resource markers
  mlngCurPos = 0
  mblnEORes = False
  mLoaded = False
  'raise the error
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, strError
End Sub



Public Function ReadWord(Optional ByVal Pos As Long = -1, Optional ByVal MSLS As Boolean = False) As Long
  
  Dim bytLow As Byte, bytHigh As Byte
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Function
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
    
  'if a position is passed,
  If Pos <> -1 Then
    'validate position
    Select Case Pos
    Case Is < 0 'no non-negative values
      lngError = 514
      GoTo ErrHandler
    Case Is >= mSize  'cant read past end
      lngError = 513
      GoTo ErrHandler
    End Select
    
    'new position is ok; assign it
    mlngCurPos = Pos
  End If
  
  'if reading in MS-LS format,
  If MSLS Then
    'read the two bytes, high first
    bytHigh = mbytData(mlngCurPos)
    bytLow = mbytData(mlngCurPos + 1)
  Else
    'read the two bytes, low first
    bytLow = mbytData(mlngCurPos)
    bytHigh = mbytData(mlngCurPos + 1)
  End If
  
  'calculate word Value
  ReadWord = CLng(CLng(bytHigh) * 256 + CLng(bytLow))
  'adjust intCount
  mlngCurPos = mlngCurPos + 2
  'check for end of resource
  mblnEORes = (mlngCurPos = mSize)
Exit Function

ErrHandler:
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Function


Public Sub SetPos(ByVal Pos As Long)
  'validate position
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
  
  Select Case Pos
  Case Is < 0 'no non-negative values
    lngError = 514
    GoTo ErrHandler
  Case Is > mSize  'cant point past eor
    lngError = 513
    GoTo ErrHandler
  End Select
  
  'new position is ok; assign it
  mlngCurPos = Pos
  'set eor Value
  mblnEORes = (mlngCurPos = mSize)
  'set bor Value
'  mblnBORes = (mlngCurPos = 0) '''???? why is this commented out?
Exit Sub

ErrHandler:
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Sub


Friend Sub WriteWord(ByVal IntegerIn As Long, Optional ByVal Pos As Long = -1, Optional blnMSLS As Boolean = False)
  
  Dim bytHigh As Byte, bytLow As Byte
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
  
  'if a location is passed,
  If Pos <> -1 Then
    'validate the new position
    Select Case Pos
    Case Is >= MAX_RES_SIZE - 1
      lngError = 516
      GoTo ErrHandler
    Case Is < 0
      lngError = 514
      GoTo ErrHandler
    Case Is > mSize
      lngError = 513
      GoTo ErrHandler
    End Select
  Else
    'otherwise, default to end of resource
    Pos = mSize
  End If
  
  'because of using long Value to simulate an unsigned integer (word in C++)
  'need to validate that the passed data does not exceed this Value
  If IntegerIn > &HFFFF& Then
    lngError = 557
    GoTo ErrHandler
  End If
  
  'negative numbers aren't allowed; only unsigned integer
  If IntegerIn < 0 Then
    lngError = 557
    GoTo ErrHandler
  End If
  
  'if past end of resource (by two bytes)
  If Pos = mSize Then
    'adjust to make room for new data being added
    mSize = mSize + 2
    ReDim Preserve mbytData(mSize - 1)
  'if past end of resource (by one byte)
  ElseIf Pos = Size - 1 Then
    'adjust to make room for new data being added
    mSize = mSize + 1
    ReDim Preserve mbytData(mSize - 1)
  End If
  
  'split integer portion into two bytes
  bytHigh = CByte(IntegerIn \ 256)
  bytLow = CByte(IntegerIn Mod 256)
  'if wrting in MSLS mode,
  If blnMSLS Then
    'save input data
    mbytData(Pos) = bytHigh
    mbytData(Pos + 1) = bytLow
  Else
    mbytData(Pos) = bytLow
    mbytData(Pos + 1) = bytHigh
  End If
  
  'set current position to Pos+1
  mlngCurPos = Pos + 2
  
  'set EORes Value
  mblnEORes = (mlngCurPos = mSize)
  
  'raise change event
  RaiseEvent Change
Exit Sub

ErrHandler:
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Sub

Friend Sub WriteByte(ByVal InputByte As Byte, Optional ByVal Pos As Long = -1)
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'this assertion will fail for a cleared resource, because the stupid array
  'has a upper bound of zero when not defined, so ubound+1=1 and size=0 is actually OK
  Debug.Assert mSize = UBound(mbytData) + 1 Or (mSize = 0 And UBound(mbytData) = 0)
  
  'if a location is passed,
  If Pos <> -1 Then
    'validate the new position
    Select Case Pos
    Case Is >= MAX_RES_SIZE
      lngError = 516
      GoTo ErrHandler
    Case Is < 0
      lngError = 514
      GoTo ErrHandler
    Case Is > mSize
      lngError = 517
      GoTo ErrHandler
    End Select
  Else
    'otherwise, default to end of resource
    Pos = mSize
  End If
  
  'if currently pointing past end of data,
  If Pos = mSize Then
    'adjust to make room for new data being added
    mSize = mSize + 1
    ReDim Preserve mbytData(mSize - 1)
  End If
  
  'save input data
  mbytData(Pos) = InputByte
  
  'set current position to Pos
  mlngCurPos = Pos + 1
  
  'set EORes Value
  mblnEORes = (mlngCurPos = mSize)
  
  'raise change event
  RaiseEvent Change
Exit Sub

ErrHandler:
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Sub

  

Public Property Get EORes() As Boolean
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  EORes = mblnEORes
End Property





Public Property Let Data(ByVal Pos As Long, ByVal NewData As Byte)
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
  
  'validate position
  Select Case Pos
  Case Is < 0 'no non-negative values
    lngError = 514
    GoTo ErrHandler
  Case Is > mSize - 1 'cant point past eor
    lngError = 515
    GoTo ErrHandler
  End Select
    
  'store new data
  mbytData(Pos) = NewData
  
  'raise change event
  RaiseEvent Change
Exit Property

ErrHandler:
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Property


Public Property Get Data(ByVal Pos As Long) As Byte

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  Debug.Assert mSize = UBound(mbytData) + 1
  
  'validate position
  Select Case Pos
  Case Is < 0 'no non-negative values
    lngError = 514
    GoTo ErrHandler
  Case Is >= mSize 'cant point past eor
    lngError = 515
    GoTo ErrHandler
  End Select
    
    'return data
    Data = mbytData(Pos)
Exit Property

ErrHandler:
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError + lngError, strErrSource, LoadResString(lngError)
End Property




Public Property Get Size() As Long
  'returns the uncompressed size of the resource
  
  'location of resource doesn't matter; should always have a size
    
  'if not established yet
'''  Debug.Assert mSize <> -1
  If mSize = -1 Then
    'can't be true if resource is loaded
    Debug.Assert mLoaded = False
    'load it to get size
    Load
    'then unload
    Unload
  End If
  
  'if still -1? not possible unless error encountered in load method
  If mSize = -1 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  'return the size
  Size = mSize
End Property

Private Sub Class_Initialize()
  strErrSource = "WINAGI.agiResource"
  ReDim mbytData(0)
  mSize = 1
  
  'VOL, Loc, and size in vol are undefined until loaded for first time
  '(or set by game property read)
  mVolume = -1
  mLoc = -1
  mSizeInVol = -1
End Sub

