VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "AGIWordList"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Collection" ,"agiWord"
Attribute VB_Ext_KEY = "Member0" ,"agiWord"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E06727D02ED"
Option Explicit
'collection of words (agiWord objects)

Private mWordCol As Collection
Private mGroupCol As Collection

Private mResFile As String
Private mDescription As String

Private mInGame As Boolean
Private mIsDirty As Boolean
Private mWriteProps As Boolean
Private mLoaded As Boolean

Private strErrSource As String

Public Sub NewWords()
  'marks the resource as loaded
  'this is needed so new resources can be created and edited
  
  On Error Resume Next
  
  'if already loaded
  If mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 642, strErrSource, LoadResString(642)
    Exit Sub
  End If
  
  'cant call NewResource if already in a game;
  'clear it instead
  If mInGame Then
    On Error GoTo 0: Err.Raise vbObjectError + 510, strErrSource, LoadResString(510)
    Exit Sub
  End If
  
  'mark as loaded
  mLoaded = True
  
  'clear resname and description
  mResFile = vbNullString
  mDescription = vbNullString
  
  'use clear method to ensure list is reset
  Clear
  
  'add default groups
  AddGroup 0
  AddWord "a", 0
  AddGroup 1
  AddWord "anyword", 1
  AddGroup 9999
  AddWord "rol", 9999
End Sub
Private Sub CompileWinAGI(CompileFile As String)

  'for the text file,
  'words are saved as one group to a line
  'in numerical order by group
  'each line contains the group number, and
  'the the group's list of words in alphabetical
  'order (separated by a tab character)
  '(empty groups are skipped during the save)
  
  'first line is version identifier
  'file description is saved at end of words
  
  Dim strTempFile As String
  Dim intFile As Integer
  Dim tmpGroup As AGIWordGroup
  Dim tmpWord As AGIWord, i As Long, j As Long
  Dim strOutput As String
  
  On Error GoTo ErrHandler
  
  'if no name
  If LenB(CompileFile) = 0 Then
    On Error GoTo 0
    'raise error
    On Error GoTo 0: Err.Raise vbObjectError + 615, strErrSource, LoadResString(615)
    Exit Sub
  End If
  
  'get temporary file
  strTempFile = TempFileName()
    
  'open file for output
  intFile = FreeFile()
  Open strTempFile For Output As intFile
  
  'print version
  Print #intFile, WINAGI_VERSION_1_2
  
  'step through all groups
  For j = 0 To Me.GroupCount - 1
    Set tmpGroup = Me.Group(j)
    'if this group has at least one word
    If tmpGroup.WordCount > 0 Then
      'add group number to output line
      strOutput = CStr(tmpGroup.GroupNum)
      'step through all words
      For i = 0 To tmpGroup.WordCount - 1
      'For Each tmpWord In tmpGroup
        'add tab character and this word
        strOutput = strOutput & vbTab & tmpGroup.Word(i)
      Next i
      'add this group to output file
      Print #intFile, strOutput
    End If
  Next j
  
  'if there is a description
  If LenB(mDescription) <> 0 Then
    'print eof marker
    Print #intFile, Chr$(255) & Chr$(255)
    'print description
    Print #intFile, mDescription
  End If
  
  'close file
  Close intFile
  
  'if CompileFile exists
  If FileExists(CompileFile) Then
    'delete it
    Kill CompileFile
  End If
  
  'copy tempfile to CompileFile
  FileCopy strTempFile, CompileFile
  
  'delete temp file
  Kill strTempFile
  Err.Clear
Exit Sub
  
ErrHandler:
  'close file
  Close intFile
  'erase the temp file
  Kill CompileFile
  'return error condition
  On Error GoTo 0: Err.Raise vbObjectError + 582, strErrSource, LoadResString(582)
End Sub

Public Sub Export(ExportFile As String, ByVal FileType As Long, Optional ByVal ResetDirty As Boolean = True)

  'exports the list of words
  '  filetype = 0 means AGI WORDS.TOK file
  '  filetype = 1 means WinAGI word list file
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  On Error GoTo ErrHandler
  
  Select Case FileType
  Case 0  'compile agi WORDS.TOK file
    Compile ExportFile
    
  Case 1  'create WinAGI word list
    CompileWinAGI ExportFile
    
  End Select
  
  'if NOT in a game,
  If Not mInGame Then
    If ResetDirty Then
      'clear dirty flag
      mIsDirty = False
    End If
    'save filename
    mResFile = ExportFile
  End If
Exit Sub
  
ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 671, strErrSrc, Replace(LoadResString(671), ARG1, CStr(lngError) & ":" & strError)
End Sub

Friend Property Get ItemByIndex(ByVal Index As Long) As AGIWord
  'used by SetWords method to retrieve words one by one
  
  'NOTE: no error checking is done, because ONLY
  'SetWords uses this function, and it ensures
  'Index is valid
  
  ItemByIndex = mWordCol(Index)
End Property

Public Property Get Loaded() As Boolean

  Loaded = mLoaded
End Property
Private Function LoadSierraFile(LoadFile As String) As Boolean

  Dim intFile As Integer
  Dim bytHigh As Byte, bytLow As Byte
  Dim lngPos As Long
  Dim bytData() As Byte
  Dim strThisWord As String
  Dim strPrevWord As String
  Dim bytVal As Byte
  Dim lngGrpNum As Long
  Dim bytPrevWordCharCount As Byte
  
  On Error GoTo ErrHandler
  
  'set load flag
  mLoaded = True
  
  'open the file
  intFile = FreeFile()
  Open LoadFile For Binary As intFile
  'read in entire resource
  ReDim bytData(LOF(intFile) - 1)
  Get intFile, 1, bytData()
  Close intFile
  
  'start at beginning of words section
  'note that words.tok file uses MSLS format for two byte-word data)
  bytHigh = bytData(0)
  bytLow = bytData(1)
  
  lngPos = CLng(bytHigh) * 256 + CLng(bytLow)
  'for first word, there is no previous word
  strPrevWord = vbNullString
  'read character Count for first word
  bytPrevWordCharCount = bytData(lngPos)
  lngPos = lngPos + 1
  
  'continue loading words until end of resource is reached
  Do Until lngPos > UBound(bytData)
    'initialize word
    strThisWord = vbNullString
    'if number of characters to Count from previous word are longer than the previous word,
    If bytPrevWordCharCount > Len(strPrevWord) Then
      'should never be??
      bytPrevWordCharCount = Len(strPrevWord)
    End If
    'if some characters should be copied from previous word
    If bytPrevWordCharCount > 0 Then
      'copy 'em
      strThisWord = Left$(strPrevWord, bytPrevWordCharCount)
    End If
    'build rest of word
    Do
      bytVal = bytData(lngPos)
      'if bit7 is clear
      If bytVal < &H80 Then
        strThisWord = strThisWord & Chr$(bytVal Xor &H7F)
      End If
      lngPos = lngPos + 1
    'continue until last character (indicated by flag) or endorresource is reached
    Loop Until (bytVal >= &H80) Or lngPos > UBound(bytData)
    
    'if end of file is reached before &H80,
    If lngPos > UBound(bytData) Then
      'invalid words.tok file!
      GoTo ErrHandler
    End If
    'add last character (after stripping off flag)
    strThisWord = strThisWord & Chr$(&HFF Xor bytVal)
    'convert any upper-case characters to lower-case (just in case)
    strThisWord = LCase$(strThisWord)
    
    'get group number
    lngGrpNum = CLng(bytData(lngPos)) * 256 + CLng(bytData(lngPos + 1))
    'set pointer to next word
    lngPos = lngPos + 2
    
    'if this word different to previous,
    '(this ensures no duplicates are added)
    If strThisWord <> strPrevWord Then
      'if this word already exists,
      If WordExists(strThisWord) Then
        'delete the old one
        RemoveWord strThisWord
      End If
      'add word
      AddWord strThisWord, lngGrpNum

      'this word is now the previous word
      strPrevWord = strThisWord
    End If
    bytPrevWordCharCount = bytData(lngPos)
    lngPos = lngPos + 1
  Loop
  
  'if no words
  If mWordCol.Count = 0 Then
    'add default words
    AddWord "a", 0
    AddWord "anyword", 1
    AddWord "rol", 9999
  End If
  
  'return true
  LoadSierraFile = True
Exit Function

ErrHandler:
  'error
  Debug.Assert False
  Err.Clear
  Resume Next
  'ensure file closed
  Close intFile
  'return false
  mLoaded = False
End Function

Private Function LoadWinAGIFile(LoadFile As String) As Boolean

  Dim intFile As Integer
  Dim strInput As String
  Dim strWords() As String, strThisWord As String
  Dim lngGrpNum As Long
  Dim i As Long
  
  'attempt to open the WinAGI resource
  intFile = FreeFile()
  Open LoadFile For Input As intFile
  
  'first line should be loader
  Line Input #intFile, strInput
  Select Case strInput
  Case WINAGI_VERSION_1_2, WINAGI_VERSION_1_0, WINAGI_VERSION_BETA
    'ok
  Case Else
    'close file
    Close intFile
    'return false
    Exit Function
  End Select
  
  'set loaded flag
  mLoaded = True
  
  'enable inline error handling to check for invalid words/groups
  On Error Resume Next
  
  'begin loading word groups
  Do Until EOF(intFile)
    Line Input #intFile, strInput
    'check for end of input characters
    If strInput = Chr$(255) & Chr$(255) Then
      Exit Do
    End If
    'split input line
    strWords = Split(strInput, vbTab)
    'if there is at least one word and one number,
    If UBound(strWords) >= 1 Then
      'get group number
      lngGrpNum = CLng(strWords(0))
      'add this group
      AddGroup lngGrpNum
      If Err.Number <> 0 Then
        'invalid word group
        Close intFile
        'return false
        Exit Function
      End If
      'add words to this group
      For i = 1 To UBound(strWords)
        AddWord strWords(i), lngGrpNum
        
        If Err.Number <> 0 Then
          'some kind of error
          lngError = Err.Number
          strError = Err.Description
          strErrSrc = Err.Source
          
          mLoaded = False
          Close intFile
          
          'pass error along
          On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
          Exit Function
        End If
      Next i
    End If
  Loop
  
  'if any lines left
  If Not EOF(intFile) Then
    'get first remaining line
    Line Input #intFile, strThisWord
    'if there are additional lines, add them as a description
    Do Until EOF(intFile)
      Line Input #intFile, strInput
'''      strThisWord = strThisWord & vbNewLine & strInput
      strThisWord = strThisWord & vbCr & strInput
    Loop
    'get description
    mDescription = strThisWord
  End If
  
  'close file
  Close intFile
  
  'if no words
  If mWordCol.Count = 0 Then
    'add default words
    AddWord "a", 0
    AddWord "anyword", 1
    AddWord "rol", 9999
  End If
  
  'return true
  LoadWinAGIFile = True
Exit Function

ErrHandler:
  'error
  Err.Clear
  'ensure file closed
  Close intFile
  'return false
  mLoaded = False
End Function

Public Sub SetWords(NewWords As AGIWordList)
  'copies word list from NewWords to
  'this word list
  
  Dim i As Long, j As Long
  Dim lngGrpNum As Long, lngCount As Long
  Dim strWord As String
  Dim tmpGroup As AGIWordGroup, tmpWord As AGIWord
  
  'if source wordlist is not loaded
  If Not NewWords.Loaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'first, clear current list
  Clear
  
  'then add all groups
  lngCount = NewWords.GroupCount
  For i = 0 To lngCount - 1
    'create new temp group
    Set tmpGroup = New AGIWordGroup
    lngGrpNum = NewWords.Group(i).GroupNum
    tmpGroup.GroupNum = lngGrpNum
    'add group to end of collection
    mGroupCol.Add tmpGroup, "g" & CStr(lngGrpNum)
  Next i
  
  'then add all words
  lngCount = NewWords.WordCount
  '(use one based, because it's easier,
  'and no other function uses ItemByIndex
  For i = 1 To lngCount
    'get word
    strWord = NewWords.ItemByIndex(i).WordText
    lngGrpNum = NewWords.ItemByIndex(i).Group
    'create new word item
    'Set tmpWord = New AGIWord
    tmpWord.WordText = strWord
    tmpWord.Group = lngGrpNum
    
    'add word to wordcol
    mWordCol.Add tmpWord, strWord
    'add word to its groupcol
    'in order to access friend properties,
    'need to create a direct reference to the AGIwordgroup class
    Set tmpGroup = mGroupCol("g" & CStr(lngGrpNum))
    
    'add this word to its group
    tmpGroup.AddWordToGroup strWord
    Set tmpGroup = Nothing
  Next i

  'copy description
  mDescription = NewWords.Description
  
  'set dirty flag
  NewWords.SetFlags mIsDirty, mWriteProps
  
  'copy filename
  mResFile = NewWords.ResFile
  
  'set loaded flag
  mLoaded = True
End Sub

Friend Sub SetFlags(ByRef IsDirty As Boolean, ByRef WriteProps As Boolean)
  
  IsDirty = mIsDirty
  WriteProps = mWriteProps
End Sub


Public Property Get Description() As String

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  Description = mDescription
End Property

Friend Property Get InGame() As Boolean

  'only used by setword method
  InGame = mInGame
End Property


Public Property Let Description(ByVal NewDescription As String)

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  'limit description to 1K
  NewDescription = Left$(NewDescription, 1024)
  
  'if changing
  If NewDescription <> mDescription Then
    mDescription = NewDescription
  
    'if in a game
    If mInGame Then
      WriteGameSetting "WORDS.TOK", "Description", mDescription
    End If
  End If
End Property


Friend Sub Init(Optional ByVal Loaded As Boolean = False)

  'this function is only called for a vocab word list
  'that is part of a game
  'it sets the ingame flag
  mInGame = True
  
  'if loaded property is passed, set loaded flag as well
  mLoaded = Loaded
  
  'it also sets the default name to 'WORDS.TOK'
  mResFile = agGameDir & "WORDS.TOK"
End Sub


Public Property Get ResFile() As String
    
  ResFile = mResFile
End Property

Public Property Let ResFile(NewSourceFile As String)
  
  'resfile cannot be changed if resource is part of a game
  
  'if in a game
  If mInGame Then
    'error- resfile is readonly for ingame resources
    On Error GoTo 0: Err.Raise vbObjectError + 680, strErrSource, LoadResString(680)
  Else
    mResFile = NewSourceFile
  End If
End Property

Public Sub Clear()
  'clears the word group, and sets up a blank list
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'reset group and word collections
  Set mGroupCol = Nothing
  Set mWordCol = Nothing
  Set mGroupCol = New Collection
  Set mWordCol = New Collection
  mDescription = vbNullString
  mIsDirty = True
End Sub



Friend Property Let InGame(NewValue As Boolean)

  mInGame = NewValue
End Property


Public Property Get GroupN(ByVal GroupNumber As Long) As AGIWordGroup
  'returns a group by its group number
  
  On Error Resume Next

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  'return group with this number
  Set GroupN = mGroupCol("g" & CStr(GroupNumber))

  If Err.Number <> 0 Or GroupN Is Nothing Then
    'invalid index or key
    On Error GoTo 0: Err.Raise vbObjectError + 575, strErrSource, LoadResString(575)
    Exit Property
  End If

End Property


Private Function GroupExists(ByVal GroupNumber As Long) As Boolean
  'if this group exists, it returns true
  On Error Resume Next

  If mGroupCol("g" & CStr(GroupNumber)).GroupNum = 0 Then
  End If

  GroupExists = (Err.Number = 0)
End Function


Public Property Let IsDirty(ByVal NewState As Boolean)
    
    mIsDirty = NewState
End Property
Public Property Get Item(vKeyIndex As Variant) As AGIWord
Attribute Item.VB_UserMemId = 0
  'access is by word string or index number
  '(in VB.NET, this will be an overload function)
  
  Dim Index As Long
  Dim Key As String
  
  On Error Resume Next
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  'only allow string or integer number
  Select Case VarType(vKeyIndex)
  Case vbByte, vbInteger, vbLong
    'retrieve via index
    
    'validate index
    If vKeyIndex < 0 Or vKeyIndex > mWordCol.Count - 1 Then
      On Error GoTo 0: Err.Raise 9, strErrSource, "Subscript out of range"
      Exit Sub
    End If
    
    Item = mWordCol(vKeyIndex + 1)
  Case vbString
    'retrieve via string key
    Item = mWordCol(vKeyIndex)
    
    If Err.Number <> 0 Then
      'word not found
      On Error GoTo 0: Err.Raise vbObjectError + 586, strErrSource, LoadResString(586)
    End If
    
  Case Else
    'not allowed
    On Error GoTo 0: Err.Raise vbObjectError + 574, strErrSource, LoadResString(574)
    Exit Property
  End Select
End Property


Public Sub Save(Optional SaveFile As String, Optional ByVal FileType As Long)
  
  'saves wordlist
  ' filetype = 0 means AGI WORDS.TOK file
  ' filetype = 1 means WinAGI word list file
  'for ingame resources, SaveFile and filetype are ignored
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  On Error GoTo ErrHandler
  
  'if in a game,
  If mInGame Then
    'compile the file
    Compile mResFile
    
  Else
    If LenB(SaveFile) = 0 Then
      SaveFile = mResFile
    End If
      
    Select Case FileType
    Case 0  'compile agi WORDS.TOK file
      Compile SaveFile
      
    Case 1  'create WinAGI word list
      CompileWinAGI SaveFile
      
    End Select
  
    'save filename
    mResFile = SaveFile
  End If
  
  'mark as clean
  mIsDirty = False
  
Exit Sub

ErrHandler:
  'return error condition
  'pass error along
  lngError = Err.Number
  strError = Err.Description
  strErrSrc = Err.Source
  
  On Error GoTo 0: Err.Raise lngError, strErrSrc, strError
End Sub
Public Sub Unload()

  'unloads the resource; same as clear, except file marked as not dirty
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  Clear
  mLoaded = False
  
  mWriteProps = False
  mIsDirty = False
End Sub

Public Property Get WordCount() As Long
    
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  WordCount = mWordCol.Count
End Property



Public Property Get NewEnum() As IUnknown
Attribute NewEnum.VB_UserMemId = -4
    
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  Set NewEnum = mWordCol.[_NewEnum]
End Property



Friend Function WordExists(aWord As String) As Boolean
  'if this word exists, it returns true
  On Error Resume Next
  
  If LenB(mWordCol(aWord).WordText) = 0 Then
  End If
  
  WordExists = (Err.Number = 0)
  Err.Clear
End Function
Private Sub Class_Initialize()
  
  Set mGroupCol = New Collection
  Set mWordCol = New Collection
  
  strErrSource = "WINAGI.AGIWordList"
End Sub



Public Sub AddGroup(ByVal GroupNumber As Long)

  Dim i As Integer
  Dim tmpGroup As AGIWordGroup

  On Error Resume Next

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'if group number is invalid
  If GroupNumber < 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 575, strErrSource, LoadResString(575)

    Exit Sub
  End If
  If GroupNumber > MAX_GROUP_NUM Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 575, strErrSource, LoadResString(575)
    Exit Sub
  End If

  'see if group already exists
  If mGroupCol("g" & CStr(GroupNumber)).GroupNum Then
  End If
  'if no error, means group already exists
  If Err.Number = 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 576, strErrSource, LoadResString(576)
    Exit Sub
  End If

  Err.Clear
  
  'if already at max number of groups
  If mGroupCol.Count = MAX_WORD_GROUPS Then
    'error
    Exit Sub
    On Error GoTo 0: Err.Raise vbObjectError + 577, strErrSource, LoadResString(577)
  End If

  'any other errors are trapped
  On Error GoTo ErrHandler

  Set tmpGroup = New AGIWordGroup
  tmpGroup.GroupNum = GroupNumber

  'if this is the first group,
  If mGroupCol.Count = 0 Then
    'add it, using the word itself as the key
    mGroupCol.Add tmpGroup, "g" & CStr(GroupNumber)
  Else
    'find correct position for this wordgroup
    If GroupNumber < mGroupCol(1).GroupNum Then
      'this group should be first
      mGroupCol.Add tmpGroup, "g" & CStr(GroupNumber), 1
    Else
      'step through rest of objects
      For i = mGroupCol.Count To 1 Step -1
        'if new groupno is greater than current groupno
        If GroupNumber > mGroupCol(i).GroupNum Then
          'this is where groupno goes
          Exit For
        End If
      Next i
      'add it, using the word itself as the key
      mGroupCol.Add tmpGroup, "g" & CStr(GroupNumber), , i
    End If
  End If
  mIsDirty = True
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 578, strErrSrc, Replace(LoadResString(578), ARG1, CStr(lngError) & ":" & strError)
End Sub


Sub RemoveWord(aWord As String)
  'deletes aWord
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  On Error GoTo ErrHandler
  
  'delete this word from its assigned group
  GroupN(mWordCol(aWord).Group).DeleteWordFromGroup aWord
  'remove it from the word collection
  mWordCol.Remove aWord
  'set dirty flag
  mIsDirty = True
Exit Sub

ErrHandler:
  'word not found
  On Error GoTo 0: Err.Raise vbObjectError + 584, strErrSource, LoadResString(584)
End Sub


Public Property Get GroupCount() As Long

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  GroupCount = mGroupCol.Count
End Property


Public Sub RemoveGroup(ByVal GroupNumber As Long)

  Dim lngGroup As Integer
  Dim i As Integer

  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  On Error GoTo ErrHandler

  'if group number is invalid
  If GroupNumber < 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 575, strErrSource, LoadResString(575)
    Exit Sub
  End If

  If GroupNumber > MAX_GROUP_NUM Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 575, strErrSource, LoadResString(575)
    Exit Sub
  End If

  'step through all words in group
  For i = 0 To mGroupCol("g" & CStr(GroupNumber)).WordCount - 1
    'delete this word from main list
    mWordCol.Remove mGroupCol("g" & CStr(GroupNumber)).Word(i)
  Next i

  'delete the group
  mGroupCol.Remove "g" & CStr(GroupNumber)

  'set dirty flag
  mIsDirty = True
Exit Sub

ErrHandler:
    'group didn't exist
    On Error GoTo 0: Err.Raise vbObjectError + 583, strErrSource, LoadResString(583)
    Exit Sub
End Sub


Private Sub Compile(CompileFile As String)
  'compiles the word list into a Sierra WORDS.TOK file
  
  Dim intFile As Integer
  Dim i As Integer
  Dim lngGroup As Long, lngWord As Long
  Dim CurByte As Byte
  Dim strWord As String
  Dim strPrevWord As String
  Dim intPrevWordCharCount As Byte
  Dim blnWordsMatch As Boolean
  Dim strFirstLetter As String
  Dim intLetterIndex(0 To 26) As Integer
  Dim strTempFile As String
  
  On Error Resume Next
  
  'now use errorhandler
  On Error GoTo ErrHandler
  
  'if no filename passed,
  If LenB(CompileFile) = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 616, strErrSource, LoadResString(616)
    Exit Sub
  End If
  
  'if not dirty AND CompileFile=resfile
  If Not mIsDirty And UCase$(CompileFile) = UCase$(mResFile) Then
    Exit Sub
  End If
  
  'if there are no word groups to add
  If mGroupCol.Count = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 672, strErrSource, Replace(LoadResString(672), ARG1, "no word groups to add")
    Exit Sub
  End If
  
  'if there are no words,
  If mWordCol.Count = 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 672, strErrSource, Replace(LoadResString(672), ARG1, "no words to add")
    Exit Sub
  End If
  
  'create temp file
  strTempFile = TempFileName()
  
  'open the file
  intFile = FreeFile()
  Open strTempFile For Binary As intFile
  
  'write letter index
  For i = 0 To 51
    Put intFile, , CByte(0)
  Next i

  'index pointer starts with 'a'
  strFirstLetter = "a"
  intLetterIndex(1) = 52

  'now step through all words in list
  For lngWord = 1 To mWordCol.Count
    'get next word
    strWord = mWordCol(lngWord).WordText
    
    'if first letter is not current first letter, AND it is 'b' thru 'z'
    '(this ensures non letter words (such as numbers) are included in the 'a' section)
    If (Left$(strWord, 1) <> strFirstLetter) And (AscW(Left$(strWord, 1)) >= 97) And (AscW(Left$(strWord, 1)) <= 122) Then
      'reset index pointer
      strFirstLetter = Left$(strWord, 1)
      intLetterIndex(AscW(strFirstLetter) - 96) = Seek(intFile) - 1
      'need to subtract 1 because data index should be zero based
      'but seek is one based
    End If
    
    'calculate number of characters that are common to previous word
    intPrevWordCharCount = 0
    i = 1
    blnWordsMatch = True

    Do
      'if not at end of word, AND current position is not longer than previous word,
      If (i <= Len(strWord)) And (Len(strPrevWord) >= i) Then
        'if characters at this position match,
        If AscW(Mid$(strPrevWord, i)) = AscW(Mid$(strWord, i)) Then
          'increment Count
          intPrevWordCharCount = intPrevWordCharCount + 1
        Else
          blnWordsMatch = False
        End If
      Else
        blnWordsMatch = False
      End If
      i = i + 1
    Loop While blnWordsMatch

    'write number of characters from previous word
    Put intFile, , intPrevWordCharCount
    
    'set previous word Value to this word
    strPrevWord = strWord
    'strip off characters that are same as previous word
    strWord = Right$(strWord, Len(strWord) - intPrevWordCharCount)
  
    'if there are two or more characters to write,
    If Len(strWord) > 1 Then
      'write all but last character
      For i = 1 To Len(strWord) - 1
        'encrypt character before writing it
        CurByte = &H7F Xor Asc(Mid$(strWord, i))
        Put intFile, , CurByte
      Next i
    End If

    'encrypt character, and set flag &H80
    CurByte = &H80 + (&H7F Xor Asc(Mid$(strWord, Len(strWord))))
    Put intFile, , CurByte
    
    'write group number (stored as 2-byte word; high byte first)
    CurByte = mWordCol(lngWord).Group \ 256
    Put intFile, , CurByte
    CurByte = mWordCol(lngWord).Group Mod 256
    Put intFile, , CurByte
  Next lngWord

  'add null character to end of file
  Put intFile, , CByte(0)

  'reset file pointer to start
  Seek intFile, 1
  
  'write index values for all 26 letters
  For i = 1 To 26
    '(two byte word, high byte first
    CurByte = intLetterIndex(i) \ 256
    Put intFile, , CurByte
    CurByte = intLetterIndex(i) Mod 256
    Put intFile, , CurByte
  Next i
  
  'close file,
  Close intFile
  
  'if CompileFile already exists
  If FileExists(CompileFile) Then
    'delete it
    Kill CompileFile
  End If
  
  'copy tempfile to CompileFile
  FileCopy strTempFile, CompileFile
  'delete temp file
  Kill strTempFile
  Err.Clear
Exit Sub

ErrHandler:
  Dim lngErr As Long
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  
  'delete temporary file
  Kill strTempFile
  'raise the error
  On Error GoTo 0: Err.Raise vbObjectError + 672, strErrSrc, Replace(LoadResString(672), ARG1, CStr(lngError) & ":" & strError)
End Sub
Public Sub Load(Optional LoadFile As String)
  
  'this function loads words for the game
  'if loading from a Sierra game, it extracts words
  'from the WORDS.TOK file;
  'if not in a game, LoadFile must be specified
  
  Dim blnSuccess As Boolean
  
  On Error GoTo ErrHandler
  
  'if already loaded,
  If mLoaded Then
    'error- resource already loaded
    On Error GoTo 0: Err.Raise vbObjectError + 511, strErrSource, LoadResString(511)
    Exit Sub
  End If
  
  If mInGame Then
    'use default filename
    LoadFile = agGameDir & "WORDS.TOK"
    'attempt to load
    If Not LoadSierraFile(LoadFile) Then
      On Error GoTo 0
      'error
      On Error GoTo 0: Err.Raise vbObjectError + 529, strErrSource, LoadResString(529)
    End If
    
    'get description, if there is one
    Debug.Assert Not (agGameProps Is Nothing)
    ReadSettingString agGameProps, "WORDS.TOK", "Description", vbNullString
  Else
    'if not in a game, must have a valid filename
    
    'if optional filename not passed,
    If LenB(LoadFile) = 0 Then
      'not in game; return error
      On Error GoTo 0
      On Error GoTo 0: Err.Raise vbObjectError + 599, strErrSource, LoadResString(599)
      Exit Sub
    End If
    
    'verify file exists
    If Not FileExists(LoadFile) Then
      On Error GoTo 0
      'error
      On Error GoTo 0: Err.Raise vbObjectError + 524, strErrSource, Replace(LoadResString(524), ARG1, LoadFile)
      Exit Sub
    End If
    
    'if extension is .agw then
    If LCase$(Right$(LoadFile, 4)) = ".agw" Then
      'assume winagi format
      If Not LoadWinAGIFile(LoadFile) Then
        'try sierra format
        If Not LoadSierraFile(LoadFile) Then
          On Error GoTo 0
          'error
          On Error GoTo 0: Err.Raise vbObjectError + 529, strErrSource, LoadResString(529)
          Exit Sub
        End If
      End If
    Else
      'assume sierra format
      If Not LoadSierraFile(LoadFile) Then
        'try winagi format
        If Not LoadWinAGIFile(LoadFile) Then
          On Error GoTo 0
          'error
          On Error GoTo 0: Err.Raise vbObjectError + 529, strErrSource, LoadResString(529)
          Exit Sub
        End If
      End If
    End If
  
    'save filename
    mResFile = LoadFile
  End If
  
  'reset dirty flag
  mIsDirty = False
Exit Sub

ErrHandler:
  Debug.Assert False
  Resume Next
End Sub
Public Property Get Group(ByVal Index As Integer) As AGIWordGroup
  
  'returns a group by its index (NOT the same as group number)
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Property
  End If
  
  On Error Resume Next
  
  'if invalid index
  If Index < 0 Or Index > mGroupCol.Count - 1 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 588, strErrSource, LoadResString(588)
    Exit Property
  End If

  'access the group by its index
  Set Group = mGroupCol(Index + 1)

End Property


Public Property Get IsDirty() As Boolean
  
  'if resource is dirty, or (prop values need writing AND in game)
  IsDirty = (mIsDirty Or (mWriteProps And mInGame))
End Property


Public Sub AddWord(ByVal WordText As String, ByVal Group As Long)

  Dim i As Integer
  Dim NewWord As AGIWord
  Dim tmpGroup As AGIWordGroup
  
  On Error Resume Next
  
  'if not loaded
  If Not mLoaded Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 563, strErrSource, LoadResString(563)
    Exit Sub
  End If
  
  'convert input to lowercase
  WordText = LCase$(WordText)
  
  'check to see if word is already in collection,
  i = mWordCol(WordText).Group
  If Err.Number = 0 Then
    On Error GoTo 0: Err.Raise vbObjectError + 579, strErrSource, LoadResString(579)
    Exit Sub
  End If
  
  Err.Clear
  
  'if group number is invalid
  If Group < 0 Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 581, strErrSource, LoadResString(581)
    Exit Sub
  End If
  
  'if number is too large
  If Group > MAX_GROUP_NUM Then
    'error
    On Error GoTo 0: Err.Raise vbObjectError + 581, strErrSource, LoadResString(581)
    Exit Sub
  End If
  
  'if this group does not yet exist,
  i = mGroupCol("g" & CStr(Group)).GroupNum
  If Err.Number <> 0 Then
    'add this group
    AddGroup Group
  End If
  
  'any other error gets trapped
  On Error GoTo ErrHandler
  
  'in order to access friend properties,
  'need to create a direct reference to the AGIwordgroup class
  Set tmpGroup = mGroupCol("g" & CStr(Group))
  
  'add this word to its group
  tmpGroup.AddWordToGroup WordText
  Set tmpGroup = Nothing
  
  'now add it to the main word collection
  NewWord.WordText = WordText
  NewWord.Group = Group
  
  'if this is the first word,
  If mWordCol.Count = 0 Then
    'add it, using the word itself as the key
    mWordCol.Add NewWord, WordText
  Else
    'find correct position for this word
    If NewWord.WordText < mWordCol(1).WordText Then
      'this word should be first
      mWordCol.Add NewWord, WordText, 1
    Else
      'step through rest of objects
      For i = mWordCol.Count To 1 Step -1
        'if new word is greater than current word
        If NewWord.WordText > mWordCol(i).WordText Then
          'this is where word goes
          Exit For
        End If
      Next i
      'add it, using the word itself as the key
      mWordCol.Add NewWord, NewWord.WordText, , i
    End If
  End If
  
  'set dirty flag
  mIsDirty = True
Exit Sub

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 673, strErrSrc, Replace(LoadResString(673), ARG1, CStr(lngError) & ":" & strError)
End Sub



Private Sub Class_Terminate()
  'destroys collections when this class is terminated
  Set mWordCol = Nothing
  Set mGroupCol = Nothing
End Sub

