Attribute VB_Name = "ResourceFunctions"
Attribute VB_Ext_KEY = "RVB_UniqueId" ,"3E24FD1D036A"
Option Explicit

'constants used in extracting compressed resources
Const TABLE_SIZE = 18041
Const START_BITS = 9

'variables used in extracting compressed resources
Private lngMaxCode As Long
Private intPrefix() As Integer
Private bytAppend() As Byte
Private lngBitsInBuffer As Long
Private lngBitBuffer As Long
Private lngOriginalSize As Long
  
'sound subclassing variables, constants, declarations
Public PrevSndWndProc As Long
Public blnPlaying As Boolean
Public PlaySndResNum As Byte
Public Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" (ByVal lpPrevWndFunc As Long, ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Public Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long
Public Declare Function mciGetErrorString Lib "winmm.dll" Alias "mciGetErrorStringA" (ByVal dwError As Long, ByVal lpstrBuffer As String, ByVal uLength As Long) As Long
Public Declare Function mciSendString Lib "winmm.dll" Alias "mciSendStringA" (ByVal lpstrCommand As String, ByVal lpstrReturnString As String, ByVal uReturnLength As Long, ByVal hwndCallback As Long) As Long
Public Const GWL_WNDPROC = (-4)
Public Const MM_MCINOTIFY = &H3B9
Public Const MCI_NOTIFY_SUCCESSFUL = &H1

'others
Private lngPos As Long
Private mMIDIData() As Byte
Public Function CompressCelData(Cel As AGICel, blnMirror As Boolean) As Byte()
  'this method compresses cel data
  'into run-length-encoded data that
  'can be written to an AGI View resource
  
  'blnMirror is used to ensure mirrored cels include enough room
  'for the flipped cel
  
  Dim bytTempRLE() As Byte
  Dim mHeight As Byte, mWidth As Byte
  Dim mCelData() As AGIColors, mTransColor As Byte
  Dim lngByteCount As Long
  Dim bytChunkColor As AGIColors, bytChunkLen As Byte
  Dim bytNextColor As AGIColors
  Dim bytOut As Byte
  Dim blnFirstChunk As Boolean
  Dim lngMirrorCount As Long
  Dim blnErr As Boolean
  
  On Error GoTo ErrHandler
  
  Dim i As Integer, j As Integer
  
  'copy cel data locally
  mHeight = Cel.Height
  mWidth = Cel.Width
  mTransColor = Cel.TransColor
  mCelData = Cel.AllCelData
  
  
  'assume one byte per pixel to start with
  '(include one byte for ending zero)
  ReDim bytTempRLE(CLng(mHeight) * (CLng(mWidth) + 1))
  'step through each row
  For j = 0 To mHeight - 1
    'get first pixel color
    bytChunkColor = mCelData(0, j)
    bytChunkLen = 1
    blnFirstChunk = True
    
    'step through rest of pixels in this row
    For i = 1 To mWidth - 1
      'get next pixel color
      bytNextColor = mCelData(i, j)
      'if different from current chunk
      If bytNextColor <> bytChunkColor Or bytChunkLen = &HF Then
        'write chunk
        bytTempRLE(lngByteCount) = bytChunkColor * &H10 + bytChunkLen
        'increment Count
        lngByteCount = lngByteCount + 1
          
        'if this is NOT first chunk or NOT transparent)
        If Not blnFirstChunk Or (bytChunkColor <> mTransColor) Then
          'increment lngMirorCount for any chunks
          'after the first, and also for the first
          'if it is NOT transparent color
          lngMirrorCount = lngMirrorCount + 1
        End If
        
        blnFirstChunk = False
        'set chunk to new color
        bytChunkColor = bytNextColor
        'reset length
        bytChunkLen = 1
      Else
        'increment chunk length
        bytChunkLen = bytChunkLen + 1
      End If
    Next i
    'if last chunk is NOT transparent
    If bytChunkColor <> mTransColor Then
      'add last chunk
      bytTempRLE(lngByteCount) = bytChunkColor * &H10 + bytChunkLen
      'increment Count
      lngByteCount = lngByteCount + 1
    End If
    'always Count last chunk for mirror
    lngMirrorCount = lngMirrorCount + 1
    'add zero to indicate end of row
    bytTempRLE(lngByteCount) = 0
    lngByteCount = lngByteCount + 1
    lngMirrorCount = lngMirrorCount + 1
  Next j
  
  'if mirroring
  If blnMirror Then
    'add zeros to make room
    Do Until lngByteCount >= lngMirrorCount
      'add a zero
      bytTempRLE(lngByteCount) = 0
      lngByteCount = lngByteCount + 1
    Loop
  End If
  
  'reset size of array
  ReDim Preserve bytTempRLE(lngByteCount - 1)
  
  'return the compressed data
  CompressCelData = bytTempRLE
Exit Function

ErrHandler:
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  'if error was due to insufficient space (subscript out of range)
  'redimension the array and try again
  If Err.Number = 9 Then
    ReDim Preserve bytTempRLE(UBound(bytTempRLE()) + 20)
    Resume
  End If
  
  Debug.Assert False
  On Error GoTo 0: Err.Raise vbObjectError + 654, strErrSrc, Replace(LoadResString(654), ARG1, CStr(lngError) & ":" & strError)
End Function




Public Sub ExpandV3ResData(ByRef bytOriginalData() As Byte, lngExpandedSize As Long)
  
  Dim intPosIn As Integer, intPosOut As Integer
  Dim intNextCode As Integer, intNewCode As Integer, intOldCode As Integer
  Dim intCount  As Integer, i  As Integer
  Dim strDat As String
  Dim strChar As String * 1
  Dim intCodeSize As Integer
  Dim bytTempData() As Byte
  
  On Error GoTo ErrHandler
  
  'initialize variables
  ReDim intPrefix(TABLE_SIZE) 'remember to correct index by 257
  ReDim bytAppend(TABLE_SIZE) 'remember to correct index by 257

  'set temporary data field
  ReDim bytTempData(lngExpandedSize - 1)
  'original size is determined by array bounds
  lngOriginalSize = UBound(bytOriginalData) + 1
  
  'reset variables used in expansion
  intPosIn = 0
  intPosOut = 0

  lngBitBuffer = 0
  lngBitsInBuffer = 0
  
  'Set initial Value for code size
  intCodeSize = NewCodeSize(START_BITS)
  'first code is 257
  intNextCode = 257
  'this seems wrong! first code should be 258, right?
  'isn't 257 the 'end' code?
  
  'Read in the first code.
  intOldCode = InputCode(bytOriginalData, intCodeSize, intPosIn)

  '!!!!why is this set???
  strChar = Chr$(0)
  
  'first code for  SIERRA resouces should always be 256
  'if first code is NOT 256
  If intOldCode <> 256 Then
    'error!
    GoTo ErrHandler
  End If

  'now begin decompressing actual data
  intNewCode = InputCode(bytOriginalData, intCodeSize, intPosIn)
  
  'continue extracting data, until all bytes are read (or end code is reached)
  Do While (intPosIn <= lngOriginalSize) And (intNewCode <> &H101)
    'if new code is &H100,
    If (intNewCode = &H100) Then
      'Restart LZW process (should tables be flushed?)
      intNextCode = 258
      intCodeSize = NewCodeSize(START_BITS)
      'Read in the first code.
      intOldCode = InputCode(bytOriginalData, intCodeSize, intPosIn)
      'the character Value is same as code for beginning
      strChar = Chr$(intOldCode)
      'write out the first character
      bytTempData(intPosOut) = intOldCode
      intPosOut = intPosOut + 1
      intCount = intCount + 1
      'now get next code
      intNewCode = InputCode(bytOriginalData, intCodeSize, intPosIn)
    Else
      ' This code checks for the special STRING+character+STRING+character+STRING
      ' case which generates an undefined code.  It handles it by decoding
      ' the last code, and adding a single Charactor to the end of the decode string.
      ' (new_code will ONLY return a next_code Value if the condition exists;
      ' it should otherwise return a known code, or a ascii Value)
      If (intNewCode >= intNextCode) Then
        'decode the string using old code
        strDat = DecodeString(intOldCode)
        'append the character code
        strDat = strDat & strChar
      Else
        'decode the string using new code
        strDat = DecodeString(intNewCode)
      End If
      'retreive the character Value
      strChar = Left$(strDat, 1)
      'now send out decoded data (it's backwards in the string, so
      'start at end and work back to beginning)
      For i = 1 To Len(strDat)
        bytTempData(intPosOut) = Asc(Mid$(strDat, i))
        intPosOut = intPosOut + 1
      Next i
      'if no more room in the current bit-code table,
      If (intNextCode > lngMaxCode) Then
        'get new code size (in number of bits per code)
        intCodeSize = NewCodeSize(intCodeSize + 1)
        intCount = intCount + 1
      End If
      
      'store code in prefix table
      intPrefix(intNextCode - 257) = intOldCode
      'store append character in table
      bytAppend(intNextCode - 257) = Asc(strChar)
      'increment next code pointer
      intNextCode = intNextCode + 1
      intOldCode = intNewCode
      'clear the decoded data string
      strDat = vbNullString
      'get the next code
      intNewCode = InputCode(bytOriginalData, intCodeSize, intPosIn)
    End If
  Loop
  
  'copy array
  ReDim bytOriginalData(lngExpandedSize - 1)
  bytOriginalData() = bytTempData()
  'free arrays
  ReDim intPrefix(0)
  ReDim bytAppend(0)
Exit Sub

ErrHandler:
  'all errors invalidate resource
  ReDim intPrefix(0)
  ReDim bytAppend(0)
  On Error GoTo 0: Err.Raise 559, Replace(LoadResString(559), ARG1, CStr(Err.Number))
End Sub


Public Function LoadResources() As Boolean
  'returns true if resources loaded with warnings
  'returns false if one or more errors occur during load
  
  Dim bytResType As Byte, bytResNum As Byte
  Dim strDirFile As String, intFile As Integer
  Dim bytBuffer() As Byte
  Dim byte1 As Byte, byte2 As Byte, byte3 As Byte
  Dim lngDirOffset As Long  ' offset of this resource's directory in Dir file (for v3)
  Dim lngDirSize As Long
  Dim intResCount As Integer
  Dim i As Integer
  Dim bytVol As Byte, lngLoc As Long
  Dim strVolFile As String, strID As String
  Dim strResID As String
  Dim blnWarnings As Boolean
  Dim intErrFile As Integer
  
  On Error Resume Next
  
  agGameEvents.RaiseEvent_LoadStatus lsResources, rtNone, 0, vbNullString
  
  'if version 3
  If agIsVersion3 Then
    'get ID
    strID = agGameID
    'get combined dir Volume
    strDirFile = agGameDir & agGameID & "Dir"
    'verify it exists
    If Not FileExists(strDirFile) Then
      On Error GoTo 0: Err.Raise vbObjectError + 524, "ResourceFunctions.LoadResources", Replace(LoadResString(524), ARG1, strDirFile)
      Exit Function
    End If
    
    'open the file, load it into buffer, and close it
    intFile = FreeFile()
    Open strDirFile For Binary As intFile
    ReDim bytBuffer(LOF(intFile))
    Get intFile, 1, bytBuffer()
    Close intFile
    
    'check for error during file manipulation
    If Err.Number <> 0 Then
      On Error GoTo 0: Err.Raise vbObject + 502, "ResourceFunctions.LoadResources", Replace(Replace(LoadResString(502), ARG1, CStr(Err.Number)), "%2", strDirFile)
      Exit Function
    End If
    'check size of file
    If UBound(bytBuffer) > 3080 Then
      On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
      Exit Function
    End If
    
    'if not enough bytes to hold at least the 4 dir pointers & 1 resource
    If UBound(bytBuffer) < 11 Then
      On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
      Exit Function
    End If
    
    'set flag if file is already expanded
    agDirExpanded = (UBound(bytBuffer) = 3079)
  Else
    'no id
    strID = vbNullString
  End If
  
  'step through all four resource types
  For bytResType = 0 To 3
    'use inline error handling
    On Error Resume Next
    'if version 3
    If agIsVersion3 Then
      'calculate offset and size of each dir component
      Select Case bytResType
      Case rtLogic
        lngDirOffset = bytBuffer(0) + bytBuffer(1) * 256
        lngDirSize = bytBuffer(2) + bytBuffer(3) * 256 - lngDirOffset
      Case rtPicture
        lngDirOffset = bytBuffer(2) + bytBuffer(3) * 256
        lngDirSize = bytBuffer(4) + bytBuffer(5) * 256 - lngDirOffset
      Case rtView
        lngDirOffset = bytBuffer(4) + bytBuffer(5) * 256
        lngDirSize = bytBuffer(6) + bytBuffer(7) * 256 - lngDirOffset
      Case rtSound
        lngDirOffset = bytBuffer(6) + bytBuffer(7) * 256
        lngDirSize = UBound(bytBuffer) - lngDirOffset
      End Select
    Else
      'no offset for version 2
      lngDirOffset = 0
      'get name of resource dir file
      strDirFile = agGameDir & ResTypeAbbrv(bytResType) & "DIR"
      'verify it exists
      If Not FileExists(strDirFile) Then
        On Error GoTo 0: Err.Raise vbObjectError + 524, "ResourceFunctions.LoadResources", Replace(LoadResString(524), ARG1, strDirFile)
        Exit Function
      End If
      
      'open the file, and load into buffer, and close it
      intFile = FreeFile()
      Open strDirFile For Binary As intFile
      ReDim bytBuffer(LOF(intFile))
      Get intFile, 1, bytBuffer()
      Close intFile
      
      'check for error during file manipulation
      If Err.Number <> 0 Then
        On Error GoTo 0: Err.Raise vbObject + 502, "ResourceFunctions.LoadResources", Replace(Replace(LoadResString(502), ARG1, CStr(Err.Number)), "%2", strDirFile)
        Exit Function
      End If
      
      'if file is too big
      If UBound(bytBuffer) > 768 Then
        On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
        Exit Function
      End If
      
      'get size
      lngDirSize = UBound(bytBuffer)
    End If
    
    'if invalid dir information, return false
    If (lngDirOffset < 0) Or (lngDirSize < 0) Then
      On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
      Exit Function
    End If
    
    'if at least one resource,
    If lngDirSize >= 3 Then
      If lngDirOffset + lngDirSize > UBound(bytBuffer) Then
        On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
        Exit Function
      End If
    End If
    
    'max size of useable directory is 768 (256*3)
    If lngDirSize > 768 Then
      On Error GoTo 0: Err.Raise vbObjectError + 542, "ResourceFunctions.LoadResources", Replace(LoadResString(542), ARG1, strDirFile)
      Exit Function
    Else
      intResCount = lngDirSize \ 3
    End If
    
    'if this resource type has entries,
    If intResCount > 0 Then
      'use error handler to check for bad resources
      On Error GoTo ErrHandler
      For i = 0 To intResCount - 1
      
        agGameEvents.RaiseEvent_LoadStatus lsResources, bytResType, i, vbNullString
        
        bytResNum = i
        'get location data for this resource
        byte1 = bytBuffer(lngDirOffset + bytResNum * 3)
        byte2 = bytBuffer(lngDirOffset + bytResNum * 3 + 1)
        byte3 = bytBuffer(lngDirOffset + bytResNum * 3 + 2)
        'ignore any &HFFFFFF sequences,
'''        If (Not ((byte1 = &HFF) And (byte2 = &HFF) And (byte3 = &HFF))) Then
        If byte1 <> &HFF Then
          'extract volume and location
          bytVol = byte1 \ 16
          strVolFile = agGameDir & strID & "VOL." & CStr(bytVol)
          lngLoc = (CLng(byte1) Mod 16) * &H10000 + CLng(byte2) * &H100 + CLng(byte3)
          strResID = agResTypeName(bytResType) & CStr(bytResNum)
          'add a resource of this res type
          Select Case bytResType
          Case rtLogic  'logic
            agLogs.LoadLogic bytResNum, bytVol, lngLoc
            'make sure it was added before attempting to set property state
            If agLogs.Exists(bytResNum) Then
              'when new resources are added, status is set to dirty; for initial load,
              'need to reset them to false
              agLogs(bytResNum).WritePropState = False
              agLogs(bytResNum).IsDirty = False
            End If
            
          Case rtPicture  'picture
            agPics.LoadPicture bytResNum, bytVol, lngLoc
            'make sure it was added before attempting to set property state
            If agPics.Exists(bytResNum) Then
              'when new resources are added, status is set to dirty; for initial load,
              'need to reset them to false
              agPics(bytResNum).WritePropState = False
              agPics(bytResNum).IsDirty = False
            End If
            
          Case rtSound  'sound
            agSnds.LoadSound bytResNum, bytVol, lngLoc
            'make sure it was added before attempting to set property state
            If agSnds.Exists(bytResNum) Then
              'when new resources are added, status is set to dirty; for initial load,
              'need to reset them to false
              agSnds(bytResNum).WritePropState = False
              agSnds(bytResNum).IsDirty = False
            End If
            
          Case rtView  'view
            agViews.LoadView bytResNum, bytVol, lngLoc
            'make sure it was added before attempting to set property state
            If agViews.Exists(bytResNum) Then
              'when new resources are added, status is set to dirty; for initial load,
              'need to reset them to false
              agViews(bytResNum).WritePropState = False
              agViews(bytResNum).IsDirty = False
            End If
          End Select
        End If
      Next i
    End If
  Next bytResType
  'return any warning codes
  LoadResources = blnWarnings
Exit Function

ErrHandler:
  
  'if error was invalid resource data, or missing vol file
  If Err.Number = vbObjectError + 595 Then
    'error trying to load this resource
    Err.Clear
    blnWarnings = True
    RecordLogEvent 0, strResID & " was skipped due to invalid resource data in " & JustFileName(strVolFile) & " at location " & CStr(lngLoc)
    Resume Next
  End If
  
  If Err.Number = vbObjectError + 606 Then
    Err.Clear
    blnWarnings = True
    RecordLogEvent 0, strResID & " was skipped because it's VOL file (" & JustFileName(strVolFile) & ") is missing."
    Resume Next
  End If
  
  'otherwise, stop loading resources
  strError = Err.Description
  strErrSrc = Err.Source
  lngError = Err.Number
  
  On Error GoTo 0: Err.Raise vbObjectError + 545, strErrSrc, Replace(LoadResString(545), ARG1, CStr(lngError) & ":" & strError)
End Function


Public Function BuildMIDI(SoundIn As AGISound) As Byte()

  Dim lngWriteTrack As Long
  Dim i As Long, j As Long
  Dim bytNote As Byte, intFreq As Integer, bytVol As Byte
  Dim lngTrackCount As Long, lngTickCount As Long
  Dim lngStart As Long, lngEnd As Long

  On Error Resume Next
  
  Debug.Assert SoundIn.Resource.Loaded

  'calculate track Count
  If Not SoundIn.Track(0).Muted And SoundIn.Track(0).Notes.Count > 0 Then
    lngTrackCount = 1
  End If
  If Not SoundIn.Track(1).Muted And SoundIn.Track(1).Notes.Count > 0 Then
    lngTrackCount = lngTrackCount + 1
  End If
  If Not SoundIn.Track(2).Muted And SoundIn.Track(2).Notes.Count > 0 Then
    lngTrackCount = lngTrackCount + 1
  End If
  If Not SoundIn.Track(3).Muted And SoundIn.Track(3).Notes.Count > 0 Then
    'add two tracks if noise is not muted
    'because the white noise and periodic noise
    'are written as two separate tracks
    lngTrackCount = lngTrackCount + 2
  End If

  'set intial size of midi data array
  ReDim mMIDIData(70 + 256)

  ' write header
  mMIDIData(0) = 77 '"M"
  mMIDIData(1) = 84 '"T"
  mMIDIData(2) = 104 '"h"
  mMIDIData(3) = 100 '"d"
  lngPos = 4


  WriteSndLong 6& 'remaining length of header (3 integers = 6 bytes)
  'write mode, trackcount and ppqn as bigendian integers
  WriteSndWord 1   'mode 0 = single track
                   'mode 1 = multiple tracks, all start at zero
                   'mode 2 = multiple tracks, independent start times
  'if no tracks,
  If lngTrackCount = 0 Then
    'need to build a 'null' set of data!!!
    
    ReDim Preserve mMIDIData(37)
    
    'one track, with one note of smallest length, and no sound
    WriteSndWord 1  'one track
    WriteSndWord 32 'pulses per quarter note
    'add the track info
    WriteSndByte 77 '"M"
    WriteSndByte 84 '"T"
    WriteSndByte 114 '"r"
    WriteSndByte 107 '"k"
    WriteSndLong 15 'track length
    'write the track number
    WriteSndDelta CLng(0)
    'write the set instrument status byte
    WriteSndByte &HC0
    'write the instrument number
    WriteSndByte 0
    'write a slight delay note with no volume to end
    WriteSndDelta 0
    WriteSndByte &H90
    WriteSndByte 60
    WriteSndByte 0
    WriteSndDelta 16
    WriteSndByte &H80
    WriteSndByte 60
    WriteSndByte 0
    'add end of track info
    WriteSndDelta 0
    WriteSndByte &HFF
    WriteSndByte &H2F
    WriteSndByte &H0
    'return
    BuildMIDI = mMIDIData()
    Exit Function
  End If
  
  WriteSndWord lngTrackCount
  
  'based on original playsound, agi sound tick is 1/64 sec
'''  WriteSndWord 32 'pulses per quarter note
                  'each tick of an AGI note is 1/64 of a second
                  'by default, MIDI defines a whole note as 2 seconds
                  'therefore, a quarter note is 1/2 second, or 32 ticks
  
  'based on NAGI, and by listening to sounds played on interpreter
  '(at least on WinXP), agi sound tick is 1/60 sec
  WriteSndWord 30 'pulses per quarter note
'''                  'each tick of an AGI note is 1/60 of a second
'''                  'by default, MIDI defines a whole note as 2 seconds
'''                  'therefore, a quarter note is 1/2 second, or 30 ticks

  'write the sound tracks
  For i = 0 To 2
    'if adding this instrument,
    If Not SoundIn.Track(i).Muted And SoundIn.Track(i).Notes.Count > 0 Then
      WriteSndByte 77 '"M"
      WriteSndByte 84 '"T"
      WriteSndByte 114 '"r"
      WriteSndByte 107 '"k"

      'store starting position for this track's data
      lngStart = lngPos
      WriteSndLong 0

      'write the track number
      WriteSndDelta CLng(lngWriteTrack)
      'write the set instrument status byte
      WriteSndByte &HC0 + lngWriteTrack
      'write the instrument number
      WriteSndByte SoundIn.Track(i).Instrument

      'step through notes in this track (5 bytes at a time)
      For j = 0 To SoundIn.Track(i).Notes.Count - 1
        'calculate note to play
        If SoundIn.Track(i).Notes(j).FreqDivisor > 0 Then
          'middle C is 261.6 HZ; from midi specs,
          'middle C is a note with a Value of 60
          'this requires a shift in freq of approx. 36.376
          'however, this offset results in crappy sounding music;
          'empirically, 36.5 seems to work best
          bytNote = CLng((Log10(111860# / CDbl(SoundIn.Track(i).Notes(j).FreqDivisor)) / dblLogConst) - 36.5)
          
          '
          'f = 111860 / (((Byte2 & 0x3F) << 4) + (Byte3 & 0x0F))
          '
          '(bytNote can never be <44 or >164)
          'in case note is too high,
          If bytNote > 127 Then
            bytNote = 127
          End If
          bytVol = 127 * (15 - SoundIn.Track(i).Notes(j).Attenuation) / 15
        Else
          bytNote = 0
          bytVol = 0
        End If

        'write NOTE ON data
        WriteSndDelta 0
        WriteSndByte &H90 + lngWriteTrack
        WriteSndByte bytNote
        WriteSndByte bytVol
        'write NOTE OFF data
        WriteSndDelta CLng(SoundIn.Track(i).Notes(j).Duration)
        WriteSndByte &H80 + lngWriteTrack
        WriteSndByte bytNote
        WriteSndByte 0
      Next j

      'write a slight delay note with no volume to end
      WriteSndDelta 0
      WriteSndByte &H90 + lngWriteTrack
      WriteSndByte 60
      WriteSndByte 0
      WriteSndDelta 16
      WriteSndByte &H80 + lngWriteTrack
      WriteSndByte 60
      WriteSndByte 0
      
      'add end of track info
      WriteSndDelta 0
      WriteSndByte &HFF
      WriteSndByte &H2F
      WriteSndByte &H0
      'save track end position
      lngEnd = lngPos
      'set cursor to start of track
      lngPos = lngStart
      'write the track length
      WriteSndLong (lngEnd - lngStart) - 4
      lngPos = lngEnd
    End If
    'increment track counter
    lngWriteTrack = lngWriteTrack + 1
  Next i

'seashore does a good job of imitating the white noise, with frequency adjusted empirically
'harpsichord does a good job of imitating the tone noise, with frequency adjusted empirically

  'if adding noise track,
  If Not SoundIn.Track(3).Muted And SoundIn.Track(3).Notes.Count > 0 Then
    'because there are two types of noise, must use two channels
    'one uses seashore (white noise)
    'other uses harpsichord (tone noise)

    For i = 0 To 1
      '0 means add tone
      '1 means add white noise

      WriteSndByte 77 '"M"
      WriteSndByte 84 '"T"
      WriteSndByte 114 '"r"
      WriteSndByte 107 '"k"

      'store starting position for this track's data
      lngStart = lngPos
      WriteSndLong 0       'place holder for chunklength

      'write track number
      WriteSndDelta CLng(lngWriteTrack)
      'write the set instrument byte
      WriteSndByte &HC0 + lngWriteTrack
      'write the instrument number
      Select Case i
      Case 0  'tone
        WriteSndByte 6 'harpsichord seems to be good simulation for tone
      Case 1  'white noise
        WriteSndByte 122 'seashore seems to be good simulation for white noise
        'crank up the volume
        WriteSndByte 0
        WriteSndByte &HB0 + lngWriteTrack
        WriteSndByte 7
        WriteSndByte 127
        'set legato
        WriteSndByte 0
        WriteSndByte &HB0 + lngWriteTrack
        WriteSndByte 68
        WriteSndByte 127
      End Select

      'reset tick counter (used in case of need to borrow track 3 freq)
      lngTickCount = 0
      
      '0 - add periodic
      '1 - add white noise
      For j = 0 To SoundIn.Track(3).Notes.Count - 1
        'add duration to tickcount
        lngTickCount = lngTickCount + SoundIn.Track(3).Notes(j).Duration
        'Fourth byte: noise freq and type
        '    In the case of the noise voice,
        '    7  6  5  4  3  2  1  0
        '
        '    1  .  .  .  .  .  .  .      Always 1.
        '    .  1  1  0  .  .  .  .      Register number in T1 chip (6)
        '    .  .  .  .  X  .  .  .      Unused, ignored; can be set to 0 or 1
        '    .  .  .  .  .  FB .  .      1 for white noise, 0 for periodic
        '    .  .  .  .  .  . NF0 NF1    2 noise frequency control bits
        '
        '    NF0  NF1       Noise Frequency
        '
        '     0    0         1,193,180 / 512 = 2330
        '     0    1         1,193,180 / 1024 = 1165
        '     1    0         1,193,180 / 2048 = 583
        '     1    1         Borrow freq from channel 3
        '
        'AGINote contains bits 2-1-0 only
        '
        'if this note matches desired type
        If (SoundIn.Track(3).Notes(j).FreqDivisor And 4) = 4 * i Then
          'if using borrow function:
          If (SoundIn.Track(3).Notes(j).FreqDivisor And 3) = 3 Then
            'get frequency from channel 3
            intFreq = GetTrack3Freq(SoundIn.Track(2), lngTickCount)
          Else
            'get frequency from bits 0 and 1
            intFreq = CInt(2330.4296875 / 2 ^ (SoundIn.Track(3).Notes(j).FreqDivisor And 3))
          End If

          'convert to midi note
          If (SoundIn.Track(3).Notes(j).FreqDivisor And 4) = 4 Then
            'for white noise, 96 is my best guess to imitate noise
            'BUT... 96 causes some notes to come out negative;
            '80 is max Value that ensures all AGI freq values convert
            'to positive MIDI note values
            bytNote = CByte((Log10(intFreq) / dblLogConst) - 80)
          Else
            'for periodic noise, 64 is my best guess to imitate noise
            bytNote = CByte((Log10(intFreq) / dblLogConst) - 64)
          End If
          'get volume
          bytVol = 127 * (15 - SoundIn.Track(3).Notes(j).Attenuation) / 15
        Else
          'write a blank
          bytNote = 0
          bytVol = 0
        End If

        'write NOTE ON data
        'no delta time
        WriteSndDelta 0
        'note on this track
        WriteSndByte &H90 + lngWriteTrack
        WriteSndByte CByte(bytNote)
        WriteSndByte CByte(bytVol)

        'write NOTE OFF data
        WriteSndDelta CLng(SoundIn.Track(3).Notes(j).Duration)
        WriteSndByte &H80 + lngWriteTrack
        WriteSndByte CByte(bytNote)
        WriteSndByte 0
      Next j

      'write a slight delay note with no volume to end
      WriteSndDelta 0
      WriteSndByte &H90 + lngWriteTrack
      WriteSndByte 60
      WriteSndByte 0
      WriteSndDelta 16
      WriteSndByte &H80 + lngWriteTrack
      WriteSndByte 60
      WriteSndByte 0

      'write end of track data
      WriteSndDelta 0
      WriteSndByte &HFF
      WriteSndByte &H2F
      WriteSndByte &H0
      'save ending position
      lngEnd = lngPos
      'go back to start of track, and write track length
      lngPos = lngStart
      WriteSndLong (lngEnd - lngStart) - 4
      lngPos = lngEnd
      'increment track counter
      lngWriteTrack = lngWriteTrack + 1
    Next i
  End If

  'remove any extra padding from the data array
  ReDim Preserve mMIDIData(lngPos)
  'return
  BuildMIDI = mMIDIData()
  'clear any errors
  Err.Clear
End Function

Private Sub WriteSndDelta(ByVal LongIn As Long)
  'writes variable delta times!!
  
  Dim i As Long
  i = LngSHR(LongIn, 21)
  If (i > 0) Then
    WriteSndByte (i And 127) Or 128
  End If
   
  i = LngSHR(LongIn, 14)
  If (i > 0) Then
    WriteSndByte (i And 127) Or 128
  End If
  
  i = LngSHR(LongIn, 7)
  If (i > 0) Then
    WriteSndByte (i And 127) Or 128
  End If
  
  WriteSndByte LongIn And 127
End Sub

Private Sub WriteSndWord(ByVal IntegerIn As Integer)
  WriteSndByte IntegerIn \ 256
  WriteSndByte IntegerIn And &HFF
End Sub

Private Sub WriteSndLong(ByVal LongIn As Long)
  
  WriteSndByte LongIn \ &H1000000
  WriteSndByte (LongIn \ &H10000) And &HFF
  WriteSndByte (LongIn \ &H100) And &HFF
  WriteSndByte (LongIn) And &HFF
End Sub

Private Function GetTrack3Freq(Track3 As AGITrack, ByVal lngTarget As Long) As Long
  'if noise channel needs the frequency of track 3,
  'must step through track three until the same point in time is found
  'then use that frequency for noise channel
  
  Dim i As Long, lngTickCount As Long
  
  'step through notes in this track (5 bytes at a time)
  For i = 0 To Track3.Notes.Count - 1
    'add duration
    lngTickCount = lngTickCount + Track3.Notes(i).Duration
    'if equal to or past current tick Count
    If lngTickCount >= lngTarget Then
      'this is the frequency we want
      GetTrack3Freq = Track3.Notes(i).FreqDivisor
      Exit Function
    End If
  Next i
  
  'if nothing found, return 0
      
End Function


Private Sub WriteSndByte(ByVal ByteIn As Byte)
  'Put intFile, , ByteIn
  mMIDIData(lngPos) = ByteIn
  lngPos = lngPos + 1
  'if at end
  If lngPos > UBound(mMIDIData) Then
    'jack it up
    ReDim Preserve mMIDIData(lngPos + 256)
  End If
End Sub


Private Function DecodeString(ByVal intCode As Integer) As String
'this function converts a code Value into its original string Value

Dim i  As Integer

  'initialize counter
  i = 0

  Do While (intCode > 255)
    'if code Value exceeds table size,
    If intCode > TABLE_SIZE Then
      Debug.Print "FATAL ERROR as  Invalid code (" & CStr(intCode) & ") in DecodeString."
      Exit Function
    Else
      'build string
      DecodeString = Chr$(bytAppend(intCode - 257)) & DecodeString
      intCode = intPrefix(intCode - 257)
    End If
  Loop
  DecodeString = Chr$(intCode) & DecodeString
End Function





Private Function InputCode(ByRef bytData() As Byte, intCodeSize As Integer, ByRef intPosIn As Integer) As Integer

  Dim lngWord As Long, lngRet As Long
  
  'this routine extracts the next code Value off the input stream
  'since the number of bits per code can vary between 9 and 12,
  'can't read in directly from the stream
  
  'unlike normal LZW, though, the bytes are actually written in so the code boundaries
  'work from right to left, NOT left to right. For example,an input stream that needs
  'to be split on a 9 bit boundary will use eight bits of first byte, plus LOWEST
  'bit of byte 2. The second code is then the upper seven bits of byte 2 and the lower
  '2 bits of byte 3 etc:
  '                          byte boundaries (8 bits per byte)
  '          byte4           byte3           byte2           byte1           byte0
  ' ...|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0
  ' ... x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x
  ' ... 3 2 1 0|8 7 6 5 4 3 2 1 0|8 7 6 5 4 3 2 1 0|8 7 6 5 4 3 2 1 0|8 7 6 5 4 3 2 1 0
  '                   code3             code2             code1             code0
  '                          code boundaries (9 bits per code)
  '
  'the data stream is read into a bit buffer 8 bits at a time (i.e. a single byte)
  'once the buffer is full of data, the input code is pulled out, and the buffer
  'is shifted.
  'the input data from the stream must be shifted to ensure it lines up with data
  'currently in the buffer.
  
  'read until the buffer is greater than 24-
  'this ensures that the eight bits read from the input stream
  'will fit in the buffer (which is a long integer==4 bytes==32 bits)
  'also stop reading data if end of data stream is reached)
  Do While (lngBitsInBuffer <= 24) And (intPosIn < lngOriginalSize)
    'get next byte
    lngWord = CLng(bytData(intPosIn))
    intPosIn = intPosIn + 1
    
    'shift the data to the left by enough bits so the byte being added will not
    'overwrite the bits currently in the buffer, and add the bits to the buffer
    lngBitBuffer = lngBitBuffer Or LngSHL(lngWord, lngBitsInBuffer)
    
    'increment Count of how many bits are currently in the buffer
    lngBitsInBuffer = lngBitsInBuffer + 8
  Loop
  
  'the input code starts at the lowest bit in the buffer
  'since the buffer has 32 bits total, need to clear out all bits above the desired
  'number of bits to define the code (i.e. if 9 bits, AND with &H1FF; 10 bits,
  'AND with &H3FF, etc.)
  lngRet = lngBitBuffer And (LngSHL(1, intCodeSize) - 1)
  'why not just use 2^intCodeSize -1 instead of the shift-left function?
  
  'now need to shift the buffer to the RIGHT by the number of bits per code
  lngBitBuffer = LngSHR(lngBitBuffer, intCodeSize)
  
  'adjust number of bits currently loaded in buffer
  lngBitsInBuffer = lngBitsInBuffer - intCodeSize
  
  'and return code Value
  InputCode = lngRet
End Function



Private Function NewCodeSize(intVal As Integer) As Integer
'this function supports the expansion of compressed resources
'it sets the size of codes which the LZW routine uses. The size
'of the code first starts at 9 bits, then increases as all code
'tables are filled. the max code size is 12; if an attempt is
'made to set a code size above that, the function does nothing
'the function also recalculates the maximum number of codes
'available to the expand subroutine. This number is used to
'determine when to call this function again.

  'max code size is 12 bits
  Const MAXBITS = 12
  
  If (intVal = MAXBITS) Then
    NewCodeSize = 11
    'this makes no sense!!!!
    'as written, it means max code size is really 11, not
    '12; an attempt to set it to 12 keeps it at 11???
    
  Else
    NewCodeSize = intVal
    lngMaxCode = LngSHL(1, NewCodeSize) - 2
   End If
End Function




Public Sub DecompressPicture(ByRef bytOriginalData() As Byte)

Dim intPosIn As Integer, intPosOut As Integer
Dim bytCurComp As Byte
Dim bytBuffer As Byte
Dim blnOffset As Boolean
Dim bytCurUncomp As Byte
Dim lngTempSize As Long
Dim bytExpandedData() As Byte
Dim lngTempCurPos As Long

  On Error GoTo ErrHandler
  
  'temporarily set size to max
  lngTempSize = MAX_RES_SIZE
  ReDim bytExpandedData(MAX_RES_SIZE)
  
  'reset variables
  intPosIn = 0
  intPosOut = 0
  blnOffset = False
  bytBuffer = 0
  lngTempCurPos = 0
  
  'decompress the picture
  Do
    'get current compressed byte
    bytCurComp = bytOriginalData(intPosIn)
    intPosIn = intPosIn + 1
    
    'if currently offset,
    If blnOffset Then
      'adjust buffer byte
      bytBuffer = bytBuffer + BytSHR(bytCurComp, 4)
      'extract uncompressed byte
      bytCurUncomp = bytBuffer
      'shift buffer back
      bytBuffer = BytSHL(bytCurComp, 4)
    Else
      'byte is not compressed
      bytCurUncomp = bytCurComp
    End If
    
    'save byte to temp resource
    bytExpandedData(lngTempCurPos) = bytCurUncomp
    lngTempCurPos = lngTempCurPos + 1
    
    'check if byte sets or restores offset
    If ((bytCurUncomp = &HF0) Or (bytCurUncomp = &HF2)) And (intPosIn < UBound(bytOriginalData) - 1) Then
      'if currently offset
      If blnOffset Then
        'write rest of buffer byte
        bytExpandedData(lngTempCurPos) = BytSHR(bytBuffer, 4)
        lngTempCurPos = lngTempCurPos + 1
        'restore offset
        blnOffset = False
      Else
        'get next byte
        bytCurComp = bytOriginalData(intPosIn)
        intPosIn = intPosIn + 1
        'save the byte, after shifting
        bytExpandedData(lngTempCurPos) = BytSHR(bytCurComp, 4)
        lngTempCurPos = lngTempCurPos + 1
        'fill buffer
        bytBuffer = BytSHL(bytCurComp, 4)
        blnOffset = True
      End If
    End If
  'continue until all original data has been read
  Loop Until intPosIn >= UBound(bytOriginalData) + 1

  'redim to arrays to actual size
  ReDim Preserve bytExpandedData(lngTempCurPos - 1)
  ReDim bytOriginalData(lngTempCurPos - 1)
  
  'copy arrays
  bytOriginalData = bytExpandedData
  
Exit Sub

ErrHandler:
  'all errors invalidate resource
  On Error GoTo 0: Err.Raise vbObjectError + 504, "ResourceFunctions.DecompressPicture", LoadResString(504)

End Sub



Public Function SndSubclassProc(ByVal hWnd As Long, ByVal uMsg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  
  Dim blnSuccess As Boolean
  Dim rtn As Long
  
  On Error Resume Next
  
  If Not blnPlaying And Not frmSndSubclass.agSndToPlay Is Nothing Then
    'set it to nothing
    Debug.Print "shouldn't be!"
  End If
  
  'check for mci msg
  Select Case uMsg
  Case MM_MCINOTIFY
    'determine success status
    blnSuccess = (wParam = MCI_NOTIFY_SUCCESSFUL)
    
    'close the sound
    rtn = mciSendString("close all", 0&, 0&, 0&)
    
    'raise the 'done' event
    frmSndSubclass.agSndToPlay.RaiseSoundComplete blnSuccess
    'agSnds(PlaySndResNum).RaiseSoundComplete blnSuccess
    
    'reset the flag
    blnPlaying = False
    
    'release the object
    Set frmSndSubclass.agSndToPlay = Nothing
  End Select
  
  'pass back to previous function
  SndSubclassProc = CallWindowProc(PrevSndWndProc, hWnd, uMsg, wParam, lParam)
End Function

Public Function UniqueResFile(ResType As AGIResType) As String

  Dim intNo As Integer
  
  On Error GoTo ErrHandler
  
  Do
    intNo = intNo + 1
    UniqueResFile = agResDir & "New" & agResTypeName(ResType) & CStr(intNo) & ".ag" & LCase$(Left$(agResTypeName(ResType), 1))
  Loop Until Dir(UniqueResFile) = vbNullString
Exit Function

ErrHandler:
  'on off chance that user has too many files
  Err.Clear
  On Error GoTo 0: Err.Raise vbObjectError, "ResourceFunctions.UniqueResFile", "Can't save- max number of files exceeded in directory. NO WAY YOU HAVE THAT MANY FILES!!!!!"
End Function


